//---------------------------------------------------------------------------

#ifndef EditAPIH
#define EditAPIH
//---------------------------------------------------------------------------

void Startup();
void Shutdown();

#ifdef EE_WINDOWS
#define EDIT_API  __declspec(dllexport) __stdcall
#else
#define EDIT_API // that is, render it merely empty space (Linux)
#endif

extern "C" int EDIT_API Edit_Initialize( int* smfID, const char* smf_name, char* smf_init_error, const int error_buffer_size, int* version_compare );
extern "C" int EDIT_API Edit_UpdateMetafile( const char* smf_name, char* smf_init_error, const int error_buffer_size );
extern "C" int EDIT_API Edit_Exit( const int smfID );

extern "C" int EDIT_API Edit_CreateNewMetafile( const char* smf_name );
extern "C" int EDIT_API Edit_CompactMetafile( const char* smf_name );
extern "C" int EDIT_API Edit_GetEngineVersion( char* version_buffer, int buffer_length );
extern "C" int EDIT_API Edit_GetMetafileStructureVersion(char* smf_structure_version, int smf_version_size);

extern "C" int EDIT_API Edit_GetMetafileVersion(int smfID, char* smf_version, int smf_version_size);
extern "C" int EDIT_API Edit_GetMetafileComment(int smfID, char* smf_comment, int smf_comment_size);

// Retrieve Engine errors
extern "C" int EDIT_API Edit_GetEngineErrorsCount( const int smfID, int* how_many_errors );
extern "C" int EDIT_API Edit_GetEngineErrorByIndex( const int smfID, const int idx, char* error_message, const int message_length);

// Primary key retrieval: Return all PKeys from a table
// First, using callback
extern "C" int EDIT_API Edit_GetPKeys( const int smfID, const char* table_name, const char* order_by, void* owner, void* callback_func );

// Or, get all of the data through cursor access functions

// For NON-PARAMETERIZED CURSORS, use:
extern "C" int EDIT_API Edit_OpenCursor( const int smfID, const char* cursor_name, const char* sql );
extern "C" int EDIT_API Edit_CloseCursor( const int smfID, const char* cursor_name );

// For PARAMETERIZED CURSORS, use:
extern "C" int EDIT_API Edit_PrepareParameterizedCursor( const int smfID, const char* cursor_name, const char* sql );
extern "C" int EDIT_API Edit_BindIntParameter( const int smfID, const char* cursor_name, const int param_num, const int value );
extern "C" int EDIT_API Edit_BindTextParameter( const int smfID, const char* cursor_name, const int param_num, const char* value );
extern "C" int EDIT_API Edit_OpenParameterizedCursor( const int smfID, const char* cursor_name );
extern "C" int EDIT_API Edit_ResetParameterizedCursor( const int smfID, const char* cursor_name );
extern "C" int EDIT_API Edit_ReleaseParameterizedCursor( const int smfID, const char* cursor_name );


// To retrieve data in EITHER parameterized OR non-parameterized cursors, use:
extern "C" int EDIT_API Edit_CursorRowCount( const int smfID, const char* cursor_name, int* row_count );
extern "C" int EDIT_API Edit_CursorColCount( const int smfID, const char* cursor_name, int* col_count );
extern "C" int EDIT_API Edit_CursorNext( const int smfID, const char* cursor_name );
extern "C" int EDIT_API Edit_CursorEof( const int smfID, const char* cursor_name, int* is_eof );
// Use a callback to retrieve the data from all columns of the current cursor tuple
extern "C" int EDIT_API Edit_ReadCursorData( const int smfID, const char* cursor_name, void* owner, void* callback_func );
// Or, get the data without a callback
// Option 1: get column data referencing by column number
extern "C" int EDIT_API Edit_GetCursorDataSizeByNumber( const int smfID, const char* cursor_name, const int col_number, int* data_size );
extern "C" int EDIT_API Edit_GetCursorDataByNumber( const int smfID, const char* cursor_name, const int col_number, char* buffer, const int buff_size );
// Option 2: get column data referencing by field name
extern "C" int EDIT_API Edit_GetCursorDataSizeByName( const int smfID, const char* cursor_name, const char* field_name, int* data_size );
extern "C" int EDIT_API Edit_GetCursorDataByName( const int smfID, const char* cursor_name, const char* field_name, char* buffer, const int buff_size );
// Let's allow for getting integer fields returned as integers... we pull PKey all the time!
extern "C" int EDIT_API Edit_GetCursorIntByNumber( const int smfID, const char* cursor_name, const int col_number, int* data );
extern "C" int EDIT_API Edit_GetCursorIntByName( const int smfID, const char* cursor_name, const char* field_name, int* data );

// Verify that all fields referenced by all edits in an edit set
// are included in the field list of a layout
extern "C" int EDIT_API Edit_CheckEditSetAgainstLayout( const int smfID, const int edit_set_pkey, const int layout_pkey, void* owner, void* callback_func );

// Get the stats about tables in the metafile.
// First, get the table names (uses a callback)
extern "C" int EDIT_API Edit_GetTableNames( const int smfID, void* owner, void* callback_func );
// Alternative means of getting the table names (no callback):
extern "C" int EDIT_API Edit_GetTablesCount( const int smfID, int* how_many_tables );
extern "C" int EDIT_API Edit_GetTableNameByIndex( const int smfID, const int idx, char* buffer, const int buff_size );

// Now, for each table, get the table structure.
// First, use a callback
extern "C" int EDIT_API Edit_GetTableStructure( const int smfID, const char* table_name, void* owner, void* callback_func );
// Alternative means of getting the table structure, without a callback
extern "C" int EDIT_API Edit_GetColumnsCount( const int smfID, const char* table_name, int* how_many_columns );
extern "C" int EDIT_API Edit_GetColumnStructure( const int smfID, const char* table_name, const int column_order, char* field_name, const int fldname_size, char* data_type, const int type_size, char* nullability, const int null_size, int* is_trigger_field);

// Get list of indexes for a table
extern "C" int EDIT_API Edit_GetIndexCount( const int smfID, const char* table_name, int* how_many_indexes );
extern "C" int EDIT_API Edit_GetIndex( const int smfID, const char* table_name, const int idx, char* index_sql, const int buff_size );

// Get list of triggers for a table
extern "C" int EDIT_API Edit_GetTriggerCount( const int smfID, const char* table_name, int* how_many_triggers );
extern "C" int EDIT_API Edit_GetTrigger( const int smfID, const char* table_name, const int idx, char* trigger_sql, const int buff_size );


// Other service functions
// Give me a table and field name, I'll tell you its column order
extern "C" int EDIT_API Edit_GetColumnOrder( const int smfID, const char* table_name,
	const char* field_name, int* col_order );
// Give me a table and column order, I'll tell you its field name
extern "C" int EDIT_API Edit_GetFieldName( const int smfID, const char* table_name,
	const int col_order, char* field_name, const int fldname_size );


// Read data from any table. The PKey identifies the row in the named table.
// First, using a callback
extern "C" int EDIT_API Edit_ReadData( const int smfID, const char* table_name, const int pkey, void* owner, void* callback_func );

// Alternative means of getting data, without a callback

// Get a "fresh look" at the data from the table; this updates the query maintained
// by the Engine for efficient calls to the Edit_GetDataBy* functions
extern "C" int EDIT_API Edit_RefreshReadData( const int smfID, const char* table_name, const int pkey );
// First, requesting the information by the column order
extern "C" int EDIT_API Edit_GetDataSizeByColumnOrder( const int smfID, const char* table_name, const int pkey, const int col_order, int* data_size );
extern "C" int EDIT_API Edit_GetDataByColumnOrder( const int smfID, const char* table_name, const int pkey, const int col_order, char* buffer, const int buff_size );
// Or, if you know the field name
extern "C" int EDIT_API Edit_GetDataSizeByFieldName( const int smfID, const char* table_name,
	const int pkey, const char* field_name, int* data_size );
extern "C" int EDIT_API Edit_GetDataByFieldName( const int smfID, const char* table_name,
	const int pkey, const char* field_name, char* buffer, const int buff_size );

// FUNCTIONS FOR Running Edits
// Load a record of case data into the Engine and prepare it for editing.
// First, using a Record Layout Tag to identify the layout...
extern "C" int EDIT_API Edit_UpdateRecordByTag( const int smfID, const char* layout_tag, const char* data );
// Alternatively, using the Record Layout Name to identify the layout
extern "C" int EDIT_API Edit_UpdateRecordByName( const int smfID, const char* layout_name, const char* data );

// Now retrieve the contents of the record (GenEDITS Plus does this to display a parsed record)
// First, using a Record Layout Tag to identify the layout, and feeding results back using a callback
extern "C" int EDIT_API Edit_DisplayDataRecordByTag( const int smfID, const char* layout_tag, void* owner, void* callback_func );
// Alternatively, using the Record Layout Name to identify the layout
extern "C" int EDIT_API Edit_DisplayDataRecordByName( const int smfID, const char* layout_name, void* owner, void* callback_func );

// For vendors who change the names of standard (e.g., NAACCR) data items
extern "C" int EDIT_API Edit_DisplayFieldInMessagesAs( const int smfID, const int field_number, const char* display_name );

// Over-ride EE_SKIPEMPTY edit or edit set option for specific fields
extern "C" int EDIT_API Edit_NeverSkipEmptyField( const int smfID, const int field_number, const int never_skip_empty);

// To retrieve the data information without using a callback, request each field by number
extern "C" int EDIT_API Edit_FieldDataByNumber( const int smfID,
	const char* layout_tag, const int field_number, int* pkey,
	int* start_pos, int* field_length,
	char* field_name, const int field_name_len,
	char* display_name, const int display_name_len,
	char* data_type, const int data_type_len,
	char* value, const int value_len );

// If caller wants to perform surgery on edit set prior to calling Edit_RunEdits,
// then first call Edit_LoadEditSet. ALSO NEEDED for resetting counters to retrieve
// summary totals and other edit details via Edit_GetRunEditsSummary
extern "C" int EDIT_API Edit_LoadEditSet( const int smfID, const char* edit_set_tag, const char* layout_tag );
extern "C" int EDIT_API Edit_UnloadEditSet( const int smfID, const char* edit_set_tag, const char* layout_tag );
extern "C" int EDIT_API Edit_GetRunEditsSummary( const int smfID, const char* edit_set_tag,
	const char* layout_tag, const int include_zero_counts,
    int* total_errors, int* total_warnings, int* total_skips, int* total_messages,
	int* total_warning_messages, void* owner, void* callback_func );

// Set/Get options for specific edits prior to a call to Edit_RunEdits
extern "C" int EDIT_API Edit_SetEditOptions( const int smfID, const char* edit_set_tag, const char* edit_tag,
	const int suppress_edit, const int skip_if_empty, const int skip_if_failed,
	const int no_skips_override );
extern "C" int EDIT_API Edit_GetEditOptions( const int smfID, const char* edit_set_tag, const char* edit_tag,
	int* is_suppressed, int* is_skip_is_empty, int* is_skip_is_failed,
	int* is_no_skips_override );
extern "C" int EDIT_API Edit_ClearEditOptions( const int smfID, const char* edit_set_tag );

// This is the all-in-one function to run edits
extern "C" int EDIT_API Edit_RunEdits( const int smfID, const char* edit_set_tag, const char* layout_tag, const char* data, const int edit_options, int* errors_count, void* owner, void* callback_func);
// Need a version that works without callbacks.
// Also need Edit_RunSingleFieldEdits and Edit_RunEditByName

// This is how GenEDITS Plus retrieves fields and values for a case
// callback func uses CallbackDisplayDataRecord format
extern "C" int EDIT_API Edit_GetEditDetails(const int smfID, const char* edit_tag, void* owner, void* callback_func);

extern "C" int EDIT_API Edit_DebugEdit(const int smfID, const char* edit_tag, const char* layout_tag, const char* data, void* owner, void* logic_callback, void* variables_callback, void* fields_callback, void* notes_callback);

extern "C" int EDIT_API Edit_RecompileEditLogic( const int smfID, const char* edit_tag );
extern "C" int EDIT_API Edit_UpdateAllAccessField( const int smfID, const char* table_name, const int pkey, const char* field_name, const char* value);

#ifdef __SQLEW__

/*  The remaining API functions are for use only by software that will write
 *  to the metafile. They control logging in and maintaining user access rights
 *  to modify objects in the metafile.
 */
extern "C" int EDIT_API Ew_Login( const int smfID, const char* agency_code, const char* password );
extern "C" int EDIT_API Ew_GetLoginsCount( const int smfID, int* how_many_logins );
extern "C" int EDIT_API Ew_GetLoginByIndex( const int smfID, const int idx, int* agency_pkey, char* agency_code, const int agency_code_size);

extern "C" int EDIT_API Ew_ValidateUnique( const int smfID, const char* table_name, const int pkey, const char* field_name, const char* value, int* data_is_unique, int* dupe_pkey );

extern "C" int EDIT_API Ew_BeginTrans( const int smfID  );
extern "C" int EDIT_API Ew_CommitTrans( const int smfID  );
extern "C" int EDIT_API Ew_RollbackTrans( const int smfID  );

extern "C" int EDIT_API Ew_SavePoint( const int smfID, const char* savepoint_name);
extern "C" int EDIT_API Ew_ReleaseSavePoint( const int smfID, const char* savepoint_name);
extern "C" int EDIT_API Ew_RollbackToSavePoint( const int smfID, const char* savepoint_name);

extern "C" int EDIT_API Ew_Edit( const int smfID, const char* table_name, const int pkey  );
extern "C" int EDIT_API Ew_Insert( const int smfID, const char* table_name  );
extern "C" int EDIT_API Ew_Delete( const int smfID, const char* table_name, const int delete_pkey  );
extern "C" int EDIT_API Ew_Post( const int smfID, const char* table_name, int* new_pkey  );
extern "C" int EDIT_API Ew_Cancel( const int smfID, const char* table_name  );

extern "C" int EDIT_API Ew_DeleteEditSetsPivot( const int smfID, const int editset_pkey, const int edit_pkey );
extern "C" int EDIT_API Ew_AddEditSetsPivot( const int smfID, const int editset_pkey, const int edit_pkey );

extern "C" int EDIT_API Ew_DeleteLayoutsPivot( const int smfID, const int layout_pkey, const int field_pkey );
extern "C" int EDIT_API Ew_AddLayoutsPivot( const int smfID, const int layout_pkey, const int field_pkey, const int start_pos );

extern "C" int EDIT_API Ew_SetFieldByName( const int smfID, const char* table_name, const char* field_name, const char* value  );
extern "C" int EDIT_API Ew_SetFieldByNumber( const int smfID, const char* table_name, const int column_number, const char* value );

extern "C" int EDIT_API Ew_CheckSyntax( const int smfID, const char* edit_logic, int* how_many_errors, int* how_many_warnings );
extern "C" int EDIT_API Ew_GetSyntaxError( const int smfID, const int error_idx, int* line_number, char* syntax_error, const int syntax_error_len);
extern "C" int EDIT_API Ew_GetSyntaxWarning( const int smfID, const int warning_idx, int* line_number, char* syntax_warning, const int syntax_warning_len);
// To retrieve Field, Table and Message items referenced in edit logic
extern "C" int EDIT_API Ew_GetLogicListCount(const int smfID, const char which_list, int* how_many_items);
extern "C" int EDIT_API Ew_GetLogicFieldNumber(const int smfID, const int idx, int* field_number);
extern "C" int EDIT_API Ew_GetLogicTableName(const int smfID, const int idx, char* table_name, const int buff_size);
extern "C" int EDIT_API Ew_GetLogicErrorMessage(const int smfID, const int idx, int* msg_func_token, char* error_text, const int buff_size);
extern "C" int EDIT_API Ew_GetLogicMessageNumber(const int smfID, const int idx, int* msg_func_token, int* msg_number);

extern "C" int EDIT_API Ew_CopyLookupTable(const int smfID, const char* source_table_name, const char* target_table_name, const int agency_pkey);

// Support for Import Metafile
extern "C" int EDIT_API Ew_OpenImportMetafile(const int smfID, const char* source_smf_name);
extern "C" int EDIT_API Ew_CloseImportMetafile(const int smfID);
extern "C" int EDIT_API Ew_ImportLookupTable( const int smfID, const char* table_name);
extern "C" int EDIT_API Ew_GetImportCandidates( const int smfID, const char* table_name, const char* match_on_field_name, void* owner, void* callback_func);
extern "C" int EDIT_API Ew_ImportWritePassword( const int smfID, const int agency_pkey, const char* the_password);


extern "C" int EDIT_API Ew_ClearEditLogicPivots( const int smfID, const int edit_pkey );

// Support for creating new/restructuring existing lookup tables
extern "C" int EDIT_API Ew_DefineTableStructure( const int smfID, const char* table_name, const int restructure_existing );
extern "C" int EDIT_API Ew_AddColumnDef(
   const int smfID,
   const char* col_name,
   const char* copy_from_col_name,
   const int data_type,
   const int data_length,
   const int is_nullable,
   const int is_trigger_field,
   const char* check_limit_to_list,
   const char* check_range_low,
   const char* check_range_high,
   const char* default_value,
   const int index_status);

extern "C" int EDIT_API Ew_CreateTable( const int smfID, int* lookup_pkey );


#endif  // __SQLEW__

#endif  // EditAPIH


