//---------------------------------------------------------------------------
#include <cstdlib> // std::atoi
#include <stdio.h>    // for sprintf
#include <algorithm>  // transform()
#include <stack>
#include <stdexcept>
#include <cstring>   // memset, strncat
#include <memory>    // smart pointers
// For the Borland compilers, use #pragma hdrstop for precompiled headers (pch)
#ifdef BCB
	#pragma hdrstop
//#else
	// For VC++, use header stdafx.h
	//#include "stdafx.h"
#endif
//---------------------------------------------------------------------------

#include "EditCompiler.h"
#include "SQLDB.h"
#include "SQLQuery.h"
#include "EngineErrorsPump.h"
#include "Utility.h"
#include "TableStats.h"
//---------------------------------------------------------------------------

#ifdef __SQLEW__

// used by tokenize()
static EEt_TokType tokgenlist [] = {

	  {"!",          NOT,     OPERATOR,'!'},
	  {"not",        NOT,     OPERATOR,'!'},
	  {"\"",         (TTokenID) STRING,  STRING,'\"'},
			 /* " + xx(length of string) + len bytes of string */
	  {"#",          FIELD,   APPFIELD,'#'},  /* # + field_num */
	  /* $ */
	  {"%",          MOD,     OPERATOR,'%'},
	  {"mod",        MOD,     OPERATOR,'%'},
	  {"&&",         AND,     OPERATOR,'&'}, /* 0x26 */
	  {"and",        AND,     OPERATOR,'&'}, /* 0x26 */
	  {"'",          CCONST,  CHARCONST,'\''}, /* 0x27 ' + char */
	  {"(",          LPAREN,  DELIMITER,'('}, /* 0x28 */
	  {")",          RPAREN,  DELIMITER,')'}, /* 0x29 */
	  {"*",          MULT,    OPERATOR,'*'},
	  {"+",          ADD,     OPERATOR,'+'},
	  {",",          COMMA,   OPERATOR,','},
	  {"-",          SUB,     OPERATOR,'-'},
	  {".",          PERIOD,  DELIMITER,'.'},
	  {"/",          DIV,     OPERATOR,'/'}, /* 0x2F */
	  /* break for numbers */
	  {":",          COLON,   DELIMITER,':'}, /* 0x3A */
	  {";",          SCOLON,  DELIMITER,';'}, /* 0x3B */
	  {"<",          LT,      OPERATOR,'<'}, /* 0x3C */
	  {"=",          ASSIGN,  OPERATOR,'='},
	  {">",          GT,      OPERATOR,'>'}, /* 0x3E */
	  {"else",       ELSE,    CONSTRUCT,  'A'}, /* 0x41 */
	  {"while",      WHILE,   CONSTRUCT,  'B'},
	  {"for",        FOR,     CONSTRUCT,  'C'},
	  {"return",     RETURN,  KEYWORD, 'D'},
	  {"==",         EQ,      OPERATOR,'E'},
	  {"if",         IF,      CONSTRUCT,  'F'},
	  {">=",         GE,      OPERATOR,'G'},
	  {"char",       EEi_CHAR,      KEYWORD, 'H'},
	  {"int",        EEi_INT,    KEYWORD, 'I'},
	  {"long",       EELONG,     KEYWORD, 'I'},
	  {"tablevar",   EEi_CHAR,      KEYWORD, 'H'},
	  {"!=",         NE,      OPERATOR,   'J'},
	  {"<=",         LE,      OPERATOR,   'K'},
	  {"do",         DO,      CONSTRUCT,  'L'},
		/* 0x4D - M + function ID letter */
	  {"libfnc",     LIBFUNC, FUNCTION,   'M'},
		/* N + xx(length of id) +	len bytes of name */
	  {"var",        VAR,     IDENTIFIER, 'N'},
	  {"dt_today",   (TTokenID) 'O',     NUMCONST,   'O'},
	  {"dt_unknown", (TTokenID) 'P',     NUMCONST,   'P'},
	  {"fail",       (TTokenID) '0',     NUMCONST,   'Q'},
	  {"pass",       (TTokenID) '1',     NUMCONST,   'R'},
	  {"dt_max",     (TTokenID) '2',     NUMCONST,   'S'},
	  {"dt_exact",   (TTokenID) '3',     NUMCONST,   'T'},
		/* U + xx(length of number) +  len bytes of number */
	  {"number",     (TTokenID) 'U',     NUMBER,     'U'},
      {"dt_empty",   (TTokenID) 'W',     NUMCONST,   'W'}, // date IOP functions
	  {"dt_valid",   (TTokenID) 'X',     NUMCONST,   'X'},
	  {"false",      (TTokenID) '0',     NUMCONST,   'Q'},
	  {"off",        (TTokenID) '0',     NUMCONST,   'Q'},
	  {"right",      (TTokenID) '0',     NUMCONST,   'Q'}, /* EEi_TRIMRIGHT */
	  {"on",         (TTokenID) '1',     NUMCONST,   'R'},
	  {"true",       (TTokenID) '1',     NUMCONST,   'R'},
	  {"dt_min",     (TTokenID) '1',     NUMCONST,   'R'},
	  {"dt_missing", (TTokenID) 'Y',     NUMCONST,   'Y'},
	  {"left",       (TTokenID) '1',     NUMCONST,   'R'}, /* EEi_TRIMLEFT */
	  {"warn",       (TTokenID) '2',     NUMCONST,   'S'},
	  {"step",       (TTokenID) '2',     NUMCONST,   'S'},
	  {"dt_error",   (TTokenID) 'Z',     NUMCONST,   'Z'},
	  {"both",       (TTokenID) '2',     NUMCONST,   'S'}, /* EEi_TRIMBOTH */
	  {"dt_day_empty",   (TTokenID) 'y',     NUMCONST,   'y'}, // date_day_iop
	  {"dt_month_empty", (TTokenID) 'z',     NUMCONST,   'z'}, // date_month_iop
	 {"\x5b",          LBRACKET,DELIMITER,  LBRACKET}, /* 0x5B [ */
	 {"\x5d",          RBRACKET,DELIMITER,  RBRACKET}, /* 0x5D ] */
		/* 0x6D - m + function ID letter */
	  {"funccall",   UDFUNCCALL, FUNCTION,   'm'},
	  {"function",   UDFUNCDECL, FUNCTION,   'n'}, /* 0x6E */
	  {"{",          LBRACE,  BLOCK,      '{'}, /* 0x7B */
	  {"or",         OR,      OPERATOR,   '|'}, /* 0x7C */
	  {"||",         OR,      OPERATOR,   '|'}, /* 0x7C */
	  {"}",          RBRACE,  BLOCK,      '}'}, /* 0x7D */
};

// used by get_next_token()
static EEt_TokType toklist [] = {
	  {"!",          NOT,     OPERATOR,'!'},
	  {"\"",         (TTokenID) STRING,  STRING,'\"'},
			 /* " + xx(length of string) + len bytes of string */
	  {"#",          FIELD,   APPFIELD,'#'},  /* # + field_num */
	  /* $ */
	  {"mod",        MOD,     OPERATOR,'%'},
	  {"&&",         AND,     OPERATOR,'&'}, /* 0x26 */
	  {"'",          CCONST,  CHARCONST,'\''}, /* 0x27 ' + char */
	  {"(",          LPAREN,  DELIMITER,'('}, /* 0x28 */
	  {")",          RPAREN,  DELIMITER,')'}, /* 0x29 */
	  {"*",          MULT,    OPERATOR,'*'},
	  {"+",          ADD,     OPERATOR,'+'},
	  {",",          COMMA,   OPERATOR,','},
	  {"-",          SUB,     OPERATOR,'-'},
	  {".",          PERIOD,  DELIMITER,'.'},
	  {"/",          DIV,     OPERATOR,'/'}, /* 0x2F */
	  /* break for numbers */
	  {":",          COLON,   DELIMITER,':'}, /* 0x3A */
	  {";",          SCOLON,  DELIMITER,';'}, /* 0x3B */
	  {"<",          LT,      OPERATOR,'<'}, /* 0x3C */
	  {"=",          ASSIGN,  OPERATOR,'='},
	  {">",          GT,      OPERATOR,'>'}, /* 0x3E */
	  {"else",       ELSE,    CONSTRUCT,  'A'}, /* 0x41 */
	  {"while",      WHILE,   CONSTRUCT,  'B'},
	  {"for",        FOR,     CONSTRUCT,  'C'},
	  {"return",     RETURN,  KEYWORD, 'D'},
	  {"==",         EQ,      OPERATOR,'E'},
	  {"if",         IF,      CONSTRUCT,  'F'},
	  {">=",         GE,      OPERATOR,'G'},
	  {"char",       EEi_CHAR,      KEYWORD, 'H'},
	  {"int",        EEi_INT,    KEYWORD, 'I'},
	  {"!=",         NE,      OPERATOR,   'J'},
	  {"<=",         LE,      OPERATOR,   'K'},
	  {"do",         DO,      CONSTRUCT,  'L'},
		/* 0x4D - M + function ID letter */
	  {"libfnc",     LIBFUNC, FUNCTION,   'M'},
		/* N + xx(length of id) +	len bytes of name */
	  {"var",        VAR,     IDENTIFIER, 'N'},
	  {"dt_today",   (TTokenID) 'O',     NUMCONST,   'O'},
	  {"dt_unknown", (TTokenID) 'P',     NUMCONST,   'P'},
	  {"fail",       (TTokenID) '0',     NUMCONST,   'Q'},
	  {"pass",       (TTokenID) '1',     NUMCONST,   'R'},
	  {"dt_max",     (TTokenID) '2',     NUMCONST,   'S'},
	  {"dt_exact",   (TTokenID) '3',     NUMCONST,   'T'},
		/* U + xx(length of number) +  len bytes of number */
	  {"number",     (TTokenID) 'U',     NUMBER,     'U'},
	  {"\x5b",          LBRACKET,DELIMITER,  LBRACKET}, /* 0x5B [ */
	  {"\x5d",          RBRACKET,DELIMITER,  RBRACKET}, /* 0x5D ] */
		/* 0x6D - m + function ID letter */
	  {"funccall",   UDFUNCCALL, FUNCTION,   'm'},
	  {"function",   UDFUNCDECL, FUNCTION,   'n'}, /* 0x6E */
	  {"{",          LBRACE,  BLOCK,      '{'}, /* 0x7B */
	  {"||",         OR,      OPERATOR,   '|'}, /* 0x7C */
	  {"}",          RBRACE,  BLOCK,      '}'}, /* 0x7D */

};

static EEt_IntFuncListType intern_funclist[] = {
	{"at", 'A', EEi_INT},                   // 'A'
	{"empty", 'K', EEi_INT},                // 'K'
	{"error_msg", 'L', EEi_INT},            // 'L'
	{"error_text", 'M', EEi_INT},           // 'M'
	{"fmtstr", 'N', EEi_CHAR},              // 'N'
	{"ilookup", 'S', EEi_INT},              // 'S'
	{"inlist", 'U', EEi_INT},               // 'U'
	{"justified", 'V', EEi_INT},            // 'V'
	{"left", 'W', EEi_CHAR},                // 'W'
	{"lookup", 'X', EEi_INT},               // 'X'
	{"lower", 'Y', EEi_CHAR},               // 'Y'
	{"match", 'Z', EEi_INT},                // 'Z'
	{"nameexpr", ']', EEi_CHAR},            // ']'
	{"right", 'e', EEi_CHAR},               // 'e'
	{"rlookup", 'f', EEi_INT},              // 'f'
	{"save_error_text", 'g', NOTHING},      // 'g' 103
	{"save_text", 'h', NOTHING},            // 'h' 104
	{"setvar", 'k', NOTHING},               // 'k'
	{"strcat", 'm', EEi_CHAR},              // 'm'
	{"strcmp", 'n', EEi_INT},               // 'n'
	{"strcpy", 'o', EEi_CHAR},              // 'o'
	{"strlen", 'p', EEi_INT},               // 'p'
	{"substr", 'q', EEi_CHAR},              // 'q'
	{"trim", 's', EEi_CHAR},                // 's'
	{"upper", 't', EEi_CHAR},               // 't'
	{"val", '', EEi_INT},                  // ''
	{"set_warning", '', NOTHING},          // ''
	{"save_warning_text", '', NOTHING},    // '' 133
	{"externaldll", '', EEi_INT},          // ''
	{"sqllookup", '{', EEi_INT},            // '{' 123 (re-using usr5)
	{"sqlrangelookup", '}', EEi_INT},       // '}' 125 (re-using usr7)
// EDITS40 Date Functions
	{"date_day_iop", '', EEi_INT},          // '' Alt+0136
	{"date_month_iop", '', EEi_INT},        // '' Alt+0137
	{"date_year_iop", '', EEi_INT},         // '' Alt+0138
	{"datecmp_iop", '', EEi_INT},           // '' Alt+0139
	{"daydiff_iop", '', EEi_INT},           // '' Alt+0140
	{"dayinterval_iop", '', EEi_INT},       // '' Alt+0141
	{"monthdiff_iop", '', EEi_INT},         // '' Alt+0142
	{"monthinterval_iop", '', EEi_INT},     // '' Alt+0143
	{"valid_date_iop", '', EEi_INT},        // '' Alt+0144
	{"yeardiff_iop", '', EEi_INT},          // '' Alt+0145
	{"yearinterval_iop", '', EEi_INT},      // '' Alt+0146
	{"allow_future_date_iop", '', NOTHING}, // '' Alt+0147
	""};
/*  DEPRECATED EDITS LANGUAGE FUNCTIONS. Re-use the tokens if needed for new ones
	{"binlookup", 'B', EEi_INT},            // 'B'
	{"check_flag", 'C', EEi_INT},           // 'C'
	{"clear_flags", 'D', NOTHING},          // 'D'
	{"copyrecord", '', NOTHING},           // ''
	{"date_day", 'E', EEi_INT},             // 'E'
	{"date_month", 'F', EEi_INT},           // 'F'
	{"date_year", 'G', EEi_INT},            // 'G'
	{"datecmp", 'H', EEi_INT},              // 'H'
	{"daydiff", 'I', EEi_INT},              // 'I'
	{"dayinterval", 'J', EEi_INT},          // 'J'
	{"monthdiff", '[', EEi_INT},            // '['
	{"monthinterval", '\\', EEi_INT},       // '\\'
	{"getche", 'O', EEi_CHAR},              // 'O'
	{"getnum", 'Q', EEi_INT},               // 'Q'
	{"getfield", 'P', EEi_CHAR},            // 'P'
	{"getvar", 'R', NOTHING},               // 'R'
	{"incr_flag", 'T', NOTHING},            // 'T'
	{"print", '^', NOTHING},                // '^'
	{"putch", '_', NOTHING},                // '_'
	{"putfield", '`', EEi_INT},             // '`'
	{"puts", 'a', NOTHING},                 // 'a'
	{"quit", 'b', NOTHING},                 // 'b'
	{"read", 'c', EEi_INT},                 // 'c'
	{"recordempty", 'd', EEi_INT},          // 'd'
	{"set_error", 'i', NOTHING},            // 'i'
	{"set_flag", 'j', NOTHING},             // 'j'
	{"showvars", 'l', NOTHING},             // 'l'
	{"trace", 'r', NOTHING},                // 'r'
	{"use", 'u', NOTHING},                  // 'u'
	{"usr2", 'x', EEi_INT},                 // 'x'
	{"usr4", 'z', EEi_INT},                 // 'z'
	{"usr1", 'v', EEi_INT},                 // 'v'
	{"usr10", 'w', EEi_INT},                // 'w'
	{"usr3", 'y', EEi_INT},                 // 'y'
	{"usr6", '|', EEi_INT},                 // '|'
	{"usr8", '~', EEi_INT},                 // '~'
	{"usr9", '', EEi_INT},                 // ''
	{"valid_date", '', EEi_INT},           // '' (note, this is NOT a blank: Alt+0129)
	{"yeardiff", '', EEi_INT},             // ''
	{"yearinterval", '', EEi_INT},         // ''
*/
/*   These are called by the destructor (note that they are not, and
 *   cannot be, members of a class). See note in destructor.
 */
void DeleteUDFDef(EEt_UDFDef* udf)
{
	if (udf)
	{
		delete udf;
		udf = NULL;
	}
}
void ClearUsedFlag(EEt_UDFDef* udf)
{
	udf->is_used_in_edit = false;
}
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
/*************************
 *** TVariableDef
 *************************/
TVariableDef::TVariableDef()
{
	FVarName.assign("");
	FVarType = '?'; // unassigned, not known
	FVarSize = 1;   // default
	FVarValue.assign("");
	FIsReported = false;
} // constructor
//---------------------------------------------------------------------------
TVariableDef::~TVariableDef()
{
	// in case I need it
} // destructor
//---------------------------------------------------------------------------
std::string TVariableDef::GetVarName()
{
	return FVarName;
}
//---------------------------------------------------------------------------
void TVariableDef::SetVarName(std::string var_name)
{
	FVarName.assign(var_name);
}
//---------------------------------------------------------------------------
char TVariableDef::GetVarType()
{
	return FVarType;
}
//---------------------------------------------------------------------------
void TVariableDef::SetVarType(char var_type)
{
	if ((var_type == EEi_INT) ||
		(var_type == EEi_CHAR))
	{
		FVarType = var_type;
	}
	else
	{
		throw std::runtime_error("TVariableDef::SetVarType - invalid var_type");
	}
}
//---------------------------------------------------------------------------
size_t TVariableDef::GetVarSize()
{
	return FVarSize;
}
//---------------------------------------------------------------------------
void TVariableDef::SetVarSize(size_t var_size)
{
	if (var_size > 0)
	{
		FVarSize = var_size;
	}
	else
	{
		throw std::runtime_error("TVariableDef::SetVarSize - var_size must be > 0");
	}
}
//---------------------------------------------------------------------------
std::string TVariableDef::GetVarValue()
{
	return FVarValue;
}
//---------------------------------------------------------------------------
void TVariableDef::SetVarValue(std::string var_value)
{
	FVarValue.assign(var_value);
}
//---------------------------------------------------------------------------
bool TVariableDef::GetIsReported()
{
	return FIsReported;
}
//---------------------------------------------------------------------------
void TVariableDef::SetIsReported(bool is_reported)
{
	FIsReported = is_reported;
}
//---------------------------------------------------------------------------
/*************************
 *** TVariablesMap
 *************************/
TVariablesMap::TVariablesMap()
{
	// just instantiate it
} // constructor
//---------------------------------------------------------------------------
TVariablesMap::~TVariablesMap()
{
	// Delete the instances of variables
	ClearVariables();
}
//---------------------------------------------------------------------------
void TVariablesMap::ClearVariables()
{
	std::map<std::string, TVariableDef*>::iterator it =
		FVariables.begin();
	while (it != FVariables.end())
	{
		delete (*it).second;
		it++;
	}

	// resize to empty
	FVariables.clear();
} // ClearVariables
//---------------------------------------------------------------------------
TVariableDef* TVariablesMap::GetVariable(std::string var_name)
{
	lowercase(var_name);
	return FVariables[var_name];
}
//---------------------------------------------------------------------------
bool TVariablesMap::AddVariable(TVariableDef* var, std::string& err_msg)
{
	std::string t_var_name = var->GetVarName();
	lowercase(t_var_name);
	// If this variable has already been declared, don't let the edit programmer
	// re-use it! (This happened with some very old AJCC edits, where TABLEVAR
	// names were reused as char[x] variables
	TVariableDef* t_test_var = FVariables[t_var_name];
	if (t_test_var)
	{
		char t_buffer[100];
		std::snprintf(t_buffer, 100,
			"variable \"%s\" already declared; variable names must be unique",
			t_var_name.c_str());
		err_msg.assign(t_buffer);
		return false;
	}
	else
	{
		FVariables[t_var_name] = var;
	}

	return true;
}
//---------------------------------------------------------------------------
/*************************
 *** TEditCompiler
 *************************/
TEditCompiler::TEditCompiler(TSQLDB* database, TEngineErrorsPump* pump,
	TTablesList* tables_list)
{
	FSQLDB = database;
	FQry = new TSQLQuery(FSQLDB);
	FEngineErrorsPump = pump;
	FTablesList = tables_list;
	FVariablesMap = new TVariablesMap();


	FAPICode = "";
	FCompiledCode = "";
	FCurrentVarName = "";
	FCurrentString = "";
	FCurrentNumber = 0;
	FNumTokens = sizeof(tokgenlist) / sizeof(EEt_TokType);
} // constructor
//---------------------------------------------------------------------------
TEditCompiler::~TEditCompiler()
{
	// When the user detaches, we want to be sure all resources are
	// freed. This syntax will delete the EEt_VarDef and EEt_UDFDef
	// objects in each element of their respective vectors, as well
	// as the std::vector element. (Thanks to Mark Storer, who posted this
	// implementation in comp.lang.c++.)
	std::for_each(FUDFDefList.begin(), FUDFDefList.end(), DeleteUDFDef);

	FUDFDefList.clear();

	delete FVariablesMap;

	clear_logic_messages_lists();

	if (FQry)
	{
		delete FQry;
		FQry = NULL;
	}
}
//---------------------------------------------------------------------------

#define iswhite(c) (c == ' '|| c == '\t'|| c == '\r'|| c == '\n')
#define isdelim(c) (strchr(" !;,+-<>'%^=()[]{}",c) || c==9 || c=='\r' || c==0)

int TEditCompiler::Compile(std::string edit_logic, int* how_many_errors,
	int* how_many_warnings)
{
	int t_idx,
		t_udf_count = 0;
	char t_LastTokenList[80];

	std::memset(t_LastTokenList, '\0', 80);

	FAPICode.assign(edit_logic);
	FAPIptr = 0;

	// Messages are cleared for EACH EDIT-- use GetErrorCount()
	// and GetErrorMessage() to retrieve them
	FErrorMessages.clear();
	// Ditto for warnings... use GetWarningCount() and GetWarningMessage()
	FWarningMessages.clear();

	// Clear any previously declared variables (when calling in a loop,
	// such as for generating an RMF, they need to be cleared between
	// each edit).
	FVariablesMap->ClearVariables();
	// Note that I am NOT going to clear FUDFDefList, because EW2
	// did not. I will, however, reset the is_used_in_edit field
	// to false until we determine the UDF is actually being used
	// in THIS edit. (Affects generating for RMF, not the one-at-a-
	// time syntax checker.)
	std::for_each(FUDFDefList.begin(), FUDFDefList.end(), ClearUsedFlag);

	// Still more! Clear the lists that collect the Field, Message and Table references
	FLogicFieldsList.clear();
	FLogicTablesList.clear();

	clear_logic_messages_lists();

	if (!FAPICode.empty())
	{
		FCompiledCode = "@";
		FCurrentVarName = "";
		FCurrentString = "";
		FCurrentNumber = 0;

		// Remove comments from code
		if (strip_comments() == crNormal)
		{

			findUDFs();

			t_udf_count = (int)FUDFDefList.size();

			/*   If there are UDFs, we need to preface the edit
			 *   with the count and their addresses. Unfortunately,
			 *   we won't know the addresses until we have finished
			 *   encoding the edit, so for now just create placeholders.
			 */
			if (t_udf_count > 0)
			{
				FCompiledCode[0] = '@' + t_udf_count;
				for (t_idx = 0; t_idx < t_udf_count; t_idx++)
				{
					FCompiledCode += "00000";
				}
			}

			// reset the pointer to the start of the code
			FAPIptr = 0;
			FCompLineNo = 0;
			FCurrentToken = 0;

			// This translates FAPICode to FCompiledCode
			int t_inlist_param = 0; // to find and (if needed) fix 2nd parameter to INLIST
			tokenize(t_inlist_param);
			while (FCurrentToken != FINISHED)
			{
				tokenize(t_inlist_param);
			}
			// Evaluate the block; skip over the UDF addresses, if any
			FCompiledPtr = ( t_udf_count > 0 ?
							 (t_udf_count*5)+1:
							 0);

			FCompLineNo = 0;
			if (balance_parens())
			{
				FCompiledPtr = ( t_udf_count > 0 ?
								 (t_udf_count*5)+1:
								 0);
				FCompLineNo = 0;
				parse_block(0);

				if (t_udf_count > 0)
				{
					char t_address[6];
					// Let's go back and get those addresses now
					for (t_idx = 0; t_idx < t_udf_count; t_idx++)
					{
						EEt_UDFDef* t_udf = FUDFDefList[t_idx];

						sprintf(t_address,"%05d",t_udf->func_addr);

						FCompiledCode.replace((t_idx*5)+1, 5, t_address);
					}
				}
			}
		} // if (strip_comments == crNormal)
	} // if (FAPICode.empty())

	*how_many_errors = (int)FErrorMessages.size();
	*how_many_warnings = (int)FWarningMessages.size();

	// Update the FLogicTablesList
	update_logic_tables();

	// Update the FLogicMessageNumbersList and FLogicErrorMessagesList
	update_logic_messages();

	/* return t_result; */
	return 0;

} // Compile
//---------------------------------------------------------------------------
std::string TEditCompiler::GetCompiledCode()
{
	return FCompiledCode;
}
//---------------------------------------------------------------------------
int TEditCompiler::GetErrorCount()
{
	return (int)FErrorMessages.size();
}
//---------------------------------------------------------------------------
char * TEditCompiler::GetErrorMessage( int index)
{
	char * t_error = NULL;
	if (index >= 0 && index < (int)FErrorMessages.size())
	{
		t_error = (char*)FErrorMessages[index].c_str();
	}
	return t_error;
}
//---------------------------------------------------------------------------

int TEditCompiler::GetWarningCount()
{
	return (int)FWarningMessages.size();
}
//---------------------------------------------------------------------------
char * TEditCompiler::GetWarningMessage( int index)
{
	char * t_warning = NULL;
	if (index >= 0 && index < (int)FWarningMessages.size())
	{
		t_warning = (char*)FWarningMessages[index].c_str();
	}
	return t_warning;
}
//---------------------------------------------------------------------------
int TEditCompiler::GetLogicListCount(char which_list)
{
	if (which_list == 'F')
	{
		return (int)FLogicFieldsList.size();
	}
	else if (which_list == 'E')
	{
    	return (int)FLogicErrorMessagesList.size();
	}
	else if (which_list == 'M')
	{
		return (int)FLogicMessageNumbersList.size();
	}
	else if (which_list == 'T')
	{
		return (int)FLogicTablesList.size();
	}
	else
	{
		FEngineErrorsPump->AddError("TEditCompiler", "GetLogicListCount",
			"Invalid value in parameter 'which_list'");
		return -1;
	}

} // GetLogicListCount
//---------------------------------------------------------------------------
int TEditCompiler::GetLogicFieldNumber(int idx)
{
	int t_result = 0;

	int t_max_size = (int)FLogicFieldsList.size();
	if ((idx >= 0) && (idx < t_max_size))
	{
		t_result = FLogicFieldsList[idx];
	}
	else
	{
		char t_buffer[100];
		std::snprintf(t_buffer, 100,
			"Parameter 'idx' (%d) out of bounds for list ", idx);
		std::strncat(t_buffer, "FLogicFieldsList", 100);

		FEngineErrorsPump->AddError("TEditCompiler", "GetLogicFieldNumber",
			t_buffer);
		throw std::runtime_error("Programmer error");
	}

	return t_result;
} // GetLogicFieldNumber
//---------------------------------------------------------------------------
std::string TEditCompiler::GetLogicTableName(int idx)
{
	std::string t_result;

	int t_max_size = (int)FLogicTablesList.size();
	if ((idx >= 0) && (idx < t_max_size))
	{
		t_result = FLogicTablesList[idx];
	}
	else
	{
		char t_buffer[100];
		std::snprintf(t_buffer, 100,
			"Parameter 'idx' (%d) out of bounds for list ", idx);

		std::strncat(t_buffer, "FLogicTablesList", 100);

		FEngineErrorsPump->AddError("TEditCompiler", "GetLogicTableName",
			t_buffer);
		throw std::runtime_error("Programmer error");
	}

	return t_result;
} // GetLogicTableName
//---------------------------------------------------------------------------
std::string TEditCompiler::GetLogicErrorMessage(int idx, int* msg_func_token)
{
	std::string t_result;

	int t_max_size = (int)FLogicErrorMessagesList.size();
	if ((idx >= 0) && (idx < t_max_size))
	{
		TLogicMessageObject* t_obj = FLogicErrorMessagesList[idx];
		*msg_func_token = t_obj->FuncToken;
		t_result = t_obj->MessageText;
	}
	else
	{
		char t_buffer[100];
		std::snprintf(t_buffer, 100,
			"Parameter 'idx' (%d) out of bounds for list ", idx);

		std::strncat(t_buffer, "FLogicErrorMessagesList", 100);

		FEngineErrorsPump->AddError("TEditCompiler", "GetLogicErrorMessage",
			t_buffer);
		throw std::runtime_error("Programmer error");
	}
	return t_result;

} // GetLogicErrorMessage
//---------------------------------------------------------------------------
int TEditCompiler::GetLogicMessageNumber(const int idx, int* msg_func_token)
{
	int t_message_number;

	int t_max_size = (int)FLogicMessageNumbersList.size();
	if ((idx >= 0) && (idx < t_max_size))
	{
		TLogicMessageObject* t_obj = FLogicMessageNumbersList[idx];
		*msg_func_token = t_obj->FuncToken;
		t_message_number = t_obj->MessageNumber;
	}
	else
	{
		char t_buffer[100];
		std::snprintf(t_buffer, 100,
			"Parameter 'idx' (%d) out of bounds for list ", idx);

		std::strncat(t_buffer, "FLogicMessageNumbersList", 100);

		FEngineErrorsPump->AddError("TEditCompiler", "GetLogicMessageNumber",
			t_buffer);
		throw std::runtime_error("Programmer error");
	}

	return t_message_number;

} // GetLogicMessageNumber
//---------------------------------------------------------------------------

void TEditCompiler::comp_sntx_err(const char *errmsg)
{
	char t_line_buf[5];
	std::string t_message ;
	std::sprintf(t_line_buf, "%4i", (int)(FCompLineNo+1));

	t_message = (std::string)"Line " + (std::string)t_line_buf + (std::string)": " + (std::string)errmsg;
	FErrorMessages.push_back(t_message);
}
//---------------------------------------------------------------------------

void TEditCompiler::comp_sntx_warn(const char *warnmsg)
{
	char t_line_buf[5];
	std::string t_message ;
	std::sprintf(t_line_buf, "%4i", (int)(FCompLineNo+1));

	t_message = (std::string)"Line " + (std::string)t_line_buf + (std::string)": " + (std::string)warnmsg;
	FWarningMessages.push_back(t_message);
}
//---------------------------------------------------------------------------
int TEditCompiler::strip_comments()
{
    std::string t_string;
    size_t t_APIlen = FAPICode.length();
    FAPIptr = 0;

    if (FAPICode.empty())
        return crEmptyAPICode;

	// skip over leading white space
    while (iswhite(FAPICode[FAPIptr]))
        FAPIptr++;

    while (FAPIptr < t_APIlen)
    {
		// skip over comments
        if (FAPIptr < t_APIlen -1 &&
            FAPICode.substr(FAPIptr,2) == "/*")
        {
            FAPIptr += 2;
            while (FAPIptr < t_APIlen-1 &&
                   FAPICode.substr(FAPIptr,2) != "*/")
            {
                // Preserve line breaks-- we need them for
                // reporting where errors occurred.
                if (FAPICode[FAPIptr] == '\n')
                {
                    t_string += FAPICode[FAPIptr];
                }
                FAPIptr++;
            }
			// if we found the end of the comments, skip past the
			// closing star-slash.
			if (FAPIptr < t_APIlen-1 &&
                FAPICode.substr(FAPIptr,2) == "*/")
            {
                FAPIptr += 2;
            }
            else
            {
                return crUnclosedComments;
            }

		} // end skip over comments

		//  Place everything that is NOT a comment into t_string
        while (FAPIptr < t_APIlen-1 && FAPICode.substr(FAPIptr,2) != "/*")
        {
            t_string += FAPICode[FAPIptr];
            FAPIptr++;
        }

		// Make sure we can bail out if comments close the edit
        if (FAPIptr >= t_APIlen -1)
        {
			if (FAPIptr == t_APIlen -1)
            {
                // Grab the last byte
                t_string += FAPICode[FAPIptr];
            }
            break;
        }

    } // while (FAPIptr < t_APIlen)

    FAPICode = t_string;

    return crNormal;
} // strip_comments
//---------------------------------------------------------------------------
void TEditCompiler::findUDFs()
{
	size_t t_idx,
		   t_marker = 0,
		   t_APIlen = FAPICode.length();

	// Shot myself in the foot by writing a UDF that embedded the words
	// "This function is never called" inside a UDF. The processor created
	// an instance of EEt_UDFDef named "is". What a mess. So I have to
	// look out for such shenanigans and ignore them.
	size_t t_func_marker,
		   t_quote_marker;

	std::string t_edits_logic = FAPICode;

	// lower-case the code so we can find what we seek
	std::transform(t_edits_logic.begin(), t_edits_logic.end(),
		t_edits_logic.begin(), tolower);

	// Start from empty with each edit that is being compiled
	std::for_each(FUDFDefList.begin(), FUDFDefList.end(), DeleteUDFDef);

	FUDFDefList.clear();


	while ( t_marker < t_APIlen)
	{
		// look for an instance of the word "function"
		t_func_marker = t_edits_logic.find("function", t_marker);
		// look for a quoted string
		t_quote_marker = t_edits_logic.find("\"", t_marker);

		if (t_quote_marker != std::string::npos)
		{
			if (t_func_marker != std::string::npos)
			{
				if (t_quote_marker < t_func_marker)
				{
					// A quoted string occurs prior to t_func_marker
					// Find out where the string ends, and make sure
					// t_func_marker is somewhere beyond it.
					t_quote_marker = t_edits_logic.find("\"", t_quote_marker+1);
					if (t_quote_marker != std::string::npos)
					{
						if (t_func_marker < t_quote_marker)
						{
							t_marker = t_func_marker+8;
							continue;
						}
					}
				}
			}
		}

		// Okay, this is legit... carry on.
		t_marker = t_func_marker;

		if (t_marker != std::string::npos)
        {
            t_marker += 9; // "function "
            /*
              The function address comes from the compiled
              code, so we have to evaluate this AFTER we
              translate from FAPICode to FCompiledCode
              t_func = "n" + ((char)('@' + t_idx)) + "(" , e.g. "nA("
              We'll use FCompiledCode.find(t_func, t_marker)
              wherever we find it, that's our address

			  in NAACCR metafile, see edit
              Primary Site, AJCC Stage Group, Ed 5, ICDO3(C/NET)
              for an example of an edit with UDFs, and note that
              the UDFs in the encoded version start with nF()
			*/

			EEt_UDFDef* t_udf = new EEt_UDFDef();
            if (t_udf)
            {
				try
				{
					std::memset(t_udf->func_name, '\0', 33);
                    t_udf->func_addr = 0;
                    t_udf->is_used_in_edit = true;

                    t_idx = 0;
                    while (t_marker < t_APIlen && t_idx < 33)
                    {
                        if (iswhite(t_edits_logic[t_marker]) ||
                            isdelim(t_edits_logic[t_marker]))
                        {
                            break;
                        }
                        t_udf->func_name[t_idx++] = t_edits_logic[t_marker++];
                    }
                    FUDFDefList.insert(FUDFDefList.end(), t_udf);
                } catch (...) {
                    delete t_udf;
                    throw;
                }
            } // if (t_udf)
        }
        else
            break;
    }
} // findUDFs
//---------------------------------------------------------------------------
int TEditCompiler::tokenize(int& inlist_param)
{
	//   tokenize() walks through the edits logic (FAPICode)
	//   and turns it into tokens based upon the contents of
	//   tokgenlist (see top of this file).
	int t_marker = 0,       // occasionally saving our place...
		t_tokidx,
		t_toklen,
		t_wordlen,
		t_result = crNormal;

	size_t t_APIlen = FAPICode.length(),
		   t_APIptr = FAPIptr; // makes sure we don't get stuck in an infinite loop

	char t_compiled;
    char t_udf_name[33];
    std::string t_word = "",
           t_field_num,
           t_construct_lf;

	bool t_isFunction;

	// eliminate blank spaces, tabs and carriage returns
    while (iswhite(FAPICode[FAPIptr]))
    {
        if (FAPICode[FAPIptr] == '\n')
        {
            FCompLineNo++;
            FCompiledCode += FAPICode[FAPIptr];
        }
		FAPIptr++;
    }

	//  Now look for matches in the tokgenlist. Jim Tebbel
	//  walked the array from the bottom up (precedence?),
	//  so we will, too.
	for (t_tokidx = FNumTokens-1; t_tokidx >= 0; t_tokidx--)
    {
        t_toklen = (int)std::strlen(tokgenlist[t_tokidx].orig);
        t_word = FAPICode.substr(FAPIptr, t_toklen);

        if (posix_stricmp(t_word.c_str(), tokgenlist[t_tokidx].orig) == 0)
        {
			//   Check to make sure we really have a match
			//   to whatever is in the tokgenlist[x].orig, e.g.,
			//   "return" and not "returnable"
			if (isalpha(t_word[0]) &&
                (isalpha(FAPICode[FAPIptr + t_toklen]) ||
                 FAPICode[FAPIptr + t_toklen] == '_'))
                continue;

			//   Let's watch out for a common syntax error that is not
			//   otherwise being caught (except in the Engine, where it
			//   barfs up an error from EEi_atom: Syntax Error)--
			//   "<>" does NOT EQUAL "!="
			if (FAPICode[FAPIptr] == '<' &&
				FAPICode[FAPIptr + t_toklen] == '>')
			{
				comp_sntx_err("Syntax error (\"<>\"); did you mean \"!=\"?");
				FAPIptr++;
				continue;
			}

			//  Here's another that has turned up with new authors:
			//  use of "=>" when they mean ">=" or "=<" for "<="
			if (FAPICode[FAPIptr] == '=' &&
				FAPICode[FAPIptr + t_toklen] == '>')
			{
				comp_sntx_err("Syntax error (\"=>\"); did you mean \">=\"?");
				FAPIptr++;
				continue;
			}

			//  Here's another that has turned up with new authors:
			//  use of "=>" when they mean ">=" or "=<" for "<="
			if (FAPICode[FAPIptr] == '=' &&
				FAPICode[FAPIptr + t_toklen] == '<')
			{
				comp_sntx_err("Syntax error (\"=<\"); did you mean \"<=\"?");
				FAPIptr++;
				continue;
			}


			//   Don't deal with APPFIELD, CHARCONST or NUMCONST
			//   within parentheses. (I don't understand why; just
			//   mimicking what has been in the processing since the
			//   1990s.)

            if (FAPICode[FAPIptr + t_toklen] == '(' &&
                (tokgenlist[t_tokidx].tok_type!=CONSTRUCT &&
                 tokgenlist[t_tokidx].tok_type!=OPERATOR &&
				 tokgenlist[t_tokidx].tok_type!=KEYWORD &&
                 tokgenlist[t_tokidx].tok_type!=DELIMITER &&
                 tokgenlist[t_tokidx].tok_name!=UDFUNCDECL))
                continue;

            FAPIptr += t_toklen;

            t_compiled = tokgenlist[t_tokidx].compiled;
            FCurrentToken = tokgenlist[t_tokidx].tok_name;
            FCurrentTokenType = tokgenlist[t_tokidx].tok_type;

            FCompiledCode += t_compiled;

            if (FCurrentTokenType != STRING)
            {
				// Again, skip the white space (but pick up all
				// instances of '\n'... we want them to help
				// the user interface give the user meaningful
				// feedback about errors).
				//
				// Note that we do NOT want to skip blanks and
				// other whitespace in quoted strings... they are
				// meant to be preserved as literals.
				while (FAPIptr < t_APIlen && iswhite(FAPICode[FAPIptr]))
                {
                    if (FAPICode[FAPIptr] == '\n')
                    {
                        FCompLineNo++;
                        if (FCurrentTokenType == CONSTRUCT)
                        {
                            // if, while, else, for... these
                            // will need to prefix their blocks
                            // with the size of the code. Don't
                            // put linefeeds in until after we
                            // have added the four blank spaces.
                            // (see below)
                            t_construct_lf += FAPICode[FAPIptr];
                        }
                        else
                        {
                            FCompiledCode += FAPICode[FAPIptr];
						}
                    }
                    FAPIptr++;
                }
            }

			if (FCurrentTokenType == CONSTRUCT)
            {
				// Tebbel's "jump table" goes here. This is the
				// distance in FCompiledCode that the pointer must
				// advance to skip past the execute block of a
				// CONSTRUCT. That block might look like this:
				// "\nDR;", which would require a jump of 4 encoded
				// bytes to get to the next executable statement (or
				// to the ELSE of an IF block).
				//
				// We set up the space for it here, and fill in a value
				// during the parse_block() portion of our program.
				FCompiledCode += "    ";
                // now add any linefeed characters accumulated above.
                FCompiledCode += t_construct_lf;
			}
            else if (FCurrentTokenType == FUNCTION &&
                     FCurrentToken == UDFUNCDECL)
            {
				std::memset(t_udf_name, '\0', 33);
                t_tokidx = 0;
                while (t_tokidx < 33 && FAPIptr < t_APIlen)
                {
                    if (iswhite(FAPICode[FAPIptr]) ||
                        isdelim(FAPICode[FAPIptr]))
                    {
                        break;
                    }
                    t_udf_name[t_tokidx++] = FAPICode[FAPIptr++];
                }
                // Find the declaration in FUDFDefList, attach
                // the token identifier, and fill in the
                // EEt_UDFDef.func_addr with the current position
                // in FCompiledCode
				for (t_tokidx = 0; t_tokidx < (int)FUDFDefList.size(); t_tokidx++)
				{
                    EEt_UDFDef* t_udf = FUDFDefList[t_tokidx];
                    if (posix_stricmp(t_udf_name, t_udf->func_name) == 0)
                    {
						FCompiledCode += 'A'+t_tokidx;

                        t_udf->func_addr = (int)FCompiledCode.size();
                        FUDFDefList[t_tokidx] = t_udf;
                        break;
                    }
                }
            }
            else
            {
                switch (t_compiled)
                {
                    case '\'':   // CCONST
                        FCompiledCode += FAPICode[FAPIptr];
                        FAPIptr++;

                        break;

                    case '"':    // STRING
						t_marker = (int)FAPICode.find_first_of("\"", FAPIptr);
                        if (t_marker >= 0)
                        {
                            t_marker -= (int)FAPIptr;

							t_word = FAPICode.substr(FAPIptr, t_marker);

							if (inlist_param == 1)
							{
								// We're processing the first parameter to INLIST.
								// If we come in here again, we're handling the
								// second parameter, which may need cleaning up
								inlist_param = 2;
							}
							else if (inlist_param == 2)
							{
								// This is a flag to indicate the function is INLIST
								// We are on the second parameter, which is the values
								// list. We need to remove extraneous spaces from
								// the string, because it messes up the processing
								// in TLibFuncs::call_inlist

								// Of course, this bollixes the list of options
								// if that list consists of blank spaces.
								std::string t_empty_str = rtrim(t_word);
								if (!t_empty_str.empty())
								{
									// strip down and rebuild the values string,
									// removing extaneous spaces, but without
									// killing the intent of making one of the
									// segments a string of blanks
									std::string t_scrap,
												t_values = t_word;
									// Carve up t_values into segments, breaking on the commas
									std::vector<std::string> t_segments;

									t_segments.reserve(20);

									size_t t_pos = t_values.find_first_of(",");
									while (t_pos != std::string::npos)
									{
										t_scrap = trim(t_values.substr(0, t_pos));
										if (!t_scrap.empty())
										{
											t_segments.push_back(t_scrap);
										}
										else
										{
											t_segments.push_back(t_values.substr(0, t_pos));
										}

										// Lop off the segment we just processed, and continue.
										t_values.assign(t_values.substr(t_pos+1, t_values.length()));
										t_pos = t_values.find_first_of(",");
									}
									// Place whatever is left as the last segment
									// (test to make sure the last character was not a comma)
									if (!t_values.empty())
									{
										t_scrap = trim(t_values);
										if (!t_scrap.empty())
										{
											t_segments.push_back(t_scrap);
										}
										else
										{
											t_segments.push_back(t_values);
										}
									}

									// Reassembly required
									size_t t_idx, t_max = t_segments.size();
									t_word = "";
									for (t_idx = 0; t_idx < t_max; t_idx++)
									{
										if (t_idx > 0)
											t_word += ",";
										t_word += t_segments[t_idx];
									}
								}
								inlist_param = 0;
							}

							t_wordlen = (int)t_word.length();

							// Although we are removing limits on strings in this new Engine,
							// the compiler has to store the length of string constants. We're
							// allowing only up to 999 bytes, which one would hope would be
							// adequate. (Remember: You can build variables of indefinite
							// length out of string constants up to 999 bytes.)
							if (t_wordlen > 999)
							{
								FCurrentToken = FINISHED;
								comp_sntx_err("string constant is too long ( > 999 characters)");
								return crError;
							}
							FCompiledCode += '0' + t_wordlen / 100;
							FCompiledCode += '0' + t_wordlen / 10;
							FCompiledCode += '0' + t_wordlen % 10;
							FCompiledCode += t_word;
							FAPIptr += t_marker+1;
						}
						else
						{
							comp_sntx_err("unterminated string (\" expected)");
							// find the end of this line
							while (FAPIptr < t_APIlen)
							{
								if (FAPICode[FAPIptr++] == '\n')
									break;
							}

						}

						break;

					case '#':    // FIELD (case insensitive)
						if (inlist_param == 1)
						{
							// We're in INLIST, and the first parameter is
							// a FIELD identifier. The next parameter will
							// MOST LIKELY be a string constant containing
							// the values list. This will need to be processed
							// to remove extraneous embedded spaces, so set
							// the flag here.
							inlist_param = 2;
						}
						// For EditEngine v5.x, we will no longer support
						// Local Name and Title in Edit Logic; these elements
						// have been removed from the EE_Fields table structure.
						//	if (FAPICode[FAPIptr] == 'S' ||
						//		FAPICode[FAPIptr] == 's' ||
						//		FAPICode[FAPIptr] == 'L' ||
						//		FAPICode[FAPIptr] == 'l' ||
						//		FAPICode[FAPIptr] == 'T' ||
						//		FAPICode[FAPIptr] == 't')
						if (FAPICode[FAPIptr] == 'S' ||
							FAPICode[FAPIptr] == 's')
						{
                            FAPIptr++;

                            while (iswhite(FAPICode[FAPIptr]))
                                FAPIptr++;

                            if (FAPICode[FAPIptr] == '"')
                            {
                                t_marker = (int)FAPICode.find_first_of("\"", FAPIptr+1);
                                if (t_marker >= 0)
                                {
                                    t_marker -= (int)FAPIptr;
                                    t_word = FAPICode.substr(FAPIptr, t_marker+1);

									// LOOK UP EE_Fields.FieldNumber
									t_field_num = field_name_to_field_num(t_word);
									FCompiledCode += t_field_num; // "0000";

									add_to_logic_fields(std::atoi(t_field_num.c_str()));

                                    FAPIptr += (t_marker+1);
                                }
                                else
                                {
									comp_sntx_err("unterminated string (\" expected)");
                                    // find the end of this line
                                    while (FAPIptr < t_APIlen)
                                    {
                                        if (FAPICode[FAPIptr++] == '\n')
                                            break;
                                    }

                                }

                                break;
                            }
						}
                        else
                        {
							comp_sntx_err("field identifiers must use form like: #S\"Primary Site\"");
                        }

						if (isdigit(FAPICode[FAPIptr]))
                        {
                            t_word = "";
                            while (isdigit(FAPICode[FAPIptr]))
                            {
                                t_word += FAPICode[FAPIptr];
                                FAPIptr++;
                            }
                            t_wordlen = (int)t_word.length();
                            FCompiledCode += t_word;
                        }
                        break;

                } // switch (t_compiled)
            }

			// we found the match; now break out of this for.. loop
            break;
		} // if (t_word.compare(tokgenlist[t_tokidx].orig) == 0)
	} // for (t_tokidx = FNumTokens-1; t_tokidx >= 0; t_tokidx--)

    if (t_tokidx < 0) // remember, we were counting DOWN from FNumTokens-1
    {
		//  We didn't find a match in tokgenlist... let's see what we
		//  have here.
		if (isalpha(FAPICode[FAPIptr]))  // function or variable
        {
			t_isFunction = false;   // until proven otherwise

            t_word = "";
            while (!isdelim(FAPICode[FAPIptr]))
            {
                t_word += tolower(FAPICode[FAPIptr]);
                FAPIptr++;
            }

            t_wordlen = (int)t_word.length();

			char t_intern_func_code = get_intern_func_code(t_word);
			if (t_intern_func_code != ' ')
			{
				FCompiledCode += 'M';
				FCompiledCode += t_intern_func_code;

				if (t_intern_func_code == 'U') // INLIST
				{
					inlist_param = 1; // next up: the first parameter to INLIST
				}

				t_isFunction = true;
			}

            if (!t_isFunction)
            {
				// Didn't find it among the LIBFUNCs; look for it
				// among the UDFs.
				for (t_tokidx = 0; t_tokidx < (int)FUDFDefList.size(); t_tokidx++)
                {
                    EEt_UDFDef* t_udf = FUDFDefList[t_tokidx];
                    if (t_word.compare(t_udf->func_name) == 0)
                    {
                        FCompiledCode += 'm';
						FCompiledCode += 'A'+t_tokidx;
                        t_isFunction = true;
                        break;
                    }
                }

				if (!t_isFunction)
				{
					// This one stumped me awhile: the APICode contained
					// dx_year = DATE_YEAR (#S"Date of Diagnosis")... that was
					// a leading blank before the open parenthesis. Eat that,
					// and THEN check to see if we're calling a deprecated
					// function
					while (FAPICode[FAPIptr] == ' ')
						FAPIptr++;

					if (FAPICode[FAPIptr] == '(')
					{
						// Okay, this is bad. This unidentified object
						// isn't one of our internal functions, and it
						// isn't one of the UDFs. Complain!
						std::string t_err =
							"call to undefined function \"" + t_word + "\" (neither edits-language nor user-defined)";
						comp_sntx_err((char*)t_err.c_str());

						// but carry on as if it's a variable, just so we
						// can finish compiling the thing
					}
				}
            }
		} // looking for functions

		if (!t_isFunction)  // not a function, so it must be a variable
        {
			FCompiledCode += 'N'; // tokgenlist for VAR
			FCompiledCode += '0' + t_wordlen / 100;
            FCompiledCode += '0' + t_wordlen / 10;
			FCompiledCode += '0' + t_wordlen % 10;
			FCompiledCode += t_word;

			// Special handling for INLIST, where we may just have
			// processed the first parameter, a variable. Notify the
			// above code that we're ready to evaluate the second parameter,
			// the values list.
			if (inlist_param == 1)
			{
				inlist_param = 2;
			}
			else if (inlist_param == 2)
			{
				std::string t_err =
					"the second parameter to INLIST must be a string constant";
				comp_sntx_err((char*)t_err.c_str());

				inlist_param = 0;
			}
		}
    } // if (t_tokidx < 0)

    if (isdigit(FAPICode[FAPIptr]))
    {
        t_word = "";
        while (isdigit(FAPICode[FAPIptr]))
        {
            t_word += FAPICode[FAPIptr];
            FAPIptr++;
        }

        t_wordlen = (int)t_word.length();

        FCompiledCode += 'U'; // tokgenlist for NUMBER
        FCompiledCode += '0' + t_wordlen / 100;
		FCompiledCode += '0' + t_wordlen / 10;
		FCompiledCode += '0' + t_wordlen % 10;
        FCompiledCode += t_word;
    }

    if (FAPIptr >= t_APIlen)
    {
        FCurrentToken = FINISHED;
    }
	else if (FAPIptr == t_APIptr)
	{
		std::string t_msg = "Illegal character (";
		t_msg += FAPICode.substr(FAPIptr,1);
		t_msg += ")";

		comp_sntx_err(t_msg.c_str());

		FAPIptr++;
	}

	return t_result;

} // tokenize()
//---------------------------------------------------------------------------
void TEditCompiler::clear_logic_messages_lists()
{
	for (size_t t_idx = 0; t_idx < FLogicMessageNumbersList.size(); t_idx++)
	{
    	delete (TLogicMessageObject*)FLogicMessageNumbersList[t_idx];
	}
	FLogicMessageNumbersList.clear();

	for (size_t t_idx = 0; t_idx < FLogicErrorMessagesList.size(); t_idx++)
	{
		delete (TLogicMessageObject*)FLogicErrorMessagesList[t_idx];
	}

	FLogicErrorMessagesList.clear();
}
//---------------------------------------------------------------------------
void TEditCompiler::add_to_logic_fields(int item)
{
	bool t_add_item = true;

	// These items may be mentioned more than once in edit logic, so we
	// want to be sure to add them to our lists only once.
	t_add_item = true;
	for (size_t t_idx = 0; t_idx < FLogicFieldsList.size(); t_idx++)
	{
		if (FLogicFieldsList[t_idx] == item)
		{
			t_add_item = false;
			break;
		}
	}

	if (t_add_item)
	{
		FLogicFieldsList.push_back(item);
	}

} // add_to_logic_fields
//---------------------------------------------------------------------------
void TEditCompiler::update_logic_tables()
{
	// We're looking for table names, which will be referenced from
	// within function calls, like this:
	//
	//  MX(  LOOKUP,         e.g., MX(#14,"014STATE.DBF.CODE)
	//  MS(  ILOOKUP,        e.g., MS(#14,"014STATE.DBF.CODE)
	//  Mf(  RLOOKUP,        e.g., Mf(N08sitehist,"023IF38IMPO.DBF.SITEHISTLO)
	//  M{(  SQLLookup,      e.g., M{("006BPLACE,"004CODE,N003t_code)
	//  M}(  SQLRangeLookup  e.g., M}("008STATEZIP,"012STATE,ZIPLOW,N10search_str)
	//
	//  Note: We are ignoring MB( BINLOOKUP, because we're not supporting
	//  binary tables in EDITS 5.0
	std::string t_parse;
	std::string t_func_token;
	std::string t_table_name;

	size_t t_pos1, t_pos2, t_idx;

	for (int t_iteration = 0; t_iteration < 5; t_iteration++)
	{
		t_parse.assign(FCompiledCode);
		if (t_iteration == 0)
			t_func_token.assign("MX(");
		else if (t_iteration == 1)
			t_func_token.assign("MS(");
		else if (t_iteration == 2)
			t_func_token.assign("Mf(");
		else if (t_iteration == 3)
			t_func_token.assign("M{(");
		else if (t_iteration == 4)
			t_func_token.assign("M}(");

		t_pos1 = t_parse.find(t_func_token);
		while (t_pos1 != std::string::npos)
		{
			t_table_name.clear();
			// Lop off the first part
			t_parse.assign(t_parse.substr(t_pos1+3, t_parse.length()));

			// The old-style lookups put the table name as the second parameter
			// The new-style (SQL) lookups put the table name first
			if (t_iteration >= 3)
			{
				// Find the comma AFTER the table name
				int t_str_len = std::atoi(t_parse.substr(1,3).c_str());

				if (t_str_len > 0)
				{
					t_table_name.assign(t_parse.substr(4, t_str_len));
				}
			}
			else
			{
				// Look for the comma between first and second parameters
				t_pos2 = t_parse.find(",\"");
				// It's a syntax error for this to fail, but we'll carry on here
				// as if we don't know that.
				if (t_pos2 != std::string::npos)
				{
					t_parse.assign(t_parse.substr(t_pos2+5, t_parse.length()));
					// We're now positioned on the first character of the name
					// of the lookup table.
					// Walk the string looking for either the period separator
					// or the function-closing parenthesis
					for (t_idx = 0; t_idx < t_parse.length(); t_idx++)
					{
						if ((t_parse[t_idx] == '.') ||
							(t_parse[t_idx] == ')'))
						{
							t_table_name = t_parse.substr(0, t_idx);
							break;
						}
					}
				}
			}

			if (!t_table_name.empty())
			{
				bool t_add_it = true;
				for (t_idx = 0; t_idx < FLogicTablesList.size(); t_idx++)
				{
					if (FLogicTablesList[t_idx].compare(t_table_name) == 0)
					{
						t_add_it = false;
						break;
					}
				}
				if (t_add_it)
				{
					FLogicTablesList.push_back(t_table_name);
				}
			}
			// and look for another instance
			t_pos1 = t_parse.find(t_func_token);
		}
	}
} // update_logic_tables
//---------------------------------------------------------------------------
void TEditCompiler::update_logic_messages()
{
	// Extract list of user tables and error messages from the
	// edit logic. Believe it or not, the most reliable way to
	// find this is by looking through the COMPILED edit for
	// specific codes:
	//
	//    Mx(  USR2,              e.g., Mx(U03505,  // WARNING: Removed in 2015
	//    ML(  ERROR_MSG,         e.g., ML(U047210)
	//    MM(  ERROR_TEXT,        e.g., MM("13Opening IF662)
	//    Mg(  SAVE_ERROR_TEXT,   e.g., Mg(U047239)
	//    Mh(  SAVE_TEXT,         e.g., Mh("58ED8300: If DX = 1994-2003, EOD--Tumor Size cannot be blank
	//    M(  SAVE_WARNING_TEXT, e.g., M(U03261)  (Alt+0133)
	//    M(  SET_WARNING,       e.g., M(U041133) (Alt+0132)
	//    Mi(  SET_ERROR,         e.g., Mi(U03305)  // WARNING: Deprecated in EDITS40, removed in 2015
	//
	// Note: This list shows the delimiters for determining the start/end
	// of each function call.
	// These, combined with the EE_CEDIT.ERR_MSG_CD determine the error messages
	// and warnings used by the edit.

	std::string t_parse;
	std::string t_func_code;
	std::string t_err_msg;
	int t_msg_number,
		t_str_len;

	// The t_func_token is returned through EDITS API functions to the caller
	// to announce which EDITS Language function invoked the message
	// Do not be confused by the apparent redundancy of codes 'E' and 'S';
	// the text functions go to a different list, and are retrieved through
	// different API calls.
	// Do not yell at me for this system; I'm carrying forward a scheme
	// that was started with the early versions of the EditEngine and
	// EditWriter. It actually works very nicely.
	int t_func_token;

	// To prevent us from adding the same message repeatedly
	TLogicMessageObject* t_found_obj;

	size_t t_pos1;

	for (int t_iteration = 1; t_iteration < 7; t_iteration++)
	{
		// Place the compiled code into our parser variable
		t_parse.assign(FCompiledCode);

		// Figure out what we're looking for...
		if (t_iteration == 1)
		{
			t_func_code.assign("ML(");     // ERROR_MSG
			t_func_token = 69;             // 'E'
		}
		else if (t_iteration == 2)
		{
			t_func_code.assign("MM(");     // ERROR_TEXT - text strings
			t_func_token = 69;             // 'E'
		}
		else if (t_iteration == 3)
		{
			t_func_code.assign("Mg(");     // SAVE_ERROR_TEXT
			t_func_token = 83;             // 'S'
		}
		else if (t_iteration == 4)
		{
			t_func_code.assign("Mh(");     // SAVE_TEXT - text strings
			t_func_token = 83;             // 'S'
		}
		else if (t_iteration == 5)
		{
			t_func_code.assign("M(");     // SAVE_WARNING_TEXT
			t_func_token = 87;             // 'W'
		}
		else if (t_iteration == 6)
		{
			t_func_code.assign("M(");     // SET_WARNING
			t_func_token = 65;             // 'A'
		}
		t_pos1 = t_parse.find(t_func_code);
		while (t_pos1 != std::string::npos)
		{
			// Lop off the t_func_code
			t_parse.assign(t_parse.substr(t_pos1+3, t_parse.length()));

			if ((t_iteration == 2) || (t_iteration == 4))
			{
				// string messages: ERROR_TEXT or SAVE_TEXT

				// We expect to see something like this:
				// "013Opening IF662
				// where the quotation mark means a string, and the next
				// three digits tell us how long the string is.
				t_str_len = std::atoi(t_parse.substr(1,3).c_str());
				t_err_msg.assign(t_parse.substr(4, t_str_len));

				// Make sure the compiled logic does not contain multiple
				// instances of this message... we need to list it only
				// once.
				t_found_obj = NULL;
				for (size_t t_idx = 0; t_idx < FLogicErrorMessagesList.size(); t_idx++)
				{
					if ((FLogicErrorMessagesList[t_idx]->FuncToken == t_func_token) &&
						(FLogicErrorMessagesList[t_idx]->MessageText.compare(t_err_msg) == 0))
					{
						t_found_obj = FLogicErrorMessagesList[t_idx];
						break;
					}
				}
				// Add it to our list
				if (!t_found_obj)
				{
					TLogicMessageObject* t_obj = new TLogicMessageObject();
					t_obj->FuncToken = t_func_token;
					t_obj->MessageNumber = 0;
					t_obj->MessageText.assign(t_err_msg);

					FLogicErrorMessagesList.push_back(t_obj);
				 }
			} // if ((t_iteration == 2) || (t_iteration == 4))
			else
			{
				// numbered messages

				// We expect to see something like this:
				// ML(U0047210) => U0047210
				// where the 'U' means a number, and the next
				// three digits tell us how long the string representation
				// of the number is (that is, it could be 001 through 999
				// digits long)
				// HOWEVER, some of these functions permit more than one parameter;
				// e.g., SAVE_ERROR_TEXT, SET_WARNING, SAVE_WARNING_TEXT.
				// For those, we need to stay in this processing loop until
				// we encounter the closing parenthesis for the parameter list.
				//size_t t_pos_u = t_parse.find("U",1);
				size_t t_pos_u = 0;
				size_t t_pos_rparen = t_parse.find(")",0);

				do
				{
					t_str_len = std::atoi(t_parse.substr(1,3).c_str());
					// Yow. Sometimes a capital 'U' is part of an identifier
					// for a function (i.e., it will be partnered with 'M',
					// and look something like this: "MU(#102". That is NOT
					// a message, so just move on.
					if (t_str_len > 0)
					{
						t_msg_number = std::atoi(t_parse.substr(4, t_str_len).c_str());

						// Make sure the compiled logic does not contain multiple
						// instances of this function + message number... we need to
						// list it only once.
						t_found_obj = NULL;
						for (size_t t_idx = 0; t_idx < FLogicMessageNumbersList.size(); t_idx++)
						{
							if ((FLogicMessageNumbersList[t_idx]->FuncToken == t_func_token) &&
								(FLogicMessageNumbersList[t_idx]->MessageNumber == t_msg_number))
							{
								t_found_obj = FLogicMessageNumbersList[t_idx];
								break;
							}
						}
						// Add it to our list
						if (!t_found_obj)
						{
							TLogicMessageObject* t_obj = new TLogicMessageObject();
							t_obj->FuncToken = t_func_token;
							t_obj->MessageNumber = t_msg_number;
							t_obj->MessageText.clear();

							FLogicMessageNumbersList.push_back(t_obj);
						 }
					 } // t_str_len > 0
					 t_pos_u = t_parse.find("U",1);
					 if ((t_pos_u != std::string::npos) &&
						 (t_pos_u < t_pos_rparen))
						t_parse = t_parse.substr(t_pos_u, t_parse.length());

				} while ((t_pos_u != std::string::npos) &&
						 (t_pos_u < t_pos_rparen));
			}

			// Now look for another instance of our function code
			t_pos1 = t_parse.find(t_func_code);

		} // while (t_pos1 != std::string::npos)
	} // for (int t_iteration = 0; t_iteration < 8; t_iteration++)

} // update_logic_messages
//---------------------------------------------------------------------------
std::string TEditCompiler::field_name_to_field_num(std::string fname)
{
	std::string t_field_num;
	// fname is going to arrive here decorated with opening and closing
	// quotes: "Abstracted By". Get rid of the quotation marks (but don't
	// assume that someday somebody won't want to name a field
	// [My name is "Earl"]
	std::string t_fname = fname;
	if (t_fname.substr(0,1) == "\"")
		t_fname = t_fname.substr(1, t_fname.length());
	if (t_fname.substr(t_fname.length()-1,1) == "\"")
		t_fname = t_fname.substr(0, t_fname.length()-1);
	// SQLize t_fname
	t_fname = stringReplace(t_fname, "'", "''");
	std::string t_sql =
		"SELECT FieldNumber FROM EE_Fields WHERE FieldName = '";
	t_sql += t_fname + "';";

	FQry->SetIsActive(false);
	// don't check whether parameterized... C/NET Solutions has a field
	// named "PATH REPORT REQUIRED?"... yep, with a question mark. That would
	// cause this to fail
	FQry->SetSQLText(t_sql, false);
	try
	{
		FQry->SetIsActive(true);
		if (FQry->RowCount() == 0)
		{
			std::string t_err = "field name " + fname + " not in data dictionary";
			comp_sntx_err((char*)t_err.c_str());
			t_field_num.assign( "0000" );
		}
		else
		{
			t_field_num.assign(FQry->GetFieldByName("FieldNumber"));
		}

		FQry->SetIsActive(false);
	}
	catch (...)
	{
		FQry->SetIsActive(false);
		std::string t_err = "failure to open query";
		FEngineErrorsPump->AddError("TEditCompiler", "field_name_to_field_num", t_err);
		t_err = "SQL: " + t_sql;
		FEngineErrorsPump->AddError("TEditCompiler", "field_name_to_field_num", t_err);
		throw;
	}
	return t_field_num;
} // field_name_to_field_num
//---------------------------------------------------------------------------
bool TEditCompiler::balance_parens()
{
    bool t_result = true; // assume the best

	std::stack<int> open_parens;
    std::stack<int> open_braces;

    get_next_token();

    while (FCurrentToken != FINISHED)
    {
        if (FCurrentTokenType == DELIMITER)
        {
            switch (FCurrentToken)
            {
                case LPAREN:
                    open_parens.push((int)FCompLineNo);
                    break;
                case RPAREN:
                    // Are we facing excessive closing parens?
                    if (open_parens.empty())
                    {
                        t_result = false;
						comp_sntx_err("closing parenthesis with no matching opening parenthesis");
                    }
                    else
                    {
                        open_parens.pop();
                    }

                    break;
            }
        }
        else if (FCurrentTokenType == BLOCK)
        {
            switch(FCurrentToken)
            {
                case LBRACE:
                    open_braces.push((int)FCompLineNo);
                    break;
                case RBRACE:
                    // Too many closing braces
                    if (open_braces.empty())
                    {
                        t_result = false;
						comp_sntx_err("closing brace with no matching opening brace");
                    }
                    else
                    {
                        open_braces.pop();
                    }
                    break;
            }

        }
        get_next_token();
    }

    size_t t_save_line;
    std::string t_msg;

    if (!open_parens.empty())
    {
        t_result = false;
        while (!open_parens.empty())
        {
            t_save_line = FCompLineNo;
            FCompLineNo = open_parens.top();
			comp_sntx_err("unbalanced parentheses; please review your code");
            FCompLineNo = t_save_line;
            open_parens.pop();
        }
    }
    if (!open_braces.empty())
    {
        t_result = false;
        while (!open_braces.empty())
        {
            t_save_line = FCompLineNo;
            FCompLineNo = open_braces.top();
            t_msg = "unresolved open brace (hint: use Ctrl+B to find matched braces... see Edit Logic menu)";
            comp_sntx_err((char*)t_msg.c_str());
            FCompLineNo = t_save_line;
            open_braces.pop();
        }
    }
    return t_result;
}  // balance_parens
//---------------------------------------------------------------------------
int TEditCompiler::get_next_token()
{
    char t_current_token,
         t_tok_idx;

    size_t t_ProgressPtr;


	// skip over blank lines
    while (FCompiledCode[FCompiledPtr] == '\n')
    {
        FCompLineNo++;
        FCompiledPtr++;
    }

	// NOTE: this function uses toklist, not tokgenlist

	t_current_token = FCompiledCode[FCompiledPtr];

    if (t_current_token == '@')
    {
        FCompiledPtr++;
		// skip over blank lines here, too
        while (FCompiledCode[FCompiledPtr] == '\n')
        {
            FCompLineNo++;
            FCompiledPtr++;
        }

        t_current_token = FCompiledCode[FCompiledPtr];
    }

    t_ProgressPtr = FCompiledPtr;

	if (t_current_token >= '!' && t_current_token <= '#')
    {
        t_tok_idx = t_current_token - '!';
        FCurrentToken     = toklist[t_tok_idx].tok_name;
        FCurrentTokenType = toklist[t_tok_idx].tok_type;

        if (t_current_token == FIELD)
        {
            FCompiledPtr++;
            if (FCompiledCode[FCompiledPtr] == '-')
            {
                FCompiledPtr++;
            }

			if (isdigit(FCompiledCode[FCompiledPtr])) // number
            {
                while (isdigit(FCompiledCode[FCompiledPtr]))
                {
                    FCompiledPtr++;
                }
            }
        }

        // FixEdits # 5092: Putting a string longer than 79 characters
        // into a call to (for example) SAVE_TEXT causes the tokenize()
        // processing to abort leaving a construct like "@Mh(\"" (no
        // terminating quote or parenthesis. That led to the t_len calculation
        // yielding a less-than-zero value, which causes the substr() call
        // to barf up exception EEFFACE (Borland VCL-speak for "hey you stupid
        // programmer, you have an uncaught exception here!".
        else if ((t_current_token == SSTRING) &&
				 ((FCompiledPtr + 1) < FCompiledCode.length()) )
        {
			/*
			unsigned int t_len = (FCompiledCode[FCompiledPtr +1] - '0') * 10;
			t_len += (FCompiledCode[FCompiledPtr +2] - '0');

			// parse_assign() needs to know what variable we're referencing
			FCurrentString = FCompiledCode.substr(FCompiledPtr+3,t_len);
			FCompiledPtr += t_len + 3;
			*/
			unsigned int t_len = (FCompiledCode[FCompiledPtr +1] - '0') * 100;
			t_len += ((FCompiledCode[FCompiledPtr +2] - '0') * 10) - t_len;
			t_len += (FCompiledCode[FCompiledPtr +3] - '0');

			// parse_assign() needs to know what variable we're referencing
			FCurrentString = FCompiledCode.substr(FCompiledPtr+4,t_len);
			FCompiledPtr += t_len + 4;
		}
		else
		{
			FCompiledPtr++;
		}

    }

    else if (t_current_token >= MOD && t_current_token <= DIV)
    {
        // Watch out for use of "++" and "--", as they are not legal
        // in the edits language. Oh yeah, watch out for "+=" and "-=", too.
        if (t_current_token == ADD || t_current_token == SUB)
        {
			if (FCompiledPtr + 1 < FCompiledCode.length())
            {
                if (t_current_token == ADD)
                {
                    if (FCompiledCode[FCompiledPtr+1] == '+')
                    {
						comp_sntx_err("syntax error \"++\"");
                    }
                    else if (FCompiledCode[FCompiledPtr+1] == '=')
                    {
						comp_sntx_err("syntax error \"+=\"");
                    }
                }
                else if (t_current_token == SUB)
                {
                    if (FCompiledCode[FCompiledPtr+1] == '-')
                    {
						comp_sntx_err("syntax error \"--\"");
                    }
                    else if (FCompiledCode[FCompiledPtr+1] == '=')
                    {
						comp_sntx_err("syntax error \"-=\"");
                    }
                } // if (t_current_token == ADD or SUB)
            } // if (FCompiledPtr + 1 < (int)FCompiledCode.length())
        }

        t_current_token   = t_current_token-NOT-1;
        FCurrentToken     = toklist[t_current_token].tok_name;
        FCurrentTokenType = toklist[t_current_token].tok_type;

        if (t_current_token == CCONST)
        {
            FCompiledPtr++;
        }
        FCompiledPtr++;
    }

    else if (t_current_token >= COLON && t_current_token <= GT)
    {
        t_current_token   = t_current_token-NOT-11;
        FCurrentToken     = toklist[t_current_token].tok_name;
        FCurrentTokenType = toklist[t_current_token].tok_type;

        FCompiledPtr++;
    }

    else if (t_current_token >= ELSE && t_current_token <= DO)
    {
        t_current_token   = t_current_token-NOT-13;
        FCurrentToken     = toklist[t_current_token].tok_name;
        FCurrentTokenType = toklist[t_current_token].tok_type;

        if (FCurrentTokenType == CONSTRUCT)
        {
            // CONSTRUCT means if..else..while..for
            // These blocks will store the SIZE of the
            // block immediately preceding the code
            // in the block... so the engine can jump
            // over it if the condition is not met to
            // execute it.
            FCompiledPtr += 5;
        }
        else
            FCompiledPtr++;
    }

    else if (t_current_token == LIBFUNC)
    {
        t_current_token   = t_current_token-NOT-13;
        FCurrentToken     = toklist[t_current_token].tok_name;
        FCurrentTokenType = toklist[t_current_token].tok_type;

        FCompiledPtr += 2;
    }

    else if (t_current_token == UDFUNCCALL ||
             t_current_token == UDFUNCDECL)
    {
        FCurrentToken = t_current_token;
        FCurrentTokenType = FUNCTION;

        FCompiledPtr += 2;
    }

    else if (t_current_token == VAR)
    {
        t_current_token   = t_current_token-NOT-13;
        FCurrentToken     = toklist[t_current_token].tok_name;
        FCurrentTokenType = toklist[t_current_token].tok_type;

        unsigned int t_len;
        // parse_assign() needs to know what variable we're referencing
		FCurrentVarName = extract_var_name(t_len);

        FCompiledPtr += t_len + 4;
    }

    else if (t_current_token >= 'O' && t_current_token <= 'T')
    {
        t_current_token   = t_current_token-NOT-13;
        FCurrentToken     = toklist[t_current_token].tok_name;
        FCurrentTokenType = toklist[t_current_token].tok_type;

        FCompiledPtr++;
    }

    else if (t_current_token == 'U')
    {
		int t_len = (FCompiledCode[FCompiledPtr +1] - '0') * 100;
		t_len += (FCompiledCode[FCompiledPtr +2] - '0') * 10;
		t_len += (FCompiledCode[FCompiledPtr +3] - '0');

        // FCurrentNumber used in parse_vardecl()
        // and in parse_assign
        FCurrentNumber =
			std::atoi((char*)FCompiledCode.substr(FCompiledPtr+4,t_len).c_str());

        FCompiledPtr += t_len + 4;

        FCurrentTokenType = NUMBER;
    }

    else if (t_current_token >= LBRACE && t_current_token <= RBRACE)
    {
        t_current_token = t_current_token-NOT-46;
        FCurrentToken     = toklist[t_current_token].tok_name;
        FCurrentTokenType = toklist[t_current_token].tok_type;

        FCompiledPtr++;
    }

    else if (t_current_token == LBRACKET ||
             t_current_token == RBRACKET)
    {
        FCurrentToken = t_current_token;
        FCurrentTokenType = DELIMITER;

        FCompiledPtr++;
    }

	else if (FCompiledPtr >= FCompiledCode.length())  // end of file
    {
        FCurrentToken = FINISHED;
        FCurrentTokenType = DELIMITER;
    }

	else if (isdigit(FCompiledCode[FCompiledPtr]))   // number
    {
        while (isdigit(FCompiledCode[FCompiledPtr]))
        {
            FCompiledPtr++;
        }
        FCurrentTokenType = NUMBER;
    }

	else if (isalpha(FCompiledCode[FCompiledPtr]))  // var name
    {
        while (!isdelim(FCompiledCode[FCompiledPtr]))
        {
            FCompiledPtr++;
        }
        FCurrentTokenType = IDENTIFIER;
    }

    else if (FCompiledPtr == t_ProgressPtr)
    {
		std::string t_msg = "Illegal character (";
		t_msg += FAPICode.substr(FAPIptr,1);
		t_msg += ")";

		comp_sntx_err(t_msg.c_str());

        FCompiledPtr++;
    }

    return FCurrentTokenType;
} // get_next_token()
//---------------------------------------------------------------------------
std::string TEditCompiler::extract_var_name(unsigned int& var_length)
{
    std::string t_var_name;
	int t_var_len;

	t_var_len  = (FCompiledCode[FCompiledPtr +1] - '0') * 100;
	t_var_len += (FCompiledCode[FCompiledPtr +2] - '0') * 10;
	t_var_len += (FCompiledCode[FCompiledPtr +3] - '0');

	if ((FCompiledPtr + t_var_len + 4) < FCompiledCode.length())
    {
		t_var_name = FCompiledCode.substr(FCompiledPtr+4,t_var_len);
        var_length = t_var_len;
    }
    else
    {
        var_length = 0;
    }

    return t_var_name;
} // extract_var_name
//---------------------------------------------------------------------------
void TEditCompiler::parse_block( int required )
{
	 //   required:
	 //      0 means parse statements until end of block
	 //      1 means require open/closing braces and parse
	 //        statements between
	 //      2 means parse block if found, else parse one statement only
	 int t_open_braces = 0;
	 size_t t_compiled_length = FCompiledCode.length();

	 // FixEdits 5268
	 int t_IF_end_block = -1;
	 int t_dummy;

	 std::vector<int> t_nested_IF_end_blocks;
	 int t_idx;

	 get_next_token();

	 if (FCurrentTokenType == BLOCK && FCurrentToken == LBRACE)
	 {
		t_open_braces = 1;
		get_next_token();
	 }
	 else if (required == 1)
	 {
		comp_sntx_err("{ expected");
	 }

	 while (FCurrentToken != FINISHED)
	 {
		if (FCurrentTokenType == KEYWORD)
		{
			switch (FCurrentToken)
			{
				case EEi_INT:
				case EEi_CHAR:
					parse_vardecl(FCurrentToken);
					break;

				case RETURN:
					parse_return();
					break;
			}
		}

		else if (FCurrentTokenType == CONSTRUCT)
		{
			switch (FCurrentToken)
			{
				case WHILE:
					parse_while();
					break;

				case FOR:
					parse_for();
					break;

				case IF:
					parse_if(t_dummy);
					if (t_dummy < t_IF_end_block)
					{
						// We're in a nested IF/ELSE situation
						// Push the larger number onto the vector,
						// and reassign t_IF_end_block for ELSE
						// processing
						t_nested_IF_end_blocks.push_back(t_IF_end_block);
					}
					t_IF_end_block = t_dummy;
					break;

				case ELSE:
					if (t_IF_end_block < (int)FCompiledPtr)
					{
						t_dummy = t_IF_end_block;

						for (t_idx = (int)t_nested_IF_end_blocks.size()-1;
							 t_idx >= 0; t_idx--)
						{
							// See if we're just done with an inner nest.
							// Look to see where the outer loop ended.
							t_dummy = t_nested_IF_end_blocks[t_idx];

							t_nested_IF_end_blocks.pop_back();

							if (t_dummy == (int)FCompiledPtr)
								break;

						}
						if (t_dummy < (int)FCompiledPtr)
						{
							// Okay, NOW complain.
							comp_sntx_err("ELSE without an IF (or preceding IF needs braces '{' and '}')");
						}
					}
					parse_else(t_dummy);
					t_IF_end_block = t_dummy;
					break;
			}
		}

		else if (FCurrentTokenType == IDENTIFIER &&
				 FCurrentToken == VAR)
		{
			get_next_token(); // advance to the ASSIGN
			parse_assign();
		}

		else if (FCurrentTokenType == FUNCTION &&
				 FCurrentToken == LIBFUNC)
		{
			parse_fncall();
			get_next_token();
			if (! (FCurrentTokenType == DELIMITER &&
				   FCurrentToken == SCOLON))
			{
				comp_sntx_err("; expected");
			}
		}

		else if (FCurrentTokenType == FUNCTION &&
				 FCurrentToken == UDFUNCCALL)
		{
			parse_fncall();
			get_next_token();
			if (! (FCurrentTokenType == DELIMITER &&
				   FCurrentToken == SCOLON))
			{
				comp_sntx_err("; expected");
			}
		}

		else if (FCurrentTokenType == FUNCTION &&
				 FCurrentToken == UDFUNCDECL)
		{
			parse_fndecl();
			parse_block(1);
		}
		else if (FCurrentTokenType == DELIMITER &&
				 FCurrentToken == SCOLON)
		{
			// do nothing, just get next token
		}
		else if (FCurrentTokenType == BLOCK &&
				 FCurrentToken == RBRACE)
		{
			t_open_braces--;
			if (required == 1 && t_open_braces == 0)
				return;

			if (t_open_braces < 0)
			{
				comp_sntx_err("unmatched }");
			}
		}
		else if (FCurrentTokenType == BLOCK &&
				 FCurrentToken == LBRACE)
		{
			t_open_braces++;
		}
		else
		{
			if (FErrorMessages.size() == 0)
			{
				comp_sntx_err("Invalid syntax.");
			}
		}

		if (required == 2)
			break;
		get_next_token();

		if (FCompiledPtr >= t_compiled_length)
		{
			FCurrentToken = FINISHED;
			break;
		}
	 } // while (FCurrentToken != FINISHED)

} // parse_block()
//---------------------------------------------------------------------------
void TEditCompiler::parse_if(int& IF_end_block)
{
	int t_StartBLOCK = (int)(FCompiledPtr-4);

    /* if ( <expr> ) { block } [else { block } ] ) */
    // When we enter this function, FCurrentToken == 'F'.
    // This call advances us to the opening parenthesis
    get_next_token();

    if (! (FCurrentTokenType == DELIMITER &&
           FCurrentToken == LPAREN))
    {
		comp_sntx_err("( expected");
    }

    // parse the IF conditional expression
    walk_expr(FCompiledCode.substr(FCompiledPtr, FCompiledCode.size()));

    walk_code_block(IF, t_StartBLOCK, IF_end_block);
}  // parse_if()
//---------------------------------------------------------------------------
void TEditCompiler::parse_else(int& IF_end_block)
{
    /* if ( <expr> ) { block } [else { block } ] ) */
    int t_StartBLOCK = (int)(FCompiledPtr-4);

    // FixEdits 5268
    walk_code_block(ELSE, t_StartBLOCK, IF_end_block);

}  // parse_else()
//---------------------------------------------------------------------------
void TEditCompiler::parse_while()
{	/* while ( <expr> ) {block} */
    int t_StartBLOCK = (int)(FCompiledPtr-4);

    // When we enter this function, the FCurrentToken is 'B';
    // this call advances us to the opening parenthesis... or
    // it should!
	get_next_token();

    if (! (FCurrentTokenType == DELIMITER &&
           FCurrentToken == LPAREN))
    {
		comp_sntx_err("( expected");
    }

    // parse the WHILE conditional expression
    walk_expr(FCompiledCode.substr(FCompiledPtr, FCompiledCode.size()));

    // FixEdits 5268
    int t_dummy = -1;
    walk_code_block(WHILE, t_StartBLOCK, t_dummy);
} // parse_while()
//---------------------------------------------------------------------------
void TEditCompiler::parse_for()
{
    /* for(initialize;test;increment) <statements> */
    int t_StartBLOCK = (int)(FCompiledPtr-4);
    parse_for_prep();
    // FixEdits 5268
    int t_dummy = -1;
    walk_code_block(FOR, t_StartBLOCK, t_dummy);
} // parse_for()
//---------------------------------------------------------------------------
void TEditCompiler::parse_for_prep()
{
    // Primarily called from parse_for() to move past the
    // for(initialize;test;increment) code, it is also called from
    // within walk_code_block() when a nested for(;;) loop is
    // embedded in another for(;;) loop.
    int t_semicolons = 0;
    int t_open_braces = 0;

    // When we enter this function, the FCurrentToken is 'C';
    // this call advances us to the opening parenthesis.
	get_next_token();

    if (! (FCurrentTokenType == DELIMITER &&
           FCurrentToken == LPAREN))
    {
		comp_sntx_err("( expected");
    }

    // count the opening brace
    t_open_braces++;

    // now find the closing brace
    while (FCurrentToken != FINISHED)
    {
        get_next_token();

        if (FCurrentTokenType == DELIMITER)
        {
            if (FCurrentToken == LPAREN)
            {
                t_open_braces++;
            }
            else if (FCurrentToken == RPAREN)
            {
                // This is the end of the for.. statement, so
                // break out of here
                if (--t_open_braces == 0)
                {
                    break;
                }
            }
            else if (FCurrentToken == SCOLON)
            {
                t_semicolons++;
            }
        }
    } // while (FCurrentToken != FINISHED)

    if (! (FCurrentTokenType == DELIMITER &&
           FCurrentToken == RPAREN))
    {
		comp_sntx_err(") expected");
    }

    if (t_semicolons != 2)
    {
		comp_sntx_err("syntax: for(initialize;test;increment)");
    }

} // parse_for_prep()
//---------------------------------------------------------------------------
void TEditCompiler::parse_assign()
{
	/* var=<expr>; */

	unsigned int t_len;
    size_t t_compiled_length = FCompiledCode.length();

    std::string t_err_msg,
           t_var_name;

    unsigned char t_funclist_token;

    // An assignment statement will make calls to get_next_token
    // which could include other variables. Stick to our knitting
    // here and make sure that we are evaluating assignments to
    // the variable that appears to the LEFT of the "=".
    std::string t_leftmost_var_name = FCurrentVarName;

    // As we parse the statement out to the closing semi-colon, we are likely
    // to encounter all manner of variables. Some will be part of a
    // computation ("j = n + 1;"), others will be buried inside
    // function calls ("j = ( n - 1 ) * 100 + VAL(t_strnum);").
    // We'll re-use t_current_var to represent them as we encounter them.
	TVariableDef* t_current_var = NULL;

    // look this up in FVarTable; check for existence
	TVariableDef* t_var = find_variable(FCurrentVarName);
    if (!t_var)
    {
        t_err_msg = "undeclared variable \"" + FCurrentVarName + "\"";
        comp_sntx_err((char*)(t_err_msg.c_str()));
        return;
    }
	else if (t_var->GetVarType() == EEi_CHAR &&
			 t_var->GetVarSize() > 1 &&
        (FCurrentTokenType == OPERATOR && FCurrentToken == ASSIGN))
    {
        t_err_msg = "assignment to variables of type ARRAY OF CHAR must use string functions (strcpy, strcat, etc.)";
        comp_sntx_err((char*)(t_err_msg.c_str()));
        return;
    }

    t_var = NULL;

	if (FCurrentTokenType == DELIMITER && FCurrentToken == LBRACKET)
	{
		// break on ']' (array declarations)
		t_var_name = extract_var_name(t_len);

        // We want to find a numeric constant or a previously-declared
        // variable in here. Start with the constant
        get_next_token();
        if (FCurrentTokenType != NUMBER)
        {
            // test for the previously-declared (and initialized!) var
            // then...
			TVariableDef* t_subscript_var = find_variable(t_var_name);
            if (!t_subscript_var)
            {
				comp_sntx_err("numeric constant or variable required for subscript");
            }
			else if (t_subscript_var->GetVarType() != EEi_INT)
            {
				comp_sntx_err("subscript variable must be of type integer (int)");
            }
            /*
            else if (t_subscript_var->var_value.empty())
            {
                comp_sntx_err("subscript variable has not been initialized");
            }
            */
        }

        // now find the right bracket
        get_next_token();

        if (!(FCurrentTokenType == DELIMITER &&
              FCurrentToken == RBRACKET))
        {
			comp_sntx_err("right bracket required for subscript");
        }

        get_next_token();
	}

	if (FCurrentTokenType == OPERATOR && FCurrentToken == ASSIGN)
	{
        // Advance to the closing semi-colon.
        // Validate tokens along the way.

        // Look for a function call-- the call to get_next_token() will
        // take us right past the "M" to the opening parenthesis, so let's
        // just pause a moment and capture what may turn out to be a
        // very important piece of information (see below, re: LIBFUNC).
        t_funclist_token = FCompiledCode[FCompiledPtr+1];
        get_next_token();

        // What should we look for? Could be a function call, an alphanumeric
        // literal, a variable...
		while (FCompiledPtr < t_compiled_length &&
			   ! (FCurrentTokenType == DELIMITER &&
				  FCurrentToken == SCOLON))
		{
			if (FCurrentTokenType == FUNCTION)
            {
                if (FCurrentToken == LIBFUNC || FCurrentToken == UDFUNCCALL)
                {
                    if (FCurrentToken == LIBFUNC)
                    {
                        // Get the variable we're currently working with
						t_current_var = find_variable(FCurrentVarName);

                        // Check the return type of the function
						if (t_current_var && (FCompiledPtr +1 < t_compiled_length))
                        {
							EEt_IntFuncListType* t_funcdef =
								get_intern_func_def(t_funclist_token);

							if (t_current_var->GetVarType() != t_funcdef->returnType)
                            {
                                t_err_msg = "variable \""+FCurrentVarName+"\" is data type ";
								if (t_current_var->GetVarType() == EEi_CHAR)
                                    t_err_msg += "CHAR";
								else if (t_current_var->GetVarType() == EEi_INT)
                                    t_err_msg += "INT";

								t_err_msg += "; function \"" + (std::string)t_funcdef->funcname + "\" returns ";

                                switch (t_funcdef->returnType)
                                {
                                    case EEi_CHAR:
                                        t_err_msg += "data type CHAR" ;
                                        break;

                                    case EEi_INT:
                                        t_err_msg += "data type INT" ;
                                        break;

                                    default: // NOTHING
                                        t_err_msg += "NOTHING";
                                        break;
                                }

                                comp_sntx_err((char*)(t_err_msg.c_str()));
                            } // if (t_current_var->var_type != t_funcdef->returnType)
                        } // if (t_current_var && (FCompiledPtr +1 < t_compiled_length))
                    } // if (FCurrentToken == LIBFUNC)

					// Parse this function call...
					parse_fncall();
					get_next_token();
				}
			} // if (FCurrentTokenType == FUNCTION)
			else if (FCurrentTokenType == BLOCK)
			{
				if (FCurrentToken == LBRACE)
					comp_sntx_err("illegal character ( '{' ) within assignment statement");
				else
					comp_sntx_err("illegal character ( '}' ) within assignment statement");
				break;
			}
			else
			{
				t_var = find_variable(t_leftmost_var_name);

                if (!t_var)
                {
                    t_err_msg = "variable \""+t_leftmost_var_name+"\" must be declared";
                    comp_sntx_err((char*)(t_err_msg.c_str()));
                }
                else
                {
                    if (FCurrentTokenType == CHARCONST)
                    {
                        // verify that the variable is a char, or
                        // the assignment is to a subscript of an
                        // array of char
						if (t_var->GetVarType() != EEi_CHAR && !t_var->GetIsReported())
                        {
							t_err_msg = "illegal assignment: variable \""+t_var->GetVarName()+"\" is declared to be NUMERIC (INT or LONG)";
                            comp_sntx_err((char*)(t_err_msg.c_str()));
							t_var->SetIsReported(true);
						}
                    }
                    else if (FCurrentTokenType == NUMBER)
                    {
                        // verify that the variable is an EEi_INT
						if ( t_var->GetVarType() != EEi_INT )
                        {
                            // get_next_token() sets a value for
                            // FCurrentNumber when FCurrentTokenType
                            // is NUMBER. Look to see if the user is
                            // assigning zero to an array of char...
                            // because that's legal (setting variable
                            // to NULL).
                            if (FCurrentNumber != 0)
                            {
								t_err_msg = "illegal assignment: variable \""+t_var->GetVarName()+"\" is declared to be CHAR";
                                comp_sntx_err((char*)(t_err_msg.c_str()));
								t_var->SetIsReported(true);
							}
                        }
                    }
                    else if (FCurrentTokenType == STRING)
                    {
                        // verify that the string is not longer
                        // than the allocated space
                        // "string too long for variable"
                        if (t_var->GetVarSize() < FCurrentString.length())
						{
							t_err_msg = "string is longer than space allocated for variable \""+ t_var->GetVarName() + "\"";
							comp_sntx_err((char*)(t_err_msg.c_str()));
							t_var->SetIsReported(true);
						}
					}
					else if (FCurrentTokenType == IDENTIFIER &&
							 FCurrentToken == VAR)
					{
						t_current_var = find_variable(FCurrentVarName);
						if (!t_current_var)
						{
							t_err_msg = "variable \""+FCurrentVarName+"\" must be declared";
							comp_sntx_err((char*)(t_err_msg.c_str()));
						}
					}
				}

				// Keeping up with function calls that could appear
				// anywhere in the assignment line
				t_funclist_token = FCompiledCode[FCompiledPtr+1];

				get_next_token();

				if (FCurrentTokenType == OPERATOR)
				{
					if (FCurrentToken == COMMA)
					{
						// This call came in from parse_vardecl(), and there
						// is a list of variables being declared. We're done
						// with this one, so bail out and return to the caller.
						break;
					}
				}
			} // else to if (FCurrentTokenType == FUNCTION)

			// We're on a mission to determine if the right-hand side of the
			// assignment forgot to place a terminating semi-colon... SO MUCH
			// is legal on the right-hand side that this is really hard to do.
			if (FCurrentTokenType == CONSTRUCT ||
				FCurrentTokenType == KEYWORD)
			{
				comp_sntx_err("; expected at the end of preceding assignment statement");

				break;
			}

		} // while we're still evaluating the right side of the assignment
	}
	else
	{
		comp_sntx_err("= expected");
	}

	// clear the current variable name
	FCurrentVarName = "";

} // parse_assign()
//---------------------------------------------------------------------------
void TEditCompiler::parse_vardecl(char var_type)
{
    /*   var_type will be either EEi_INT or EEi_CHAR */
    size_t t_compiled_length = FCompiledCode.length(),
        t_len;

	while (FCompiledPtr < t_compiled_length &&
           ! (FCurrentTokenType == DELIMITER &&
              FCurrentToken == SCOLON) )
    {
		TVariableDef* t_var = new TVariableDef();
		if (t_var)
        {
			try {
				t_len = (FCompiledCode[FCompiledPtr +1] - '0') * 100;
				t_len += (FCompiledCode[FCompiledPtr +2] - '0') * 10;
				t_len += (FCompiledCode[FCompiledPtr +3] - '0');

				if (t_len >= 0 && ((FCompiledPtr+4 + t_len) < t_compiled_length))
				{
					t_var->SetVarName(FCompiledCode.substr(FCompiledPtr+4,t_len));
					t_var->SetVarType(var_type);
				}
				else
				{
					comp_sntx_err("syntax error in declaration");
				}

				get_next_token();

				if (FCurrentTokenType == OPERATOR && FCurrentToken == MULT)
				{
					get_next_token();
				}

				if (FCurrentTokenType != IDENTIFIER)
				{
					// Be on the look-out for repeating variable types,
					// such as "int t_foo1, int t_foo2"... the second
					// "int" will hit here as a KEYWORD
					if (FCurrentTokenType == KEYWORD &&
					   (FCurrentToken == EEi_CHAR || FCurrentToken == EEi_INT))
					{
						comp_sntx_err("syntax error (variable type in middle of declaration)");
					}
					else
					{
						comp_sntx_err("syntax error (token should be an IDENTIFIER)");
					}
				}

				get_next_token();

				if (FCurrentTokenType == OPERATOR)
				{
					// It's legal to have a comma-separated list in the
					// variable declarations, but no other operator is
					// allowed (we're specifically watching out for
					// assignments in the declaration, e.g. "int t_foo = 72;"
					if (FCurrentToken != COMMA)
					{
						if (FCurrentToken == ASSIGN)
						{
							comp_sntx_err("syntax error: assignment not allowed in declaration");
						}
						else
						{
							comp_sntx_err("syntax error in declaration");
						}
					}
				}
				else if (FCurrentTokenType == DELIMITER && FCurrentToken == LBRACKET)
				{
					get_next_token();

					if (FCurrentTokenType != NUMBER)
					{
						comp_sntx_err("numeric constant required");
					}

					// get_next_token() assigned a value to FCurrentNumber
					// for this variable declaration
					t_var->SetVarSize(FCurrentNumber);

					// now clear FCurrentNumber so it doesn't get used somewhere
					// else that is not appropriate
					FCurrentNumber = 0;

					get_next_token();

					if (!(FCurrentTokenType == DELIMITER &&
						  FCurrentToken == RBRACKET))
					{
						comp_sntx_err("right bracket required");
					}

					get_next_token();
                }

				if (t_var->GetVarType() == '?')
				{
					delete t_var;
					t_var = NULL;
				}
				else
				{
					std::string t_err_msg;
					if (!FVariablesMap->AddVariable(t_var, t_err_msg))
					{
						comp_sntx_err(t_err_msg.c_str());
					}
				}
			}
			catch (...)
			{
                delete t_var;
                throw;
            }
        } // if (t_var)

        /* It is ILLEGAL to make an assignment to a variable at the same
         * time it is being declared. (This was the behavior in EW2, and
         * it causes the engine to choke.
        if (FCurrentTokenType == OPERATOR && FCurrentToken == ASSIGN)
        {
            parse_assign();
        }
        */

		if (FCurrentTokenType == OPERATOR && FCurrentToken == COMMA)
        {
            // variable declarations may be broken over multiple
            // lines. skip over the linefeeds
            while (FCompiledCode[FCompiledPtr] == '\n')
            {
                FCompLineNo++;
                FCompiledPtr++;
            }
			continue;
        }

        if ( ! (FCurrentTokenType == DELIMITER && FCurrentToken == SCOLON) )
        {
		    comp_sntx_err("syntax error in declaration");
        }
	}
} // parse_vardecl()
//---------------------------------------------------------------------------
void TEditCompiler::parse_fndecl()
{
	/* function fn () */
	get_next_token();

	if(! (FCurrentTokenType == DELIMITER && FCurrentToken == LPAREN))
    {
		comp_sntx_err("( expected");
    }

	get_next_token();

	if(!(FCurrentTokenType == DELIMITER && FCurrentToken == RPAREN))
    {
		comp_sntx_err(") expected");
    }
} // parse_fndecl()
//---------------------------------------------------------------------------
void TEditCompiler::parse_fncall()
{
	/* fn ( <expr> [, <expr>] ) */
    int t_open_braces = 0; // counting open/close parentheses
	// Don't check parameters for UDFs (FCurrentToken == 'm')
	// ... there are none.
	bool t_is_LIBFUNC = (FCurrentToken == 'M');
	/*  When we arrive here, we are positioned at the
	 *  opening parenthesis that followed the function code.
	 *
	 *  For example, "Mo(N07t_table,\"11LYMPH98.BIN);"
	 *   you are HERE   ^
	 *
	 *  Back off one and look it up in intern_funclist...
	 *  like this...
	 */
	//std::string t_func_name = get_intern_func_name(FCompiledCode[FCompiledPtr-1]);
	/*  By the way, if you're wondering, the above sample translates to:
	 *  "strcpy(t_table, \"LYMPH98.BIN");"
	 */
	char t_func_code = FCompiledCode[FCompiledPtr-1];
	get_next_token();
	if(!(FCurrentTokenType==DELIMITER && FCurrentToken==LPAREN))
	{
		comp_sntx_err("( expected");
	}
	t_open_braces++;

	if (t_is_LIBFUNC)
	{
		check_params(t_func_code, FCompiledCode.substr(FCompiledPtr, FCompiledCode.size()));
	}

	size_t t_compiled_length = FCompiledCode.length();

	while (t_open_braces > 0 && FCompiledPtr < t_compiled_length)
    {
        get_next_token();

        if (FCurrentTokenType == DELIMITER)
        {
            if (FCurrentToken == LPAREN)
            {
				t_open_braces++;
            }
            else if (FCurrentToken == RPAREN)
            {
                if (--t_open_braces < 0)
                {
                    comp_sntx_err("closing ) encountered without opening (");
                }
            }
            if (FCurrentToken == SCOLON)
            {
                if (t_open_braces > 0)
                {
                    comp_sntx_err(") expected");
                }
                break;
            }
        }
    }

} // parse_fncall()
//---------------------------------------------------------------------------
void TEditCompiler::parse_return()
{
	/* return <expr>; */
    // We want constants, variables or function calls
	size_t t_compiled_length = FCompiledCode.length();
    int t_open_parens = 0;

    int t_save_line = -1;

    bool t_need_operator_or_semicolon = false;

    // FixEdits 5267: Rewritten to catch missing semi-colons that
    // were slipping through.

    // When we come in here, the FCurrentTokenType is KEYWORD
    // and FCurrentToken is 'D' (return)
    while (FCompiledPtr < t_compiled_length &&
           ! (FCurrentTokenType == DELIMITER &&
              FCurrentToken == SCOLON))
    {

        get_next_token();

        if (t_save_line == -1)
        {
            // If I have to report a missing semi-colon, I need to know which
            // line contained the word "RETURN". We're on it now, so this is
            // what we'll use. See bottom of this function for how it is
            // given to comp_sntx_err function.
            t_save_line = (int)FCompLineNo;
        }

        if (t_need_operator_or_semicolon)
        {
            /*
            if (FCurrentTokenType == DELIMITER &&
                FCurrentToken == SCOLON)
            {
                // we're done... get out of here
                break;
			}
            */
            if (FCurrentTokenType == DELIMITER)
            {
                if (FCurrentToken == SCOLON)
                {
                    // we're done... get out of here
                    break;
				}
				else if (FCurrentToken == RPAREN)
				{
					// Otherwise, continue processing. Might be a bouquet of
					// parentheses
				}
				else //if (FCurrentToken == COLON)
				{
					comp_sntx_err("RETURN statements must end with semi-colon (;)");
				}
			}
            else if (!(FCurrentTokenType == OPERATOR))
            {
                // this is an error condition, but we'll handle it below
                break;
            }

            // otherwise, reset this flag and go around again
            t_need_operator_or_semicolon = false;
        }

        if (FCurrentTokenType == NUMCONST)
        {
			// 12/05/2016: Loosened this to check for a semi-colon or operator
			// following a RETURN statement plus ANY NUMCONST.
			t_need_operator_or_semicolon = true;
			/*
			// 'Q' is the compiled value for '0', and it means fail,
			// 'R' is the compiled value for '1', and it means pass.
			// Make sure it's followed by a semi-colon
			if ((FCurrentToken == '0') ||
				(FCurrentToken == '1'))
			{
				t_need_operator_or_semicolon = true;
			}
			*/
		}
		else if (FCurrentTokenType == NUMBER)
		{
			t_need_operator_or_semicolon = true;
		}
        else if (FCurrentTokenType == DELIMITER)
        {
            if (FCurrentToken == LPAREN)
            {
                t_open_parens++;
            }
            else if (FCurrentToken == RPAREN)
            {
                if (--t_open_parens < 0)
                {
                    comp_sntx_err("closing ) encountered without opening (");
                }

                t_need_operator_or_semicolon = true;
            }
        }
        else if (FCurrentTokenType == OPERATOR)
        {
            if (FCurrentToken == ASSIGN)
            {
                comp_sntx_err("Possibly incorrect assignment");
            }
            else if (FCurrentToken == COMMA)
            {
                comp_sntx_err("COMMA not valid in RETURN statement");
            }
        }
        else if (FCurrentTokenType == FUNCTION)
        {
            // could be: RETURN INLIST(...);
            // but could ALSO be: RETURN MATCH(...) OR EMPTY(...);
            // In order to test for the terminating semi-colon,
            // we have to let it continue the processing looking
            // for more OPERATORs, or finding a semi-colon.
			parse_fncall();

			// If the function was MATCH or INLIST, we have increased
			// the size of the code (translated match pattern to format
			// for std::regex); reset the controller for this loop
			t_compiled_length = FCompiledCode.length();

            t_need_operator_or_semicolon = true;
        }


        else if (FCurrentTokenType == CHARCONST)
        {
            get_next_token();
            if (FCurrentTokenType != IDENTIFIER)
            {
                comp_sntx_err("expected IDENTIFIER");
            }
            get_next_token();
            if (FCurrentTokenType != CHARCONST)
            {
                comp_sntx_err("unterminated character constant (\"'\" expected)");
            }

            t_need_operator_or_semicolon = true;
        }
        else if (FCurrentTokenType == IDENTIFIER &&
                 FCurrentToken == VAR)
        {
            // look this up in FVarTable; check for existence
			TVariableDef* t_var = find_variable(FCurrentVarName);
            if (!t_var)
            {
                std::string t_err_msg = "undeclared variable \"" + FCurrentVarName + "\"";
                comp_sntx_err((char*)(t_err_msg.c_str()));
            }
            FCurrentVarName = "";

            t_need_operator_or_semicolon = true;
        }
    }
    // Did we close all open parentheses?
    if (t_open_parens > 0)
    {
        comp_sntx_err(") expected");
    }
    // If we didn't finish on the semi-colon, report it.
    if (! (FCurrentTokenType == DELIMITER &&
           FCurrentToken == SCOLON))
    {
        // In order to get this to display the correct line number,
        // we have to lie to comp_sntx_err.
        size_t t_save2 = FCompLineNo;
        FCompLineNo = t_save_line;
		comp_sntx_err("return statement must end with semi-colon");
		// but let's not allow subsequent errors to be off-line.
		FCompLineNo = t_save2;

	}
} // parse_return()
//---------------------------------------------------------------------------
std::string TEditCompiler::get_intern_func_name(char func_code)
{
	unsigned int t_func_count = sizeof(intern_funclist) / sizeof(EEt_IntFuncListType);
	std::string t_func_name;
	// intern_funclist has an empty element at the end; decrement t_func_count
	for (unsigned int t_idx = 0; t_idx < t_func_count-1; t_idx++)
	{
		if (intern_funclist[t_idx].token == func_code)
		{
			t_func_name.assign(intern_funclist[t_idx].funcname);
			break;
		}
	}
	return t_func_name;
} // get_intern_func_name
//---------------------------------------------------------------------------
char TEditCompiler::get_intern_func_code(std::string func_name)
{
	unsigned int t_func_count = sizeof(intern_funclist) / sizeof(EEt_IntFuncListType);
	char t_func_code = ' ';
	// intern_funclist has an empty element at the end; decrement t_func_count
	for (unsigned int t_idx = 0; t_idx < t_func_count-1; t_idx++)
	{
		if (std::strcmp(intern_funclist[t_idx].funcname, func_name.c_str()) == 0)
		{
			t_func_code = intern_funclist[t_idx].token;
			break;
		}
	}

	return t_func_code;
} // get_intern_func_code
//---------------------------------------------------------------------------
EEt_IntFuncListType* TEditCompiler::get_intern_func_def(char func_code)
{
	unsigned int t_func_count = sizeof(intern_funclist) / sizeof(EEt_IntFuncListType);
	EEt_IntFuncListType* t_func_def = NULL;

	// intern_funclist has an empty element at the end; decrement t_func_count
	for (unsigned int t_idx = 0; t_idx < t_func_count-1; t_idx++)
	{
		if (intern_funclist[t_idx].token == func_code)
		{
			t_func_def = &intern_funclist[t_idx];
			break;
		}
	}

	if (!t_func_def)
	{
		char t_func_code[2];
		t_func_code[0] = func_code;
		t_func_code[1] = '\x0';
		std::string t_error_msg = "Syntax checker cannot find EDITS Language Function coded '";
		t_error_msg += t_func_code;
		t_error_msg += "'";

		FEngineErrorsPump->AddError("TEditCompiler", "get_intern_func_def",
			t_error_msg);

		throw std::runtime_error("Something bad has happened");
	}
	return t_func_def;

} // get_intern_func_def
//---------------------------------------------------------------------------
void TEditCompiler::check_params(unsigned char func_code, std::string the_code)
{
	//  This function is called by parse_block() when we establish that
	//  we have come upon a function call.
	//
	//  Note that we want to make sure that functions that do lookups
	//  into Tables or Messages use parameters that exist in the
	//  database. Note that fields are easier to spot, and are
	//  already being verified in function field_name_to_field_num().
	//
	//  Tables:    BINLOOKUP, ILOOKUP, LOOKUP, RLOOKUP
	//  Messages:  ERROR_MSG, SAVE_ERROR_TEXT, SAVE_WARNING_TEXT, SET_WARNING

	unsigned int t_idx,
				 t_param_cnt,
				 t_expected_type; // EEi_INT or EEi_CHAR

	std::string t_func_name,
           t_err_msg,
           t_one_param;

    std::vector<std::string> t_params;
	bool t_isVariable;

	// for SQLLookup and SQLRangeLookup
	TTableDef* t_table;

	t_func_name = get_intern_func_name(func_code);

	// uppercase the function name for use in error messages
	std::transform(t_func_name.begin(), t_func_name.end(), t_func_name.begin(), toupper);

	if (parse_params(the_code, t_params))
	{
        // validate number of parameters for each case
        t_param_cnt = (int)t_params.size();

        switch ((char)func_code)
        {
            case 'A':   /* at*/
                // long AT(search_for_str, search_in_str [, width])
                if (verify_param_count(t_func_name, t_param_cnt, 2, 3))
                {
                    // AT: Verify data types
                    for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
                    {
                        switch (t_idx)
                        {
							case 0:
                            case 1:
                                t_expected_type = EEi_CHAR;
                                break;
                            case 2:
                                t_expected_type = EEi_INT;
                                break;
                        }
                        check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
                    } // for (...)

                } // verify_param_cnt
                break;
			case 'K':   /* empty*/
			case ']':   /* nameexpr */
				// bool EMPTY(string | field)
				// char* NAMEEXPR(string)
				if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
                {
                    check_param_type(t_func_name, t_params[0], 1, EEi_CHAR);
                }
                break;
            case 'L':   /* error_msg */
				// int ERROR_MSG( number )
                if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
                {
                    check_param_type(t_func_name, t_params[0], 1, EEi_INT);
					message_number_is_valid(t_func_name, t_params[0]);
				}
                break;
            case 'M':   /* error_text*/
                // int ERROR_TEXT( string )
                if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
                {
                    check_param_type(t_func_name, t_params[0], 1, EEi_CHAR);
                }
                break;
            case 'N':   /* fmtstr */
                // char* FMTSTR(var_name, format_str, number)
                if (verify_param_count(t_func_name, t_param_cnt, 3, 3))
                {
                    // Validate the parameters (data type)
                    for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0:
							case 1:
								t_expected_type = EEi_CHAR;
								break;
							case 2:
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (...)
				}
				break;
/*
			case 'R':   // getvar //
				// CHAR* or NUMERIC GETVAR(var_name [, element])
				if (verify_param_count(t_func_name, t_param_cnt, 1, 2))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0:
								t_expected_type = EEi_CHAR;
								break;
							case 1:
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (...)
				}
				break;
*/
			case 'U':   /* inlist */
				// int INLIST(string|field, list of values or ranges
				//             [,match_pattern
				//             [,start_pos
				//             [,length ] ] ] )
				if (verify_param_count(t_func_name, t_param_cnt, 2, 5))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0: // string | field
							case 1: // list of values
							case 2: // optional match pattern
								t_expected_type = EEi_CHAR;
								break;
							case 3: // optional start_pos
							case 4: // optional length
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
						// Alan Houser noted an error that slipped through a COC edit
						// where the INLIST call included a range like "00,82-85-89,99"
						// when what was intended was                  "00,82,85-89,99"
						// Let's issue a warning if the list of values
						// displays two range indicators within an area delimited
						// by surrounding quotes or commas
						if (t_idx == 1 && t_expected_type == EEi_CHAR)
						{
							int t_hyphen_count = 0;
							std::string t_param = t_params[t_idx];
							for (size_t t_kdx = 0; t_kdx < t_param.length(); t_kdx++)
							{
								if (t_param[t_kdx] == '-')
									t_hyphen_count += 1;
								else if (t_param[t_kdx] == ',')
								{
									if (t_hyphen_count > 0)
										t_hyphen_count -= 1;
								}

								if (t_hyphen_count > 1)
								{
									comp_sntx_warn("ambiguous call to INLIST");
									break;
								}
							}
						}
					} // for (...)
				}

				break;
			case 'V':   /* justified */
				// int JUSTIFIED(string|field, LEFT | RIGHT)
				if (verify_param_count(t_func_name, t_param_cnt, 2, 2))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0:
								t_expected_type = EEi_CHAR;
								break;
							case 1:
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (...)
				}

				break;
			case 'W':   /* left */
			case 'e':   /* right*/
				// char *LEFT(string|field, num_chars)
				// char *RIGHT(string|field, num_chars)
				if (verify_param_count(t_func_name, t_param_cnt, 2, 2))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0:
								t_expected_type = EEi_CHAR;
								break;
							case 1:
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (...)
				}
				break;
			case 'S':   /* ilookup*/
			case 'X':   /* lookup */
			case 'f':   /* rlookup*/
			// ILOOKUP(<field>, <char *table_name>|<char *table_name.item>)
			// LOOKUP(<field>, <char *table_name>|<char *table_name.item>)
			// RLOOKUP (<field>, <char *table_name>|<char *table_name.item>)
			// For example:
			// (#S"TNM Path Stage Group","AJC5STGP.DBF.GRPSTAGE")
			// or:
			// ( SomeVar, "AJC5STGP.DBF.GRPSTAGE")
				if (verify_param_count(t_func_name, t_param_cnt, 2, 2))
				{
					// Verify data types
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						if (check_param_type(t_func_name, t_params[t_idx], t_idx+1, EEi_CHAR))
						{
							if (t_idx == 1) // table name
							{
								t_one_param = t_params[t_idx]; // table name
								t_isVariable = (t_one_param[0] == 'N'); // IDENTIFIER VAR

								if (t_isVariable)
								{
									t_one_param = t_one_param.substr(4, t_one_param.size());
									// look these up in FVarTable; check for existence and content
									TVariableDef* t_var = find_variable(t_one_param);
									if (t_var)
									{
										std::string t_var_value = t_var->GetVarValue();
										if (!t_var_value.empty())
										{
											find_table(t_func_name, t_var_value.c_str(), ".DBF");
										}
									}
								} // if (t_isVariable)
								else
								{
									if (t_one_param[0] == '"')
									{
										t_one_param = t_one_param.substr(4, t_one_param.size());
										find_table(t_func_name, t_one_param, ".DBF");
									}
								}
							} // if (t_idx == 0)
						} // if (check_param_type(...)

					} // for (...)
				} // lookup, ilookup, rlookup
				break;

			case '{':   /* sqllookup */
			case '}':   /* sqlrangelookup */
				// SQLLookup(<char* table_name>, <char* columns list>, <char* value>)
				// SQLRangeLookup(<char* table_name>, <char* columns list>, <char* value>)
				// For example:
				// SQLRangeLookup("STATEZIP", "State, ZipLow", "CA94402")
				// or
				// SQLRangeLookup(TableNameVar, ColListVar, t_search_str)
				if (verify_param_count(t_func_name, t_param_cnt, 3, 3))
				{
					if (check_param_type(t_func_name, t_params[0], 1, EEi_CHAR))
					{
						// Check for existence of the TableName variable
						t_one_param = t_params[0];
						t_isVariable = (t_one_param[0] == 'N'); // IDENTIFIER VAR
						if (t_isVariable)
						{
							t_one_param = t_one_param.substr(4, t_one_param.size());
							// look these up in FVarTable; check for existence and content
							TVariableDef* t_var = find_variable(t_one_param);
							if (t_var)
							{
								std::string t_var_value = t_var->GetVarValue();
								if (!t_var_value.empty())
								{
									t_table = find_sql_table(t_var_value.c_str());
								}
							}
						}
						else
						{
							if (t_one_param[0] == '"')
							{
								t_one_param = t_one_param.substr(4, t_one_param.size());
								t_table = find_sql_table(t_one_param);
							}
						}
					} // if (check_param_type(... for first param (Table Name)
// 08/14/2017: This is the fix for the AV caused when SQLLookup/SQLRangeLookup
// reference a table that is not in the metafile. This would be Engine version "5.0.0.18"
					// Now check the field or fields list
					if (t_table && check_param_type(t_func_name, t_params[1], 1, EEi_CHAR))
					{
						t_one_param = t_params[1];
						t_isVariable = (t_one_param[0] == 'N'); // IDENTIFIER VAR
						if (t_isVariable)
						{
							t_one_param = t_one_param.substr(4, t_one_param.size());
							// look these up in FVarTable; check for existence and content
							TVariableDef* t_var = find_variable(t_one_param);
							if (t_var)
							{
								std::string t_var_value = t_var->GetVarValue();
								if (!t_var_value.empty())
								{
									find_sql_fields(t_table, t_var_value.c_str());
								}
							}
						}
						else
						{
							if (t_one_param[0] == '"')
							{
								t_one_param = t_one_param.substr(4, t_one_param.size());
								find_sql_fields(t_table, t_one_param);
							}
						}
					}

					// Finally, check the type of the last parameter
					check_param_type(t_func_name, t_params[2], 1, EEi_CHAR);
				}
				break;



			case 'Y':   /* lower*/
				// char *LOWER(string|field)
				if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
				{
					check_param_type(t_func_name, t_params[0], 1, EEi_CHAR);
				}
				break;
			case 'Z':   /* match*/
				// int MATCH(string, pattern);
				// Example: MATCH(#S"Accession Number--Hosp","ddddddddd")
				// MZ(#16,\"09ddddddddd)

				// MATCH: Verify number of parameters
				if (verify_param_count(t_func_name, t_param_cnt, 2, 2))
				{
					// MATCH: Verify data types
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, EEi_CHAR);
					}  // MATCH: for (t_idx...
				} // verifying t_param_cnt
				break;
			case 'h':   /* save_text             104 */
				// int SAVE_TEXT(msg1 [, msg2 [...msgn]])
				if (verify_param_count(t_func_name, t_param_cnt, 1, 99))
				{
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, EEi_CHAR);
					}
				}
				break;
			case 'g':   /* save_error_text       103 */
			case '':   /* save_warning_text     133 */
				// int SAVE_ERROR_TEXT(msg1 [, msg2 [...msgn]])
				// int SAVE_WARNING_TEXT(msg1 [, msg2 [...msgn]])
				if (verify_param_count(t_func_name, t_param_cnt, 1, 99))
				{
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, EEi_INT);
						message_number_is_valid(t_func_name, t_params[t_idx]);
					}
				}
				if ((char)func_code == '') /* save_warning_text     133 */
				{
					comp_sntx_warn("save_warning_text is deprecated; set_warning is preferred (see documentation)");
				}
				break;
			//case 'i':   /* set_error */
			case '':   /* set_warning */
				// int SET_WARNING(flag1 [, flag2 [...flagn]])
				if (verify_param_count(t_func_name, t_param_cnt, 1, 99))
				{
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, EEi_INT);
						message_number_is_valid(t_func_name, t_params[t_idx]);
					}
				}
				break;
			case 'k':   /* setvar */
				// void SETVAR(var_name,value [, element])

				// Verify correct number of parameters
				if (verify_param_count(t_func_name, t_param_cnt, 2, 3))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0: // var_name
							case 1: // value
								t_expected_type = EEi_CHAR;
								break;
							case 2: // element
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (t_idx...)
				} // verify_param_count()

				break;
			case 'm':   /* strcat */
			case 'n':   /* strcmp */
			case 'o':   /* strcpy */
				// char *STRCAT(str1, str2 [,num])
				// int STRCMP(str1, str2 [,num])
				// int STRCPY(dest, source [,[-]num])
				// Example using strcpy:
				// "Mo(N07t_table,\"11LYMPH98.BIN);"

				// Verify correct number of parameters
				if (verify_param_count(t_func_name, t_param_cnt, 2, 3))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0: // dest
							case 1: // source
								t_expected_type = EEi_CHAR;
								break;
							case 2: // num
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (t_idx...)
				} // verify_param_count()
				break;
			case 'p':   /* strlen */
				// int STRLEN(string|field)
				if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
				{
					check_param_type(t_func_name, t_params[0], 1, EEi_CHAR);
				}
				break;
			case 'q':   /* substr */
				// char *SUBSTR(string|field, start_pos [,num_chars])
				if (verify_param_count(t_func_name, t_param_cnt, 2, 3))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0: // string|field
								t_expected_type = EEi_CHAR;
								break;
							case 1: // start_pos
							case 2: // num_chars
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (t_idx...)
				} // verify_param_count()

				break;
			case 's':   /* trim */
				// char *TRIM(string|field, LEFT | RIGHT | BOTH )
				if (verify_param_count(t_func_name, t_param_cnt, 2, 2))
				{
					// Verify data types
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0:
								t_expected_type = EEi_CHAR;
								break;
							case 1:
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (...)
				} // verify_param_cnt
				break;
			case 't':   /* upper*/
				// char *UPPER(string|field)
				if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
				{
					check_param_type(t_func_name, t_params[0], 1, EEi_CHAR);
				}
				break;
			case '':   /* val*/
				// long VAL(string|field)
				if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
				{
					check_param_type(t_func_name, t_params[0], 1, EEi_CHAR);
				}
				break;
			case '':  /* externaldll Alt+0135 */
				// can we verify?
				break;

			case '':   /* allow_future_date_iop Alt+0147 */
				if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
				{
					check_param_type(t_func_name, t_params[0], 1, EEi_INT);

					if (t_params[0][0] == '-')
						comp_sntx_err("allow_future_date_iop parameter must be >= 0");

				}
				break;

			case '':   /* date_day_iop Alt+0136 */
			case '':   /* date_month_iop Alt+0137 */
			case '':   /* date_year_iop Alt+0138 */
			case '':   /* valid_date_iop Alt+0144 */
				if (verify_param_count(t_func_name, t_param_cnt, 1, 1))
				{
					check_param_type(t_func_name, t_params[0], 1, EEi_CHAR);
					// make sure date field parameter is type D1
					check_date_type(t_func_name, t_params[0]);
				}
				break;

			case '':   /* datecmp_iop Alt+0139 */
			case '':   /* daydiff_iop Alt+0140 */
			case '':   /* monthdiff_iop Alt+0142 */
			case '':   /* yeardiff_iop Alt+0145 */
				if (verify_param_count(t_func_name, t_param_cnt, 3, 3))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0:
							case 1:
								t_expected_type = EEi_CHAR;
								break;
							case 2:
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (...)
					// make sure any date field parameters are type D1
					check_date_type(t_func_name, t_params[0]);
					check_date_type(t_func_name, t_params[1]);
				}

				break;

			case '':   /* dayinterval_iop Alt+0141 */
			case '':   /* monthinterval_iop Alt+0143 */
			case '':   /* yearinterval_iop Alt+0146 */
				if (verify_param_count(t_func_name, t_param_cnt, 5, 5))
				{
					// Validate the parameters (data type)
					for (t_idx = 0; t_idx < t_param_cnt; t_idx++)
					{
						switch (t_idx)
						{
							case 0:
							case 1:
								t_expected_type = EEi_CHAR;
								break;
							case 2:
							case 3:
							case 4:
								t_expected_type = EEi_INT;
								break;
						}
						check_param_type(t_func_name, t_params[t_idx], t_idx+1, t_expected_type);
					} // for (...)

					// make sure any date field parameters are type D1
					check_date_type(t_func_name, t_params[0]);
					check_date_type(t_func_name, t_params[1]);
				}

				break;
		}
	} // if (parse_params...)
} // check_params
//---------------------------------------------------------------------------
bool TEditCompiler::parse_params(std::string the_code,
    std::vector<std::string>& params)
{
    bool t_result = true;
    int t_idx,
        t_code_len = (int)the_code.size(),
        t_start_pos = 0,
        t_end_pos,
        t_open_braces = 0,
        t_str_len,
        t_nested_func_cnt = 0; // function calls in parameters
    std::string t_code,
           t_param;

    t_end_pos = -1; // the_code.find_first_of(")", 0)-1;

    // count the opening parenthesis
    t_open_braces++;
    // find the LAST closing parenthesis (they may be nested)
    for (t_idx = 0; t_idx < t_code_len; t_idx++)
    {
        if (the_code[t_idx] == 'M' &&
            (t_idx+2 < t_code_len && the_code[t_idx+2] == '('))
        {
            t_nested_func_cnt++;
        }
        else if (the_code[t_idx] == '(')
        {
            t_open_braces++;
        }
		else if (the_code[t_idx] == ')')
        {
            if (--t_open_braces == 0)
            {
                t_end_pos = t_idx;
                break;
            }
        }

    }

    if (t_end_pos > t_start_pos)
    {
        t_code = the_code.substr(t_start_pos, t_end_pos - t_start_pos);
        t_code_len = (int)t_code.size();

        while (t_start_pos < t_code_len)
        {
			if (params.size() > 0)
			{
				// After processing the first parameter, and if we are still
				// evaluating parameters, make sure we have a comma between the
				// last parameter we saw and the one coming up.
				if ((t_end_pos < t_code_len) &&
					(t_code[t_end_pos] != ','))
				{
					comp_sntx_err("Syntax error: review function parameters");
				}
			}

			if (t_nested_func_cnt > 0 &&
                (t_code[t_start_pos] == 'M' &&
                 (t_start_pos +2 < t_code_len && t_code[t_start_pos+2] == '(')))
            {
                // Function calls can be nested many layers deep.
                // For now, just take one entire call (even if it
                // contains additional calls) and return it as a
                // single parameter. (We'll evaluate the return data
                // type of a nested function call in check_param_type().)
                for (t_end_pos = t_start_pos; t_end_pos < t_code_len; t_end_pos++)
                {
                    if (t_code[t_end_pos] == '(')
                        t_open_braces++;
                    else if (t_code[t_end_pos] == ')')
                    {
                        if (--t_open_braces == 0)
                        {
                            t_end_pos++;
                            break;
                        }
                    }
                }

                t_nested_func_cnt--;
            }
            else if (t_code[t_start_pos] == '"')
            {
                // Parameter is a string, which could contain all manner
                // of "trigger" characters (commas, quotes, parentheses)
                // Fortunately, the next two bytes tell us how long the
                // string is, so advance to the end of string.
/*
				if (t_start_pos + 2 < t_code_len)
				{
					t_str_len = std::atoi((t_code.substr(t_start_pos+1,2)).c_str());
					t_end_pos = t_start_pos + t_str_len + 3;
				}
*/
				if (t_start_pos + 3 < t_code_len)
				{
					std::string t_str_size = t_code.substr(t_start_pos+1,3);
					// This is NOT a string representation of a literal number.
					// We have to perform the dance to turn each ASCII byte back
					// into an integer value.

					//t_str_len = std::atoi((t_code.substr(t_start_pos+1,3)).c_str());
					t_str_len = (t_str_size[0] - '0') * 100;
					t_str_len += ((t_str_size[1] - '0') * 10) - t_str_len;
					t_str_len += (t_str_size[2] - '0');

					t_end_pos = t_start_pos + t_str_len + 4 ;
				}
				else
                {
                    comp_sntx_err("parse_params() encountered unexpected error parsing string parameter");
                    break;
                }
            }
            else
            {
                // The rest of you men, follow me... look for the next comma,
                // or run to the end of the parameter list.
                for (t_end_pos = t_start_pos; t_end_pos < t_code_len; t_end_pos++)
                {
                    if (t_code[t_end_pos] == ',')
                        break;
                }
            }
            if (t_end_pos >= t_start_pos)
            {
                // Note: If t_end_pos == t_start_pos, somebody
                // passed us an empty parameter (NULL).
                t_param = t_code.substr(t_start_pos, t_end_pos - t_start_pos);
                params.insert(params.end(), t_param);
                t_start_pos = t_end_pos +1;
                // If a function call is broken over multiple lines,
                // we have to skip past the linefeeds
				//while (iswhite(t_code[t_start_pos]))
				while (t_start_pos <= t_code_len &&
				       iswhite(t_code[t_start_pos]))
                    t_start_pos++;
            }
        }
    }
    else
    {
		comp_sntx_err("Function call without closing parenthesis");
        t_result = false;
    }
    return t_result;
} // parse_params
//---------------------------------------------------------------------------
bool TEditCompiler::verify_param_count(
    std::string func_name,    // edits logic function name
    int param_count,     // parameters found by parse_params()
    int low_expected,    // fewest number of parameters expected
    int hi_expected)     // greatest number of params expected
                         // if this number is 99, unlimited params allowed
{
    // Called by check_params, this function reports any problems
    // with the number of parameters to a function call. If it
    // returns false, the compiler does not try to analyze the
    // function call further.

    bool t_result = true; // until proven otherwise
    std::string t_err_msg = "";

    // param_count is how many parameters we found using parse_params().
    // expected is how many parameters
    if (param_count < low_expected)
    {
        t_err_msg = func_name + " error: too few parameters";
    }
    else
    {
        if (param_count > hi_expected)
        {
            if (hi_expected < 99)
            {
                t_err_msg = func_name + " error: too many parameters";
            }
        }
    }

    if (t_err_msg.size() > 0)
    {
        comp_sntx_err((char*)(t_err_msg.c_str()));
        t_result = false;
    }

    return t_result;
} // verify_param_count
//---------------------------------------------------------------------------
bool TEditCompiler::message_number_is_valid(std::string func_name,
	std::string full_param)
{
	// Called by check_params for functions ERROR_MSG, SET_WARNING,
	// SAVE_ERROR_TEXT and SAVE_WARNING_TEXT to ensure that the message
	//  numbers in the parameters list exist in EE_Messages
	bool t_found = false;

	int t_msg_number = std::atoi(full_param.substr(4, full_param.length()).c_str());
	char t_buffer[500];

	std::unique_ptr<TSQLQuery> t_qry(new TSQLQuery(FSQLDB));
	std::snprintf(t_buffer, 500,
		"SELECT [PKey] FROM [EE_Messages] WHERE [MessageNumber] = %d;",
		t_msg_number);

	t_qry->SetSQLText(t_buffer);
	t_qry->SetIsActive(true);
	if (!t_qry->Eof())
	{
		if (t_qry->GetIntFieldByName("PKey") > 0)
		{
			t_found = true;
		}
	}
	t_qry->SetIsActive(false);

	if (!t_found)
	{
		std::snprintf(t_buffer, 500,
			"in function %s, Message #%d does not exist",
			func_name.c_str(), t_msg_number);
		comp_sntx_err(t_buffer);
	}

	return t_found;
} // message_number_is_valid
//---------------------------------------------------------------------------
bool TEditCompiler::check_param_type(std::string func_name,
    std::string full_param, int param_num, int expected_type)
{
    // Called by check_params for each parameter to determine
    // whether the appropriate data type was used. Some functions
    // need to know whether there are any problems with the
    // parameters (the lookup functions), so we'll return a status.

    bool t_result = true; // until proven otherwise

    // Beware embedded operators in parameters
    std::string t_ops[10] = { "+" ,  // ADD
							  "-" ,  // SUB
							  "*" ,  // MULT
							  "/" ,  // DIV
							  "==",  // EQ
							  "!=",  // NE
							  "<" ,  // LT
							  "<=",  // LE
							  ">" ,  // GT
							  ">="}; // GE

	size_t t_len = full_param.size(),
		   t_op_idx;  // pointer into t_ops array

	/* bcc64.exe (CLANG) reported this warning when t_op_pos declared unsigned int:
		 Warning comparison of constant 18446744073709551615 with expression of type 'unsigned int' is always true

	   Found this explanation:
	   http://www.cplusplus.com/forum/general/135598/

	   Look for explanation from MiiNiPa on June 14, 2014:
		 npos is often implemented as maximum value of string::size_type,
		 which is unsigned long long on 64bit machines.
		 unsigned is often 32 bit in size
		 When you are asigning result of search to unsigned int it gets truncated
		 and subsequent comparsion for inequality returns true even if element
		 is not found. Then you are trying to access element UINT_MAX

	   So I changed the declaration as shown here, and now the warning is gone.
	   In 64-bit, size_t is "unsigned long long"
	 */
    size_t t_op_pos;  // position of an operator in the parameter

    bool t_isVariable = full_param[0] == 'N',
         t_isConstant = (full_param[0] >= 'O' && full_param[0] <= 'T'),
         t_isField    = full_param[0] == '#',
         t_isFunction = (full_param[0] == 'M' && (t_len > 2 && full_param[2] == '(')),
         t_isNumber   = full_param[0] == 'U',
         t_isString   = full_param[0] == '"',
         t_has_op     = false;

    std::string t_one_param,
           t_err_msg = "";

    unsigned int t_start_pos = (t_isField ? 1 : 4);

    // Verify that a constant is expected at this position
    if (t_isConstant)
    {
        return true;
	}


//////////////////////////////////////////////
	if (t_isString)
	{
		char t_buff[4];
		t_buff[0] = full_param[1];
		t_buff[1] = full_param[2];
		t_buff[2] = full_param[3];
		t_buff[3] = '\x0';

		int t_value_size = std::atoi(t_buff);
		if (t_value_size == 0)
		{
			// parameter is an empty string
			return true;
		}
	}
//////////////////////////////////////////////

	if (t_start_pos >= t_len)
    {
        t_err_msg = "check_param_type error: Function-" +
            func_name + ", Parameter-" +
            full_param;
        comp_sntx_err((char *)(t_err_msg.c_str()));
        return false;
    }

    t_one_param = full_param.substr(t_start_pos, full_param.size());

    // Does this parameter have an embedded operator?
    for (t_op_idx = 0; t_op_idx < 10; t_op_idx++)
    {
        t_op_pos = t_one_param.find_first_of(t_ops[t_op_idx]);
        if (t_op_pos != std::string::npos)
        {
            t_has_op = true;
            break;
        }
    }

    if (t_isVariable)
    {
        // look these up in FVarTable; check for existence and content
		TVariableDef* t_var;
        if (t_has_op)
			t_var = find_variable(t_one_param.substr(0, t_op_pos));
        else
			t_var = find_variable(t_one_param);
        if (!t_var)
        {
            no_such_variable(param_num, func_name, t_one_param);
        }
        else
        {
			if (t_var->GetVarType() != expected_type)
            {
                t_err_msg = func_name + " error: Variable \"" + t_one_param + "\" must be data type ";
                if (expected_type == EEi_CHAR)
                    t_err_msg += "CHAR[] (string)";
                else
                    t_err_msg += "INT or LONG";

            }
            /* RATS! We can't do this, because of declaration data type "tablevar".
               That is a reference to fields in a table, and pre-declares the
               field names (comparable to old Clipper 5 FIELD declarations). They
			   are NEVER declared as arrays of char, and the tokenize() process is
               obliged to convert them to EEi_CHAR (see translation tables, above),
               so there is no way for me to detect them here.

               Consequently, we have to let the engine barf up an error message
               when somebody uses syntax like this:

                char t_foo;
                strcpy(t_foo, "howdy"); // <== buffer overflow

            else if (t_var->var_type == EEi_CHAR)
            {
                if (t_var->var_size == 1)
                {
                    t_err_msg = func_name + " error: Variable \"" + t_one_param + "\" must be declared as an ARRAY OF CHAR";
                }
            }
            */
        }
    } // if (t_isVariable)
    else
    {
        if (t_isFunction)
        {
			unsigned char t_token = full_param[1];
			EEt_IntFuncListType* t_funcdef = get_intern_func_def(t_token);
			std::string t_param_func = t_funcdef->funcname;

            // Validate the function call for the function that is a
            // parameter to the main function identified as func_name.
			check_params(t_token, full_param.substr(3, full_param.length()));

/*
			// Function getvar can return EITHER EEi_INT OR EEi_CHAR, so we
			// won't try to process it. But all others will return one of
			// NOTHING, EEi_INT or EEi_CHAR.
			if (t_param_func.compare("getvar") != 0)
            {
				if (t_funcdef->returnType != expected_type)
                {
                    t_err_msg = func_name + " error: function parameter\"" + t_param_func + "\" returns ";
					switch (t_funcdef->returnType)
                    {
                        case EEi_CHAR:
                            t_err_msg += "a CHAR; " ;
                            break;

                        case EEi_INT:
                            t_err_msg += "an INT; " ;
                            break;

                        default: // NOTHING
                            t_err_msg += "NOTHING; ";
                            break;
                    }

                    t_err_msg += "expected data type is ";

                    if (expected_type == EEi_CHAR)
                        t_err_msg += "CHAR";
                    else
                        t_err_msg += "INT";
                }
			}
*/
        }
        else if (!t_isField)
        {
            if (t_isString && expected_type != EEi_CHAR)
            {
                t_err_msg = func_name + " error: Parameter \"" + t_one_param + "\" must be data type INT or LONG";
            }
            else if (t_isNumber && expected_type != EEi_INT)
            {
                t_err_msg = func_name + " error: Parameter \"" + t_one_param + "\" must be data type CHAR[] (string)";
            }
		}
	}

	if (t_err_msg.size() > 0)
	{
		comp_sntx_err((char*)(t_err_msg.c_str()));
		t_result = false;
	}
	return t_result;
} // check_param_type
//---------------------------------------------------------------------------
TVariableDef* TEditCompiler::find_variable(std::string var_name)
{
	lowercase(var_name);
	return FVariablesMap->GetVariable(var_name);
} // find_variable
//---------------------------------------------------------------------------
void TEditCompiler::find_table(std::string func_name,
	std::string table_name, std::string extension)
{
	std::string t_lookup_idx = "",
				 t_err_msg = "";

	size_t t_pos;

	uppercase(table_name);
	uppercase(func_name);
	uppercase(extension);

	// Unless we have EMF_Convert remove ".DBF" from table names, and maybe
	// even if we do, we will simply ignore the disk file extension by stripping
	// it off of the table_name parameter. For the time being, the caller
	// will continue passing in ".DBF" as the extension.
	if (!extension.empty())
	{
		size_t t_dbf_pos = table_name.rfind(extension);
		if (t_dbf_pos != std::string::npos)
		{
			table_name.replace(t_dbf_pos, extension.length(), "");
		}
	}
	if (func_name.find("LOOKUP") != std::string::npos)
	{
		// Look for a column name following table name
		t_pos = table_name.rfind(".");
		if (t_pos != std::string::npos)
		{
			t_lookup_idx = table_name.substr(t_pos+1, table_name.length());
			table_name = table_name.substr(0, t_pos);
		}

		if ((func_name.compare("RLOOKUP") == 0) || (func_name.compare("ILOOKUP") == 0))
		{
			if (t_lookup_idx.empty())
			{
				t_err_msg = func_name + " always returns FALSE if no index key is specified. See help for " +
					func_name;
				comp_sntx_err((char*)t_err_msg.c_str());
			}
		}
	}

	TTableDef* t_table = FTablesList->GetTable(table_name);

	if (t_table)
	{
		// For LOOKUP, ILOOKUP and RLOOKUP, it is critical to get valid index
		// key names.
		// Otherwise, the Engine could cough up an access violation.
		// To keep things nice and complicated, however, LOOKUP may be called
		// without an index key.
		if ((func_name.find("LOOKUP") != std::string::npos) &&
			(!t_lookup_idx.empty()))
		{
			if (!find_index_name(t_table, t_lookup_idx))
			{
				if (func_name.compare("LOOKUP") == 0)
				{
					t_err_msg  = "Index name " + t_lookup_idx + " does not exist. ";
					t_err_msg += "Performance will be improved by creating this index on table " + table_name;
					comp_sntx_warn((char*)t_err_msg.c_str());
				}
				else
				{
					t_err_msg  = "Index name " + t_lookup_idx + " is not valid. ";
					t_err_msg += " Open table "+ table_name +" in Table form to find correct index names.";
					comp_sntx_err((char*)t_err_msg.c_str());
				}
			}
		}

	} // if (t_table)
	else
	{
		t_err_msg = "Table " + table_name + " not found in this metafile";
		comp_sntx_err((char*)t_err_msg.c_str());
	}
} // find_table
//---------------------------------------------------------------------------
bool TEditCompiler::find_index_name(TTableDef* table, std::string compare_idx_name)
{
	bool t_found_it = false;
	int t_idx_count = table->GetIndexCount();
	std::string t_idx_name;
	size_t t_start_pos, t_end_pos;

	// When a table is indexed on a column of data, that data may not be
	// unique within the column. SQL databases like to see data sorted in
	// a predictable fashion, if possible. So we post-fix the PKey value
	// to each index to ensure this reliable sorting.
	uppercase(compare_idx_name);
	std::string t_idx_with_pkey = compare_idx_name + std::string("], [PKEY");

	if (t_idx_count > 0)
	{
		for (int t_idx = 0; t_idx < t_idx_count; t_idx++)
		{
			t_idx_name = table->GetIndex(t_idx);
			// Index naming convention is to use the name of the table,
			// prefaced with lowercase 'x', underscore, then "([" followed
			// by the indexed field name... which will coincide with the
			// index name that was used in the old EMF version of EditWriter.
			// For example:
			// CREATE INDEX xAJC6STGP_AJCSITGP ON AJC6STGP([AJCSITGP])
			//
			t_start_pos = t_idx_name.find("(["); // 43
			t_end_pos = t_idx_name.find("])");   // 53
			if ((t_start_pos != std::string::npos) &&
				(t_end_pos != std::string::npos) &&
				(t_start_pos +2 < t_end_pos))
			{
				t_start_pos += 2;
				t_idx_name = t_idx_name.substr(t_start_pos, t_end_pos-t_start_pos);

				uppercase(t_idx_name);
				if ((t_idx_name.compare(compare_idx_name) == 0) ||
				    (t_idx_name.compare(t_idx_with_pkey) == 0))
				{
					t_found_it = true;
					break;
				}
			}
			/* Decided to make this a warning, to explain that performance would
			 * be improved if the author would add an index to the table
			else
			{
				FEngineErrorsPump->AddError(
					"TEditCompiler", "find_table", "Engine programmer issue: failure to parse index def for name");
				throw std::runtime_error("Something bad has happened");
			}
			*/
		}
	}

	return t_found_it;

} // find_index_name
//---------------------------------------------------------------------------
TTableDef* TEditCompiler::find_sql_table(std::string table_name)
{
	TTableDef* t_table = FTablesList->GetTable(table_name);

	if (!t_table)
	{
		std::string t_err_msg = "Table " + table_name + " not found in this metafile";
		comp_sntx_err((char*)t_err_msg.c_str());
	}

	return t_table;

}  // find_sql_table
//---------------------------------------------------------------------------
void TEditCompiler::find_sql_fields(TTableDef* table, std::string fields_list)
{
	std::string t_fields_list,
				t_fld_name,
				t_err_msg,
				t_table_name = table->GetTableName();

	TFieldDef* t_fld;
	// Something parsable, with blank spaces squeezed out
	t_fields_list.assign(stringReplace(fields_list," ",""));
	size_t t_pos = t_fields_list.find_first_of(",");
	while (t_pos != std::string::npos)
	{
		t_fld_name = t_fields_list.substr(0, t_pos);

		t_fld = table->GetColumn(t_fld_name.c_str());
		if (!t_fld)
		{
			t_err_msg = std::string("Table \"") + t_table_name +
						std::string("\" has no column named \"") +
						t_fld_name + "\"";
			comp_sntx_err((char*)t_err_msg.c_str());
		}
		else if (!find_index_name(table, t_fld_name))
		{
			t_err_msg  = std::string("Field \" ") + t_fld_name + std::string("\" is not indexed. ");
			t_err_msg += std::string("Performance will be improved by creating this index on table ") + t_table_name;
			comp_sntx_warn((char*)t_err_msg.c_str());
		}

		t_fields_list = t_fields_list.substr(t_pos+1, t_fields_list.length());

		t_pos = t_fields_list.find_first_of(",");
	}

	// Evaluate up the last (or only) field in the list
	t_fld_name = t_fields_list;

	t_fld = table->GetColumn(t_fld_name.c_str());
	if (!t_fld)
	{
		t_err_msg = std::string("Table \"") + t_table_name +
					std::string("\" has no column named \"") +
					t_fld_name + "\"";
		comp_sntx_err((char*)t_err_msg.c_str());
	}
	else if (!find_index_name(table, t_fld_name))
	{
		t_err_msg  = std::string("Field \" ") + t_fld_name + std::string("\" is not indexed. ");
		t_err_msg += std::string("Performance will be improved by creating this index on table ") + t_table_name;
		comp_sntx_warn((char*)t_err_msg.c_str());
	}

} // find_sql_fields
//---------------------------------------------------------------------------

bool TEditCompiler::check_date_type(std::string func_name,
	std::string param)
{
	bool t_retval = true;
	std::string t_field_name, t_data_type;
	std::string t_err_msg;
	// Make sure that if the parameter is a field, it's a D1 date field

// FIX ME: This will change when Susan and Kathleen T. decide upon the
//         other date formats they want the Engine to support. 12/06/2013-KB
	if (param.substr(0,1).compare("#") == 0)
	{
		std::string t_field_number = param.substr(1, param.length());
		std::string t_sql =
			"SELECT FieldName, DataType FROM EE_Fields WHERE FieldNumber = ";
		t_sql += t_field_number + ";";

		FQry->SetIsActive(false);
		FQry->SetSQLText(t_sql);
		try
		{
			FQry->SetIsActive(true);
			if (FQry->RowCount() == 0)
			{
				t_err_msg = "field number ";
				t_err_msg += t_field_number;
				t_err_msg += " is not in the data dictionary";
				comp_sntx_err((char*)t_err_msg.c_str());

				t_retval = false;
			}
			else
			{
				t_data_type.assign(FQry->GetFieldByName("DataType"));
// FIX ME: Switch statement for all the date types?
				t_retval = (t_data_type.compare("D1") == 0);
				if (!t_retval)
				{
					t_field_name.assign(FQry->GetFieldByName("FieldName"));

					t_err_msg = "Field \"";
					t_err_msg += t_field_name;
					t_err_msg += "\" must be data type D1 for call to " + func_name;
					comp_sntx_err((char*)(t_err_msg.c_str()));
				}
			}

			FQry->SetIsActive(false);

		}
		catch (...)
		{
			FQry->SetIsActive(false);
			std::string t_err = "failure to open query";
			FEngineErrorsPump->AddError("TEditCompiler", "check_date_type", t_err);
			t_err = "SQL: " + t_sql;
			FEngineErrorsPump->AddError("TEditCompiler", "check_date_type", t_err);
			throw;
		}
    } // if the parameter is a field (not a variable or a string literal)
	return t_retval;
}
//---------------------------------------------------------------------------
void TEditCompiler::no_such_variable(int param_num,
    std::string func_name, std::string var_name)
{
	std::string t_err = func_name +
		" error: Parameter #"+ intToString(param_num)+
		" references undeclared variable \"" +
		var_name + "\"";
	comp_sntx_err((char*)t_err.c_str());
}
//---------------------------------------------------------------------------
void TEditCompiler::walk_code_block(char token, int start_block,
    int& IF_end_block)
{
    /*  Called by parse_* functions for CONSTRUCTs:
     *  parse_if(), parse_else(), parse_while() and parse_for()
     *  We're walking the block of executable code in FCompiledCode
     *  to find out how many bytes it consumes. Needed by the edits
     *  engine so it knows how far to skip ahead when the data fails
     *  the condition of the code block.
     */
    char t_jump_buf[5];
    int t_CompLineNo = (int)FCompLineNo;

    EEt_NestedConstruct* t_nest;

    std::vector<EEt_NestedConstruct*> t_constructs;

	int t_idx,      // index into t_constructs, points to last one added
		t_jump;     // the number we're calculating in this function

	size_t t_CompiledPtr = FCompiledPtr,
		   t_compiled_len = FCompiledCode.length();

    size_t t_save_line;
	std::string t_msg;

    try {
        t_nest = new EEt_NestedConstruct();
        if (t_nest)
        {
			try
			{
				t_nest->token = token;
				t_nest->construct_line_no = (int)FCompLineNo;
				t_nest->open_brace_line_no = 0;
				t_nest->close_brace_line_no = 0;
				t_constructs.push_back(t_nest);
			}
			catch (...)
			{
				delete t_nest;
				throw;
			}
		} // if (t_nest)

		// This is the start of the block of code to be executed
		// when CONSTRUCT condition is true
		t_CompiledPtr = FCompiledPtr;

		do
		{
			t_idx = (int)(t_constructs.size()) -1;
			if (t_idx < 0)
				break;

			// Find out if we are looking at a block or a single statement--
			get_next_token();

			if (FCurrentTokenType == CONSTRUCT)
			{
				// For each nested construct, add an entry to
				// t_constructs, defaulting its open-brace count
				// to zero.
				t_nest = new EEt_NestedConstruct();
				if (t_nest)
				{
					try
					{
						t_nest->token = FCurrentToken;
						t_nest->construct_line_no = (int)FCompLineNo;
						t_nest->open_brace_line_no = 0;
						t_nest->close_brace_line_no = 0;
						t_constructs.push_back( t_nest );
					}
					catch(...)
					{
						delete t_nest;
						throw;
					}
				} // if (t_nest)
				// FixEdits # 5131: If the FCurrentToken is FOR ('C'),
				// we are at the START of the for(;;) loop. Move past it!
				if (FCurrentToken == FOR)
                    parse_for_prep();
            }
            else if (FCurrentTokenType == BLOCK &&
                FCurrentToken == LBRACE)
            {
                // Don't remove this item from the std::vector
                // until we find the closing brace
                t_constructs[t_idx]->open_brace_line_no = (int)FCompLineNo;
            }
            else
            {
                bool t_pop = false;

                if (FCurrentTokenType == BLOCK &&
                    FCurrentToken == RBRACE)
                {
					if (t_constructs[t_idx]->open_brace_line_no == 0)
                    {
                        comp_sntx_err("} found without matching {");
                    }

                    t_constructs[t_idx]->close_brace_line_no = (int)FCompLineNo;

                    t_pop = true;
                }
                else if (FCurrentTokenType == DELIMITER &&
                         FCurrentToken == SCOLON)
                {
                    // If the current (nested) construct has an
                    // open brace, don't remove it until the
                    // closing brace (RBRACE) is found. Otherwise,
                    // a semi-colon marks the end of the block.
                    if (t_constructs[t_idx]->open_brace_line_no == 0)
                    {
                        t_pop = true;
                    }
                }
                // Keep popping these off the stack as long
                // as we were waiting for a semi-colon and
                // not an RBRACE to end the block.
                if (t_pop)
                {
                    bool t_else_popped = false;
                    int t_jdx = (int)t_idx;

                    while (t_jdx >= 0)
                    {
                        t_nest = t_constructs[t_jdx];
                        if (t_nest->token == IF)
                        {
                            // Have we popped an ELSE construct?
                            // If so, we can pop the IF now, too
                            if (t_else_popped)
                            {
                                // If ELSE is popped, then we already
								// processed the IF, with or without
                                // braces. It's safe to pop it now.
                                // But reset the flag to false in case
                                // we're in a deeply nested series of
                                // CONSTRUCTs.
                                t_else_popped = false;

                                // Before we pop this IF, make sure
                                // it had proper brace-balancing
                                if (t_nest->open_brace_line_no > 0)
                                {
                                    if (t_nest->close_brace_line_no == 0)
                                    {
                                        t_save_line = FCompLineNo;
                                        FCompLineNo = t_nest->open_brace_line_no;
                                        t_msg = "unresolved open brace (hint: use Ctrl+B to find matched braces... see Edit Logic menu)";
                                        comp_sntx_err((char*)t_msg.c_str());
                                        FCompLineNo = t_save_line;
                                    }
                                }

                                if (t_nest->close_brace_line_no > 0)
                                {
                                    if (t_nest->open_brace_line_no == 0)
                                    {

                                        t_save_line = FCompLineNo;
                                        FCompLineNo = t_nest->close_brace_line_no;
                                        t_msg = "close brace without open brace (hint: use Ctrl+B to find matched braces... see Edit Logic menu)";
                                        comp_sntx_err((char*)t_msg.c_str());
                                        FCompLineNo = t_save_line;
                                    }
                                }

                                // std::vector::pop_back invokes the element's destructor
                                //delete t_nest;
                                t_constructs.pop_back();
                            }
                            // Are we on the first nested CONSTRUCT?
							else if (t_jdx == 0)
                            {
                                if ((t_nest->open_brace_line_no > 0) &&
                                    (t_nest->close_brace_line_no == 0))
                                {
                                    // Get out of this loop... we have to
                                    // keep processing until we find its
                                    // closing brace.
                                    break;
                                }
                                else
                                {
                                    // std::vector::pop_back invokes the element's destructor
                                    //delete t_nest;
                                    t_constructs.pop_back();
                                }
                            }
                            else
                            {
                                // Now for the fun part... if we have finished
                                // an IF CONSTRUCT, we need to see if it is
                                // immediately followed by an ELSE... but
                                // only if we're still working with a NESTED IF;
                                // otherwise, we want to break out of here.

                                size_t t_marker = FCompiledPtr,
                                    t_line_no = FCompLineNo;

                                get_next_token();

                                if (FCurrentTokenType == CONSTRUCT &&
                                    FCurrentToken == ELSE)
                                {
                                    //***************************************
                                    // FixEdits 5268:
                                    // Let's look for a preceding IF.
                                    // And, if we find it, make sure that
                                    // it does not have mismatched braces
                                    if (t_constructs.size() == 0)
									{
                                        comp_sntx_err("ELSE without an IF");
                                    }
                                    else
                                    {
                                        bool t_found_if = false;
                                        for (size_t t_kdx = t_constructs.size() -1;
                                             t_kdx >= 0; --t_kdx)
                                        {
                                            if (t_constructs[t_kdx]->token == IF)
                                            {
                                                if ((t_constructs[t_kdx]->open_brace_line_no > 0) &&
                                                    (t_constructs[t_kdx]->close_brace_line_no == 0))
                                                {
                                                    comp_sntx_err("mismatched braces; expected '}' before ELSE");
                                                }
                                                t_found_if = true;
                                                break;
                                            }
                                        }
                                        if (!t_found_if)
                                        {
                                            comp_sntx_err("ELSE without an IF");
                                        }
                                    }
                                    //***************************************

                                    EEt_NestedConstruct* t_else_nest = new EEt_NestedConstruct();
                                    if (t_else_nest)
                                    {
                                        try {
                                            t_else_nest->token = FCurrentToken;
                                            t_else_nest->construct_line_no = (int)FCompLineNo;
                                            t_else_nest->open_brace_line_no = 0;
                                            t_else_nest->close_brace_line_no = 0;
                                            t_constructs.push_back( t_else_nest );

                                            // Get out of here and process the
                                            // ELSE statement
											break;

                                        } catch(...) {
                                            delete t_else_nest;
                                            throw;
                                        }
                                    }
                                }
                                else
                                {
                                    // Put the FCompiledPtr back; we may be
                                    // ready to break out of the controlling loop.
                                    FCompiledPtr = t_marker;
                                    FCompLineNo  = t_line_no;

                                    // Now the usual processing...
                                    if ((t_nest->open_brace_line_no > 0) &&
                                        (t_nest->close_brace_line_no == 0))
                                    {
                                        // Get out of this loop... we have to
                                        // keep processing until we find its
                                        // closing brace.
                                        break;
                                    }
                                    else
                                    {
                                        // std::vector::pop_back invokes the element's destructor
                                        //delete t_nest;
                                        t_constructs.pop_back();
                                    }
                                }
                            }
                        }
                        else if (t_nest->token == ELSE)
                        {
                            if ((t_nest->open_brace_line_no > 0) &&
                                (t_nest->close_brace_line_no == 0))
                            {
                                // Get out of this loop... we have to
								// keep processing until we find its
                                // closing brace.
                                break;
                            }
                            else
                            {
                                // Set the flag to tell this loop
                                // controller to get rid of the
                                // preceding IF (even if other CONSTRUCTs
                                // lie between it and this ELSE)
                                t_else_popped = true;
                                // std::vector::pop_back invokes the element's destructor
                                //delete t_nest;
                                t_constructs.pop_back();
                            }
                        }
                        else
                        {
							// For all others, look to see if we're
                            // in a brace-bound section of code
                            if ((t_nest->open_brace_line_no > 0) &&
                                (t_nest->close_brace_line_no == 0))
                            {
                                // Get out of this loop... we have to
                                // keep processing until we find its
                                // closing brace.
                                break;
                            }
                            else
                            {
                                // If this nested CONSTRUCT has no braces,
                                // it's done... delete it.

                                // std::vector::pop_back invokes the element's destructor
                                //delete t_nest;
                                t_constructs.pop_back();
                            }
                        }
                        // decrement the counter
						t_jdx--;
					} // while (t_jdx >= 0)

				} // if (t_pop)
			}
		} while (FCompiledPtr < t_compiled_len);

		for (t_idx = 0; t_idx < (int)t_constructs.size(); t_idx++)
		{
			// Look for open brace without matching closing brace
			if (t_constructs[t_idx]->open_brace_line_no > 0)
			{
				if (t_constructs[t_idx]->close_brace_line_no == 0)
				{
					t_save_line = FCompLineNo;
					FCompLineNo = t_constructs[t_idx]->open_brace_line_no;
					t_msg = "unresolved open brace (hint: use Ctrl+B to find matched braces... see Edit Logic menu)";
					comp_sntx_err((char*)t_msg.c_str());
					FCompLineNo = t_save_line;
				}
			}

			// Look for closing brace without matching opening brace
			if (t_constructs[t_idx]->close_brace_line_no > 0)
			{
				if (t_constructs[t_idx]->open_brace_line_no == 0)
				{
					t_save_line = (int)FCompLineNo;
					FCompLineNo = t_constructs[t_idx]->open_brace_line_no;
					t_msg = "close brace without open brace (hint: use Ctrl+B to find matched braces... see Edit Logic menu)";
					comp_sntx_err((char*)t_msg.c_str());
					FCompLineNo = t_save_line;
				}
			}
	   } // for...
	} catch(std::runtime_error& e) {

		while (!t_constructs.empty())
		{
			// std::vector::pop_back invokes the element's destructor
			t_constructs.pop_back();
		}
		FEngineErrorsPump->AddError("TEditCompiler", "walk_code_block", e.what());
		throw;
	}

	while (!t_constructs.empty())
	{
		// std::vector::pop_back invokes the element's destructor
		t_constructs.pop_back();
	}

	// Calculate the size of the CONSTRUCT block, and write it
	// to the segment of FCompiledCode preceding the CONSTRUCT token
    t_jump = (int)(FCompiledPtr - t_CompiledPtr);
    sprintf(t_jump_buf, "%4i", t_jump);

    FCompiledCode = FCompiledCode.substr(0,start_block) +
                    (std::string)t_jump_buf +
                    FCompiledCode.substr(start_block +4);

    if ((token == IF) ||
        (token == ELSE))
    {
        // FixEdits 5268: Return the end-of-block marker to parse_if,
        // or parse_else
        // and through those to parse_block.
        get_next_token();
        IF_end_block = (int)FCompiledPtr;
    }

    // Move the pointer back to the start of the block of
    // executable code so that any nested IF, WHILE, FOR
    // clauses can be parsed.
    FCompiledPtr = t_CompiledPtr;
    FCompLineNo  = t_CompLineNo;
} // walk_code_block()
//---------------------------------------------------------------------------
void TEditCompiler::walk_expr(std::string the_code)
{
    /*   Called by parse_if() and parse_while()
     *   Walk the code between the parentheses of if () or while ()
     *   and evaluate for function calls.
     */

    // FixEdits # 5014, preventing reference to APPFIELDs outside a function call
    bool t_isFunctionCall = false;

    int t_open_parens = 0, // the first paren after 'if' or 'while'
		 t_func_parens = 0; // counting parens inside a function call

	size_t t_code_len = FCompiledCode.size();

    std::string t_var_name,
           t_err_msg;

    char t_func_code;

    TVariableDef* t_var = NULL;

    while (FCompiledPtr < t_code_len)
    {
        // When we come in here, the FCompiledCode[FCompiledPtr] is pointing
        // to the first token following the "if(" or "while(". Start evaluating
        // from there, and move on to the next token at the bottom of this loop.
        //get_next_token();

        if (FCurrentTokenType == DELIMITER)
        {
            // token is '(' or ')'

            if (FCurrentToken == LPAREN)
            {
                t_open_parens++;
            }

            else if (FCurrentToken == RPAREN)
            {
                t_open_parens--;

                // If we found the outer closing parenthesis, we're done. Get out!
                if (t_open_parens == 0)
                    break;

                // FixEdits 5014: Watch out for references to
                // APPFIELDs outside a function call
                if (t_isFunctionCall)
                {
                    if (--t_func_parens == 0)
                        t_isFunctionCall = false;
                }

                /* Look for nested conditions */

                /* find out what's next up */
                get_next_token();

                /* If the next token is a closing parenthesis,
                   all is well. But if it isn't... */
                if (! (FCurrentTokenType == DELIMITER &&
                       FCurrentToken == RPAREN))
                {
                    /* The only legal token here is one of the OPERATORs */
                    if (! (FCurrentTokenType == OPERATOR))
                    {
                        comp_sntx_err("operator expected");
                    }
                    else
                    {
                        /* FixEdits # 5285:
                         * However, just because it's an OPERATOR doesn't
                         * mean we're out of the woods. There is a tendency
                         * to repeat the IF or WHILE unnecessarily, and that
                         * leads to a syntax error being thrown by the Engine.
                         * Let's try to catch that here.
                         */
                        int t_save_ptr = (int)FCompiledPtr;
                        TTokenType t_save_type = FCurrentTokenType;
                        char t_save_token = FCurrentToken;

                        try
                        {
                            get_next_token();

                            if (FCurrentTokenType == CONSTRUCT &&
                                ((FCurrentToken == IF) ||
                                 (FCurrentToken == WHILE)))
                            {
                                if (FCurrentToken == IF)
                                    comp_sntx_err("improper nesting of IF");
                                else
                                    comp_sntx_err("improper nesting of WHILE");
                            }
                        }
						catch(std::runtime_error& e)
                        {
							FEngineErrorsPump->AddError(
								"TEditCompiler", "walk_expr", e.what());
							// Put everything back the way it was
							FCompiledPtr = t_save_ptr;
							FCurrentTokenType = t_save_type;
							FCurrentToken = t_save_token;
							throw;
						}
						// Put everything back the way it was
						FCompiledPtr = t_save_ptr;
						FCurrentTokenType = t_save_type;
						FCurrentToken = t_save_token;

                    }

                }

                // since we already invoked get_next_token(), loop
                // back to the top for further processing.
                continue;

            }
        }

        else if (FCurrentTokenType == FUNCTION &&
                 FCurrentToken == LIBFUNC)
        {
            // token is 'M'

            t_isFunctionCall = true;
            t_func_parens++;

            /*  get_next_token will position you on the opening parenthesis
             *  of the function call. The identifier for the specific function
             *  will be one character to the left... so back up and grab it.
             */
            t_func_code = FCompiledCode[FCompiledPtr-1];

            // skip over the opening parenthesis to the function call
            get_next_token();

            if (FCurrentTokenType == DELIMITER && FCurrentToken == LPAREN)
            {
                t_open_parens++;
                t_func_parens++;
            }
            else
                comp_sntx_err("opening parenthesis expected following function name");


            check_params(t_func_code, FCompiledCode.substr(FCompiledPtr, t_code_len));
        }

        else if (FCurrentTokenType == IDENTIFIER &&
                 FCurrentToken == VAR)
        {
            // token is 'N'
            /*  Catching undeclared variables in expressions:
                int t_bar;
                while (foo)
                {
                    t_bar = t_bar +1;
                }
             */

            // get_next_token() will have assigned the name of the
            // variable to FCurrentVarName.

            // t_var_idx is passed by reference, receives
            // a non-zero value if the variable has been
            // declared
            t_var = find_variable(FCurrentVarName);

            if (!t_var)
            {
                t_err_msg = "variable \""+FCurrentVarName+"\" must be declared";
                comp_sntx_err((char*)(t_err_msg.c_str()));
			}

		} // check for variables ('N')

		else if (FCurrentTokenType == OPERATOR)
		{
			if (FCurrentToken == ASSIGN)
			{
				if (!t_var)
				{
					t_err_msg = "assignment operator without a variable (did you mean \"==\"?)";
					comp_sntx_err((char*)(t_err_msg.c_str()));
				}
				else
				{
					// This function is called from parse_if and parse_while
					// to evaluate the control expression... I think. BIG risks
					// here, but let's see what happens
					if (t_open_parens > 0) // we're in the control expression
					{
						t_err_msg = "possible incorrect assignment (did you mean \"==\"?)";
						comp_sntx_warn((char*)(t_err_msg.c_str()));
					}
				}
			}
		} // if (FCurrentTokenType == OPERATOR)

		else if (FCurrentTokenType == APPFIELD)
		{
			// FixEdits # 5014: protect against a reference to a field
			// that is not made from within a function call.
			if (!t_isFunctionCall)
			{
				t_err_msg = "syntax error: fields must be accessed through function calls";
				comp_sntx_err((char*)(t_err_msg.c_str()));
			}

        }

        // See note at the top of this loop; now it's time to advance to the
        // next token.
        get_next_token();

    } // while (t_open_parens > 0)

} // walk_expr()
//---------------------------------------------------------------------------

#endif

