//---------------------------------------------------------------------------

#ifndef EditErrorsH
#define EditErrorsH
//---------------------------------------------------------------------------
#include <string>
#include <map>
#include <vector>

#include "edit_external.h"


class TSQLDB;
class TFastSQLQuery;
class TEngineErrorsPump;
class TEdit;
class TLayout;
class TLibFuncs;

class TErrorObject
{
private:
	std::string FErrorType;
	std::string FErrorMessage;
	std::string FFunctionName;
public:
	TErrorObject(std::string error_type, std::string error_message, std::string func_name);

	std::string GetErrorType() { return FErrorType; };
	std::string GetErrorMessage() { return FErrorMessage; };
	std::string GetFunctioName() { return FFunctionName; };

};

class TEditErrors
{
private:
	TSQLDB* FSQLDB;
	TFastSQLQuery* FMessageQry;
	TEngineErrorsPump* FEngineErrorsPump;
	void* FCallbackOwner;
	CallbackRunEdits FCallbackFunction;

	int FError_Text_Edit_PKey;
	std::string FError_Text_Message;

	// Store the pointer to TLibFuncs so that we can interact
	// At a minimum, we need access to TLibFuncs::GetDateErrorCode()
	TLibFuncs* FLibFuncs;

	TLayout* FLayout;

	int FAltMessageNumber;
	int FSuppressWarnings;

	std::map< int, std::vector<TErrorObject*> > FErrorsMap;

	void clear_errors_map();
	std::string expand_msg_macros(std::string raw_message, TEdit* running_edit);
	void add_error_to_map(int edit_pkey, TErrorObject* error_object);

	bool create_message_query();

public:
	TEditErrors(TSQLDB* database, TEngineErrorsPump* pump);
	~TEditErrors();

	void SetLibFuncs(TLibFuncs* lib_funcs);

	int FailedEditsCount();
	void Initialize(void* owner, void* callback_func, TLayout* layout, int suppress_warnings);
	void ReportDefaultError(TEdit* running_edit, std::string error_type);
	void ReportSaveErrorText(TEdit* running_edit, int msg_num);
	void ReportSetWarning(TEdit* running_edit, int msg_num);
	void ReportSaveWarningText(TEdit* running_edit, int msg_num);
	void SetErrorText(TEdit* running_edit, std::string error_message);
	void SetSaveText(TEdit* running_edit, std::string save_message);
	void SetAltMessageNumber(int alt_message_number);
};

#endif
