//---------------------------------------------------------------------------
#include <cstring> // strncpy

// For the Borland compilers, use #pragma hdrstop for precompiled headers (pch)
#ifdef BCB
	#pragma hdrstop
#endif

#include "EngineErrorsPump.h"
//---------------------------------------------------------------------------
TEngineErrorsPump::TEngineErrorsPump()
{
	// constructor
}
//---------------------------------------------------------------------------

TEngineErrorsPump::~TEngineErrorsPump()
{
	ClearPump();
} // destructor
//---------------------------------------------------------------------------

int TEngineErrorsPump::ClearPump()
{
	// This function is called between runs of API functions (from InitCheck
	// in EditCAPI) to clear out unclaimed messages before starting another
	// call.
	for (unsigned int t_idx = 0; t_idx < FPump.size(); t_idx++)
	{
		delete FPump[t_idx];
	}

	FPump.clear();

	return 0; // all is well
} // ClearPump
//---------------------------------------------------------------------------

int TEngineErrorsPump::AddError(TEngineError* error)
{
	// I want to store these errors in the chronological order in which they
	// were thrown, so we'll insert to the end of vector FPump instead of
	// pushing onto the stack.
	FPump.insert(FPump.end(), error);
	// size_t t_how_many = FPump.size();
	return 0; // all is well
} // AddError
//---------------------------------------------------------------------------
int TEngineErrorsPump::AddError(std::string class_name,
	std::string function_name, std::string error_message)
{
	TEngineError* t_error = new TEngineError();

	t_error->EngineClassName.assign(class_name);
	t_error->FunctionName.assign(function_name);
	t_error->ErrorMessage.assign(error_message);

	return AddError(t_error);

} // overloaded AddError
//---------------------------------------------------------------------------
int TEngineErrorsPump::GetErrorsCount(int* how_many_errors)
{
	*how_many_errors = (int)FPump.size();

	return 0; // all is well
} // GetErrorsCount
//---------------------------------------------------------------------------
int TEngineErrorsPump::GetErrorByIndex(int idx, char* error_message,
	int message_length)
{
	std::string t_error_message;
	// NOTE: This class assumes the idx is part of a 0-based counter.
	// But the TSMF class will assume the programmer is thinking 1-based
	// (and API documentation will say that's what it is), so the idx
	// will come to TSMF as 1 for the first entry, and to TEngineErrorsPump
	// as 0.
	if ((idx >= 0) && (idx < (int)FPump.size()))
	{
		TEngineError* t_error = FPump[idx];
		t_error_message  = t_error->EngineClassName + "::" + t_error->FunctionName;
		t_error_message += "" + t_error->ErrorMessage;

		std::strncpy(error_message, (char*)t_error_message.c_str(), message_length);
	}
	else
	{
		// Handling an error while talking to the error pump... if the index
		// is out of range, put THAT information in the error_message that goes
		// back to the caller.
	        std::strncpy(error_message, "TEngineErrorsPump::GetErrorByIndex - index out of range",
			message_length);
	}
	return 0;
} // GetErrorByIndex
//---------------------------------------------------------------------------


