#include <stdexcept>
#include <cstdlib>  // std::atoi
#include <cstdlib>   // std::snprintf
#include <cstring>  // memset
//---------------------------------------------------------------------------

#include "Ew_AccessController.h"
#include "SQLDB.h"
#include "SQLQuery.h"
#include "Utility.h"
#include "EngineErrorsPump.h"

//---------------------------------------------------------------------------
#ifdef __SQLEW__

TEw_AccessController::TEw_AccessController(TEngineErrorsPump* pump, TSQLDB* database)
{
	FEngineErrorsPump = pump;
	FSQLDB = database;
	FAgenciesQry = NULL;
} // constructor
//---------------------------------------------------------------------------
TEw_AccessController::~TEw_AccessController()
{
	if (FAgenciesQry)
	{
		delete FAgenciesQry;
		FAgenciesQry = NULL;
	}
	FLoginMap.clear();
	FLoginList.clear();
}  // destructor
//---------------------------------------------------------------------------
bool TEw_AccessController::is_logged_in(std::string agency_code)
{
	std::string t_compare;

	for (unsigned int t_idx = 0; t_idx < FLoginList.size(); t_idx++)
	{
		t_compare.assign(FLoginList[t_idx]);

		if (agency_code.compare(t_compare) == 0)
		{
			return true;
		}
	}
	return false;
} // is_logged_in (std::string)
//---------------------------------------------------------------------------
bool TEw_AccessController::IsLoggedIn( int agency_pkey )
{
	bool t_result = false;

	if (agency_pkey == 1)
	{
		// EE_Agencies.PKey =1 is the "NONE" agency, and is always valid.
		t_result = true;
	}
	else if (FLoginMap.find(agency_pkey) != FLoginMap.end())
	{
		t_result = true;
	}
	else if (is_logged_in("system"))
	{
		t_result = true;
	}

	return t_result;

} // IsLoggedIn (int)
//---------------------------------------------------------------------------
int TEw_AccessController::get_logged_in_pkey()
{
	int t_pkey = 1; // Default to the "NONE" agency

	std::string t_code;
	// Iterate the FLoginMap, looking for something OTHER THAN "NONE"
	// to return to the caller. The first good value will be used,
	// even if there are multiple logins.
	for (std::map<int, std::string>::iterator it = FLoginMap.begin();
		 it != FLoginMap.end(); it++)
	{
		t_code.assign((*it).second);
		if (t_code.compare("NONE") != 0)
		{
			t_pkey = (*it).first;
			break;
		}
	}
	return t_pkey;
} // get_logged_in_pkey
//---------------------------------------------------------------------------
bool TEw_AccessController::is_uber_user(std::string agency_code,
	std::string password)
{
	char t_uber_pw[10];

	t_uber_pw[0]  = 'k';
	t_uber_pw[1]  = 'a';
	t_uber_pw[2]  = 't';
	t_uber_pw[3]  = 's';
	t_uber_pw[4]  = 'u';
	t_uber_pw[5]  = 'e';
	t_uber_pw[6]  = '\x0';

	if (agency_code.compare("system") == 0)
	{
		if (password.compare(t_uber_pw) == 0)
		{
			return true;
		}
	}

	return false;

} // is_uber_user
//---------------------------------------------------------------------------
bool TEw_AccessController::is_valid_login(std::string& agency_code,
	std::string password, int* agency_pkey)
{
	bool t_result = false;
	// default the agency_pkey to a bogus value
	*agency_pkey = -1;

	if ((agency_code.length() > 0) && (password.length() > 0))
	{
		if (!FAgenciesQry)
		{
			FAgenciesQry = new TSQLQuery(FSQLDB);
		}

		std::string t_sql = "SELECT PKey, Password, AdminCode FROM EE_Agencies ";
		t_sql += "WHERE AdminCode = '" + stringReplace(agency_code,"'", "''") + "';";

		FAgenciesQry->SetSQLText(t_sql.c_str());
		try
		{
			try
			{
				FAgenciesQry->SetIsActive(true);

				if (!FAgenciesQry->Eof())
				{
					// This may come back to bite us, but the password
					// scheme dating back to EW2 padded the password
					// to ten characters, and the padding became part
					// of the munging. So we had best follow suit.

					// Translate the plain-text password into munged
					char t_munged_pw[11];
					memset(t_munged_pw, '\0', 11);
					// Munge everything except the NULL terminator
					MungedPassword(password, t_munged_pw, 10);

					// Retrieve the password stored in the metafile
					std::string t_stored_pw = FAgenciesQry->GetFieldByName("Password");

					if (t_stored_pw.compare(t_munged_pw) == 0)
					{
						// Update the agency PKey with the associated password.
						*agency_pkey = FAgenciesQry->GetIntFieldByName("PKey");
						agency_code.assign(FAgenciesQry->GetFieldByName("AdminCode"));
						t_result = true;
					}
				}

			}
			catch (std::runtime_error& e)
			{
				FAgenciesQry->SetIsActive(false);
				throw std::runtime_error(e.what());
			}

			FAgenciesQry->SetIsActive(false);
		}
		catch (std::runtime_error& e)
		{
			FEngineErrorsPump->AddError("TEw_AccessController", "is_valid_login",
				e.what());
		}
	} // test for empty parameters
	return t_result;
} // is_valid_login
//---------------------------------------------------------------------------
void TEw_AccessController::MungedPassword(std::string rawpw,
	char* cookedpw, unsigned int cooked_len)
{
	// This is a C++ version of what originated as FoxPro code
	// used by EditWriter v2 (DOS). We have to be backward
	// compatible, so that users can log into their upgraded
	// metafiles. And this isn't a terrible algorithm.

	std::string t_temp;
	// The FoxPro version forced passwords to all uppercase;
	// the pre-uppercase characters falling between 48 and 122 (inclusive).
	// ('0' through 'z').
	t_temp.assign(rawpw);
	uppercase(t_temp);
	unsigned int t_len = (unsigned int)t_temp.length();

	int t_code[] = {137,
					 46,
					127,
					119,
					212,
					142,
					189,
					125,
					201,
					101};

	char k;
	for (unsigned int i=0, j=0; i < cooked_len; i++)
	{
		t_code[i] = t_code[i] - t_len + j;
		// As long as we have not walked off the end of t_temp,
		// use the character at the i'th position:
		if (i < t_len)
		{
			k = t_temp.c_str()[i];
		}
		else
		{
			// Once t_temp.length() has been hit, use a blank space:
			k = ' '; // 32
		}
		j = k - 32;
		cookedpw[i] = (char)((k + t_code[i]) % 256);
	}
} // munged_password
//---------------------------------------------------------------------------
int TEw_AccessController::Login(std::string agency_code, std::string password)
{
	int t_result = 0;
	// First, make sure the parameters don't have leading or trailing
	// spaces.
	std::string t_agency_code = trim(agency_code);
	std::string t_password = trim(password);

	int t_agency_pkey;

	if (t_agency_code.length() == 0)
	{
		// This is legal... gives them access to non-agency owned data.
		// Establish the agency_code as "NONE", which matches
		// AgencyPKey 1, a required value.
		FLoginList.push_back("NONE");
		// We also need to push this onto the FLoginMap
		FLoginMap[1] = "NONE";
		return t_result;
	}

	// Next, determine whether this login has already been resolved
	if (is_logged_in(t_agency_code))
	{
		// The FoxPro version forced passwords to all uppercase;
		// all characters falling between 48 and 122 (inclusive).
		// ('0' through 'z').
		uppercase(t_password);

		if (!is_valid_login(t_agency_code, t_password, &t_agency_pkey))
		{
			t_result = -1;
		}
	}
	else
	{
		if (is_uber_user(t_agency_code, t_password))
		{
			FLoginList.push_back(t_agency_code);
		}
		else
		{
			// The FoxPro version forced passwords to all uppercase;
			// all characters falling between 48 and 122 (inclusive).
			// ('0' through 'z').
			uppercase(t_password);

			if (is_valid_login(t_agency_code, t_password, &t_agency_pkey))
			{
				FLoginList.push_back(t_agency_code);
				FLoginMap[t_agency_pkey] = t_agency_code;
			}
			else
			{
				FEngineErrorsPump->AddError("TEw_AccessController",
					"Login", "Agency code and/or password not valid");

				// The default login is "NONE"; let's get that set up
				// so that we have something in the LoginList.
				if (!is_logged_in("NONE"))
				{
					FLoginList.push_back("NONE");
				}
				t_result = -1;
			}
		}
	} // if (is_logged_in)
	return t_result;
} // Login
//---------------------------------------------------------------------------
int TEw_AccessController::GetLoginsCount(int* how_many_logins)
{
	*how_many_logins = (int)FLoginList.size();
	return 0;
}
//---------------------------------------------------------------------------
int TEw_AccessController::GetLoginByIndex( const int idx, int* agency_pkey,
	char* agency_code, const int agency_code_size )
{
	int t_result = 0;
	// default the agency_pkey
	*agency_pkey = 0;
	if (idx < (int)FLoginList.size())
	{
		strncpy(agency_code, FLoginList[idx].c_str(), agency_code_size);
		std::string t_test;
		std::map<int, std::string>::iterator it = FLoginMap.begin();
		while (it != FLoginMap.end())
		{
			t_test = (*it).second;
			if (t_test.compare(agency_code) == 0)
			{
				*agency_pkey = (*it).first;
				break;
			}

			it++;
		}
	}
	else
	{
		char t_buffer[500];
		std::snprintf(t_buffer, 500,
			"Index [%d] out of range; must be less than total logins. Call Ew_GetLoginsCount first",
			idx);
		FEngineErrorsPump->AddError("TEw_AccessController", "GetLoginByIndex",
			t_buffer);

		t_result = -1;
	}

	return t_result;

} // GetLoginByIndex
//---------------------------------------------------------------------------

int TEw_AccessController::GetLoggedInAgencyPKey( const int current_agency_pkey,
	int* use_this_agency_pkey )
{
	if (is_logged_in("system"))
	{
		// If the user is logged in under the super-user account,
		// then use the existing agency_pkey
		*use_this_agency_pkey = current_agency_pkey;
	}
	else if (IsLoggedIn(current_agency_pkey))
	{
		// If the object's current AgencyPKey is one for which the user
		// is logged in, go ahead and use it.
		*use_this_agency_pkey = current_agency_pkey;
	}
	else
	{
		// Otherwise, use the PKey for the agency used to login
		*use_this_agency_pkey = get_logged_in_pkey();
	}

	return 0;
} // HaveAccessRights
//---------------------------------------------------------------------------
// Support for Import Metafile: Provide temporary login to metafile as the
// super user, and remove it when finished with importing.
void TEw_AccessController::AddImportSuperUser(bool& was_super_user)
{
	// The caller is responsible for retaining the state of was_super_user
	// and passing it in to RemoveImportSuperUser to reset access rights
	// for normal usage.
	was_super_user = is_logged_in("system");
	if (!was_super_user)
	{
		FLoginList.push_back("system");
	}
} // AddImportSuperUser
//---------------------------------------------------------------------------
void TEw_AccessController::RemoveImportSuperUser(bool was_super_user)
{
	// If the user was not already logged in as the super user,
	// remove "system" from FLoginList
	if (!was_super_user)
	{
		std::string t_compare,
					t_system = "system";

		for (unsigned int t_idx = 0; t_idx < FLoginList.size(); t_idx++)
		{
			t_compare.assign(FLoginList[t_idx]);

			if (t_compare.compare(t_system) == 0)
			{
				FLoginList.erase(FLoginList.begin() + t_idx);
				break;
			}
		}

	}
} // RemoveImportSuperUser
//---------------------------------------------------------------------------


#endif // __SQLEW__

