//---------------------------------------------------------------------------
// NOTE: Tracking open DLLs is done outside instances of TSMF.
// See comments at top of header file for this unit.
#include "ExternalDLLTracker.h"
#include "Utility.h"

static size_t DLLMaxSize = 0;

//---------------------------------------------------------------------------
// This function will be called from DllMain/ShutDown (EditAPI.cpp)
// When the DLL is unloaded, we'll free all of the libraries
void ClearExternalDLLs()
{
	if (DLLMaxSize > 0)
	{
		for (size_t t_idx = 0; t_idx < DLLMaxSize; t_idx++)
		{
			EEt_ExternalDLL *t_ExternalDLL = FExternalDLLs[t_idx];

#ifdef EE_WINDOWS
			// release the DLL resource
			FreeLibrary(t_ExternalDLL->hLib);
#else
			dlclose(t_ExternalDLL->hLib);
#endif
			delete t_ExternalDLL;
		}

		FExternalDLLs.clear();
	}
} // ClearExternalDLLs
//-----------------------------------------------------------------------------
EEt_ExternalDLL* GetExternalDLL(std::string dll_name)
{
	EEt_ExternalDLL *t_ExternalDLL = NULL;

	if (DLLMaxSize > 0)
	{
		for (size_t t_idx = 0; t_idx < DLLMaxSize; t_idx++)
		{
			EEt_ExternalDLL *t_test = FExternalDLLs[t_idx];

			//if (posix_stricmp(t_test->DLLname.c_str(), dll_name.c_str()) == 0)
			if (t_test->DLLname.compare(dll_name) == 0)
			{
				t_ExternalDLL = t_test;
				break;
			}

		}

    }
	return t_ExternalDLL;
} // GetExternalDLL
//-----------------------------------------------------------------------------
#ifdef EE_WINDOWS
void AddExternalDLL(std::string dll_name, HINSTANCE hLib, void* func_ptr)
#else
void AddExternalDLL(std::string dll_name, void* hLib, void* func_ptr)
#endif
{
	EEt_ExternalDLL* t_ExternalDLL = new EEt_ExternalDLL();
	t_ExternalDLL->DLLname.assign(dll_name);
	t_ExternalDLL->hLib = hLib;
	t_ExternalDLL->EditsInterfaceFunc = func_ptr;

	FExternalDLLs.push_back(t_ExternalDLL);

	DLLMaxSize = FExternalDLLs.size();
} // AddExternalDLL
//-----------------------------------------------------------------------------

