#ifndef FASTSQLQUERYH
#define FASTSQLQUERYH
#include "sqlite3.h"

// TFastSQLQuery is a variation on TSQLQuery. It doesn't build the column lists
// and you cannot retrieve data by column name. You must be mindful of the
// column order in your query. Used by the RunEdits process to expedite processing.

// forward declaration
class TSQLDB;

class TFastSQLQuery
{
private:
	std::string FSQLText;
	std::string FLastErrorMessage;

	sqlite3_stmt* FStatement; // for sqlite3_prepare_v2, sqlite3_step
	int FColCount;
	//int FRowCount;
	bool FEof;

	bool FIsActive;

	bool FIsParameterized;

	// This will require testing... but if the sqlite3_prepare_v2 calls
	// must always be matched by sqlite3_finalize calls, could we not
	// keep track of them and run all the finalize calls when we destroy
	// the query object?
	int FPreparedCount;

	//std::string get_ucase_sql();

	bool reset_statement();
	bool open_query();
	bool close_query();
	bool finalize_query();

	void set_is_parameterized(bool is_parameterized);

	int get_prepared_count();
	void incr_prepared_count();
	void decr_prepared_count();
	void force_finalize_query();

public:
	TFastSQLQuery(TSQLDB* database);
	~TFastSQLQuery();

	//bool ExecSQL();
	bool FetchNext();
	// Return data as string
	char* GetFieldByNumber(int column_number);
	// For int fields, return data as integer
	int GetIntFieldByNumber(int column_number);

	TSQLDB* SQLDB;
	char* GetSQLText();
	void SetSQLText(std::string sql_text, bool check_parameterized = true);

	bool GetIsActive();
	bool SetIsActive(bool is_active);

	bool GetIsParameterized();

	bool Eof();
	int ColCount();
	//int RowCount();

	char* LastErrorMessage();

	// Support for parameterized queries
	bool PrepareParameterizedQuery();
	bool ResetParameterizedQuery();
	bool BindIntParameter(const int param_num, const int value);
	bool BindTextParameter(const int param_num, const char* value);
	bool OpenParameterizedQuery();
	bool ReleaseParameterizedQuery();

}; // class TFastSQLQuery
//---------------------------------------------------------------------------
#endif

