//---------------------------------------------------------------------------

#ifndef LayoutH
#define LayoutH
//---------------------------------------------------------------------------
/*   This class is for referencing Record Layouts for the Run Edits process.
 */

#include <string>
#include <map>
#include <vector>

// Forward declarations
class TSQLDB;
class TFastSQLQuery;
class TEngineErrorsPump;
class TLayout;

class TLayoutField
{
private:
	TLayout* FLayout;
	TSQLDB* FSQLDB;
	TFastSQLQuery* FSetupFieldQry;
	TEngineErrorsPump* FEngineErrorsPump;

	int FPKey;
	int FStartPos;
	int FFieldNumber;
	std::string FFieldName;
	std::string FDisplayName;
	int FFieldLength;
	std::string FDataType;
	// The value in FFailedSingleFieldEdit is set to false when the
	// TLayout::LoadData reassigns values to the layout fields,
	// and to true when the single-field edit fails
	bool FFailedSingleFieldEdit;
	// FNeverSkipImpty is an override of the edit or edit set
	// instructions to SKIP EMPTY. Used mostly to modify the
	// SKIP EMPTY instruction for multi-field edits involving
	// NAACCR "Over-ride" fields, which will frequently be
	// empty. This ensures that if the ONLY empty field is the
	// Over-ride, the edit will run, but if some other field is
	// also empty, the edit will NOT run.
	bool FNeverSkipEmpty;

	void setup_layout_field();
	bool create_setup_field_query();

public:
	TLayoutField(TLayout* layout, TSQLDB* database, TEngineErrorsPump* pump, int field_pkey, int start_pos);
	~TLayoutField();

	int GetPKey();
	int GetStartPos();
	int GetFieldNumber() ;

	char* GetFieldName();
	char* GetDisplayName();
	int SetDisplayName(const char* display_name);
	int GetFieldLength();
	char* GetDataType();
	char* GetValue();

	bool GetFailedSingleFieldEdit() { return FFailedSingleFieldEdit; }
	void SetFailedSingleFieldEdit(bool failed) { FFailedSingleFieldEdit = failed; }

	bool GetNeverSkipEmpty() { return FNeverSkipEmpty; }
	void SetNeverSkipEmpty(bool never_skip) { FNeverSkipEmpty = never_skip; }

};



class TLayout
{
private:
	TSQLDB* FSQLDB;
	TEngineErrorsPump* FEngineErrorsPump;
	std::string FLayoutName;
	std::string FLayoutTag;
	int FPKey;

	//std::string FData;
	char* FData;
	size_t FDataLength;

	// FFieldsByStartPos is merely a list of all the fields sorted
	// by starting position, where overlay blocks are further
	// sorted to list the field with the greater length before
	// the field(s) overlaid within it.
	std::vector<TLayoutField*> FFieldsByStartPos;
	std::map<int, TLayoutField*> FFieldsByNumber;


	void create_layout_fields();

public:
	TLayout(TSQLDB* database, TEngineErrorsPump* pump);
	~TLayout();

	// scratch variable used by TLayoutField for retrieving string elements
	char* LayoutFieldBuffer;
	size_t LayoutFieldBuffSize;
	void SetupLayoutFieldBuffer(size_t buff_size);

	int LoadLayout(const std::string layout_tag);
	int LoadData(const std::string data);
	int DisplayOneField(int field_number, int field_macro_order, void* owner, void* callback_func);
	int DisplayDataRecord( void* owner, void* callback_func );
	int DisplayFieldInMessagesAs( const int field_number, const std::string display_name );

	int NeverSkipEmptyField( const int field_number, const int never_skip_empty );

	int FieldDataByNumber(
		const int field_number, int* pkey,
		int* start_pos, int* field_length,
		char* field_name, const int field_name_len,
		char* display_name, const int display_name_len,
		char* data_type, const int data_type_len,
		char* value, const int value_len );

	TLayoutField* GetFieldByFieldNumber( int field_number );

	char* FillFieldValue(const int start_pos, const int field_length);


};
#endif
