//---------------------------------------------------------------------------

#include <memory>   // std::unique_ptr
#include <cstring> // strncpy

// For the Borland compilers, use #pragma hdrstop for precompiled headers (pch)
#ifdef BCB
	#include <cstdio>   // snprintf
	#pragma hdrstop
#else
	// Visual Studio 2012 uses _snprintf from stdio.h instead of the STL definition
	// Do a global search-and-replace looking for "std::snprintf" and
	// replacing with "_snprintf"
	#include <stdio.h>
#endif

#include "SMFTables.h"
#include "SQLDB.h"
#include "SQLQuery.h"

/*  This unit creates the SMF version of the metafile constituent objects control
 *  tables... that is, the ones with names prefixed with "EE_".
 *
 *  Calls into this class come from TEMFtoSMF::CreateSMFTriggers().
 */

//---------------------------------------------------------------------------

TSMFTables::TSMFTables(TSQLDB* database)
{
    FSQLDB = database;
    FQry = new TSQLQuery(FSQLDB);
} // constructor
//---------------------------------------------------------------------------
TSMFTables::~TSMFTables()
{
    if (FQry)
    {
        delete FQry;
        FQry = NULL;
    }
} // destructor
//---------------------------------------------------------------------------

bool TSMFTables::CreateSMFTables(const char* mf_structure_version)
{
	bool t_result = true;
	if (!create_table(tidControl, mf_structure_version))
    {
			  t_result = false;
    }

	if (t_result && !create_indexes(tidControl))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidAgencies))
    {
			  t_result = false;
    }

	if (t_result && !create_indexes(tidAgencies))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidEdits))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidEdits))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidEditSets))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidEditSets))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidFields))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidFields))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidLayouts))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidLayouts))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidMessages))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidMessages))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidLookupTables))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidLookupTables))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    // Now get the pivot tables
    if (t_result && !create_table(tidEditSetEditsPivot))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidEditSetEditsPivot))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidEditErrorsPivot))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidEditErrorsPivot))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidEditMessagesPivot))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidEditMessagesPivot))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidEditFieldsPivot))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidEditFieldsPivot))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidEditTablesPivot))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidEditTablesPivot))
    {
			  t_result = false;
    }
    ///////////////////////////////////////////////////////////////////////////
    if (t_result && !create_table(tidLayoutFieldsPivot))
    {
			  t_result = false;
    }

    if (t_result && !create_indexes(tidLayoutFieldsPivot))
    {
			  t_result = false;
    }

	return t_result;
} // CreateSMFTables
//---------------------------------------------------------------------------

bool TSMFTables::create_table(TSMFTableId TableId,
	const char* mf_structure_version /* = NULL */)
{
    bool t_result = true;
    switch (TableId)
    {
		 case tidControl:
			t_result = create_table_EE_Control(mf_structure_version);
            break;
         case tidAgencies:
            t_result = create_table_EE_Agencies();
            break;
         case tidEdits:
            t_result = create_table_EE_Edits();
            break;
         case tidEditSets:
            t_result = create_table_EE_EditSets();
            break;
         case tidFields:
            t_result = create_table_EE_Fields();
            break;
         case tidLayouts:
            t_result = create_table_EE_Layouts();
            break;
         case tidMessages:
            t_result = create_table_EE_Messages();
            break;
         case tidLookupTables:
            t_result = create_table_EE_LookupTables();
            break;
         case tidEditSetEditsPivot:
            t_result = create_table_EE_EditSetEditsPivot();
            break;
         case tidEditErrorsPivot:
            t_result = create_table_EE_EditErrorsPivot();
            break;
         case tidEditMessagesPivot:
            t_result = create_table_EE_EditMessagesPivot();
            break;
         case tidEditFieldsPivot:
            t_result = create_table_EE_EditFieldsPivot();
            break;
         case tidEditTablesPivot:
            t_result = create_table_EE_EditTablesPivot();
            break;
         case tidLayoutFieldsPivot:
            t_result = create_table_EE_LayoutFieldsPivot();
            break;
         default:
            // Caller used an invalid identifier; respond with failure
            t_result = false;
            break;
    }

    return t_result;
} // create_table
//---------------------------------------------------------------------------
bool TSMFTables::create_indexes(TSMFTableId TableId)
{
    bool t_result = true;

    FSQL.clear();

    switch (TableId)
    {
         case tidControl:
            FSQL += "create unique index xEE_Control_ItemName on EE_Control(ItemName);";
            break;
         case tidAgencies:
            FSQL += "create unique index xEE_Agencies_AdminCode on EE_Agencies(AdminCode);";
            FSQL += "create unique index xEE_Agencies_OrgName on EE_Agencies(OrgName);";
            break;
         case tidEdits:
            FSQL += "create index xEE_Edits_AgencyPKey on EE_Edits(AgencyPKey);";
            FSQL += "create index xEE_Edits_MessagePKey on EE_Edits(MessagePKey);";
            FSQL += "create unique index xEE_Edits_EditTag on EE_Edits(EditTag);";
            FSQL += "create unique index xEE_Edits_EditName on EE_Edits(EditName);";
            break;
         case tidEditSets:
            FSQL += "create index xEE_EditSets_AgencyPKey on EE_EditSets(AgencyPKey);";
            FSQL += "create unique index xEE_EditSets_EditSetTag on EE_EditSets(EditSetTag);";
            FSQL += "create unique index xEE_EditSets_EditSetName on EE_EditSets(EditSetName);";
            break;
         case tidFields:
            FSQL += "create index xEE_Fields_AgencyPKey on EE_Fields(AgencyPKey);";
            FSQL += "create unique index xEE_Fields_FieldNumber on EE_Fields(FieldNumber);";
			FSQL += "create unique index xEE_Fields_FieldName on EE_Fields(FieldName);";
			FSQL += "create unique index xEE_Fields_LocalName on EE_Fields(LocalName);";
			break;
		 case tidLayouts:
			FSQL += "create index xEE_Layouts_AgencyPKey on EE_Layouts(AgencyPKey);";
			FSQL += "create unique index xEE_Layouts_LayoutTag on EE_Layouts(LayoutTag);";
			FSQL += "create unique index xEE_Layouts_LayoutName on EE_Layouts(LayoutName);";
			break;
		 case tidMessages:
			FSQL += "create index xEE_Messages_AgencyPKey on EE_Messages(AgencyPKey);";
			FSQL += "create unique index xEE_Messages_MessageNumber on EE_Messages(MessageNumber);";
			FSQL += "create unique index xEE_Messages_Message on EE_Messages(Message);";
			break;
		 case tidLookupTables:
			FSQL += "create index xEE_LookupTables_AgencyPKey on EE_LookupTables(AgencyPKey);";
			FSQL += "create unique index xEE_LookupTables_TableName on EE_LookupTables(TableName);";
			break;
		 case tidEditSetEditsPivot:
			FSQL += "create index xEE_EditSetEditsPivot_EditSetPKey on EE_EditSetEditsPivot(EditSetPKey);";
			FSQL += "create index xEE_EditSetEditsPivot_EditPKey on EE_EditSetEditsPivot(EditPKey);";
			FSQL += "create unique index xEE_EditSetEditsPivot_NoDupes on EE_EditSetEditsPivot(EditSetPKey,EditPKey);";
			break;
		 case tidEditErrorsPivot:
			FSQL += "create index xEE_EditErrorsPivot_EditPKey on EE_EditErrorsPivot(EditPKey);";
			break;
		 case tidEditMessagesPivot:
			FSQL += "create index xEE_EditMessagesPivot_EditPKey on EE_EditMessagesPivot(EditPKey);";
			FSQL += "create index xEE_EditMessagesPivot_MessagePKey on EE_EditMessagesPivot(MessagePKey);";
			FSQL += "create unique index xEE_EditMessagesPivot_NoDupes on EE_EditMessagesPivot(EditPKey,MessagePKey);";
			break;
		 case tidEditFieldsPivot:
			FSQL += "create index xEE_EditFieldsPivot_EditPKey on EE_EditFieldsPivot(EditPKey);";
			FSQL += "create index xEE_EditFieldsPivot_FieldPKey on EE_EditFieldsPivot(FieldPKey);";
			FSQL += "create unique index xEE_EditFieldsPivot_NoDupes on EE_EditFieldsPivot(EditPKey,FieldPKey);";
			break;
		 case tidEditTablesPivot:
			FSQL += "create index xEE_EditTablesPivot_EditPKey on EE_EditTablesPivot(EditPKey);";
			FSQL += "create index xEE_EditTablesPivot_LookupTablePKey on EE_EditTablesPivot(LookupTablePKey);";
			FSQL += "create unique index xEE_EditTablesPivot_NoDupes on EE_EditTablesPivot(EditPKey,LookupTablePKey);";
			break;
		 case tidLayoutFieldsPivot:
			FSQL += "create index xEE_LayoutFieldsPivot_LayoutPKey on EE_LayoutFieldsPivot(LayoutPKey);";
			FSQL += "create index xEE_LayoutFieldsPivot_FieldPKey on EE_LayoutFieldsPivot(FieldPKey);";
			FSQL += "create unique index xEE_LayoutFieldsPivot_NoDupes on EE_LayoutFieldsPivot(LayoutPKey, FieldPkey);";

            break;
         default:
            // Caller used an invalid identifier; respond with failure
            t_result = false;
            break;
    }

    if (FSQL.length() > 0)
    {
		FQry->SetSQLText(FSQL.c_str());
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}
	}
    return t_result;
} // create_indexes
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_Control(const char* mf_structure_version)
{
	bool t_result = true;

	FSQL  = "DROP TABLE IF EXISTS EE_Control;\r\n";

	FSQL += "CREATE TABLE EE_Control( \r\n";
	FSQL += "[PKey] integer not null primary key, \r\n";
	FSQL += "[ItemName] varchar(30) not null, \r\n";
	FSQL += "[ItemValue] varchar(50) null \r\n";
	FSQL += ");";

	FQry->SetSQLText(FSQL.c_str());
	if (FQry->ExecSQL())
	{
		FSQL = "INSERT INTO EE_Control (ItemName, ItemValue) ";
		//FSQL += "VALUES ('MFStructureVersion', '5.00');\n";
		FSQL += "VALUES ('MFStructureVersion', '";
		FSQL += mf_structure_version;
		FSQL += "');\n";

		FSQL += "INSERT INTO EE_Control (ItemName, ItemValue) ";
		FSQL += "VALUES ('MFContentVersion', 'Metafile author to provide this');\n";

		FSQL += "INSERT INTO EE_Control (ItemName, ItemValue) ";
		FSQL += "VALUES ('MFComment', 'Metafile comment goes here');\n";

		FQry->SetSQLText(FSQL.c_str());
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}

	}
	else
	{
		t_result = false;
	}
	return t_result;
} // create_table_EE_Control()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_Agencies()
{
	bool t_result = true;

	FSQL  = "DROP TABLE IF EXISTS EE_Agencies;\r\n";

	FSQL += "CREATE TABLE EE_Agencies( \r\n";
	FSQL += "[PKey] integer not null primary key, \r\n";
	FSQL += "[AdminCode] varchar(12) not null collate nocase, \r\n";
	FSQL += "[OrgName] varchar(50) not null collate nocase, \r\n";

	// FSQL += "[Address1] varchar(50) null collate nocase, \r\n";
	// FSQL += "[Address2] varchar(50) null collate nocase, \r\n";
	// FSQL += "[City] varchar(30) null collate nocase, \r\n";
	// FSQL += "[State] char(2) null collate nocase, \r\n";
	// FSQL += "[PostalCode] varchar(10) null collate nocase, \r\n";
	// FSQL += "[ContactName] varchar(30) null collate nocase, \r\n";
	// FSQL += "[PhoneNumber] varchar(20) null collate nocase, \r\n";

	FSQL += "[Password] varchar(10) null collate nocase, \r\n";
	FSQL += "[LastModDate] datetime null, \r\n";
	FSQL += "[Comments] text null \r\n";
	// FSQL += "[AGENCY_ID] char(3) not null collate nocase default '000' \r\n";
	FSQL += ");";


	FQry->SetSQLText(FSQL.c_str());
	if (FQry->ExecSQL())
	{
		std::string t_insert_fields  = "[AdminCode],[OrgName],";
		t_insert_fields += "[Password],";
		t_insert_fields += "[LastModDate],[Comments]";

		// We want to ensure we have PKey 1 means AdminCode 'NONE' so we can use
		// it as a default for AgencyPKey in other control tables.
		std::string t_insert_values  = "'NONE','No Standard Setter', ";
		t_insert_values += "";
		t_insert_values += "NULL,NULL,NULL";

		FSQL = "INSERT INTO EE_Agencies(" + t_insert_fields + ") ";
		FSQL += "VALUES(" + t_insert_values + ");\n";

		FQry->SetSQLText(FSQL.c_str());
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}
	}
	else
	{
		t_result = false;
	}
	return t_result;
} // create_table_EE_Agencies()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_Edits()
{
	bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_Edits;\r\n";

    FSQL += "CREATE TABLE EE_Edits( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
	// EE_Agencies will set up a first record with PKey=1, meaning "NONE"
    FSQL += "[AgencyPKey] integer not null default 1 references EE_Agencies(PKey), \r\n";
    FSQL += "[IsMultiField] integer not null default 0, \r\n";
    FSQL += "[EditTag] varchar(20) not null collate nocase, \r\n";

    FSQL += "[EditName] varchar(100) not null collate nocase, \r\n";

    // EE_Messages will set up a first record with PKey=1, meaning "This edit has no default error message."
    FSQL += "[MessagePKey] integer not null default 1 references EE_Messages(PKey), \r\n";

    FSQL += "[LastModDate] datetime null, \r\n";
    FSQL += "[Description] text null, \r\n";
    FSQL += "[EditLogic] text null, \r\n";
    FSQL += "[CompiledLogic] text null, \r\n";
    FSQL += "[AdminNotes] text null \r\n";
    FSQL += ");";


    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_Edits()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_EditSets()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_EditSets;\r\n";

    FSQL += "CREATE TABLE EE_EditSets( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    // EE_Agencies will set up a first record with PKey=1, meaning "NONE"
    FSQL += "[AgencyPKey] integer not null default 1 references EE_Agencies(PKey), \r\n";
    FSQL += "[EditSetTag] varchar(20) not null collate nocase, \r\n";

    FSQL += "[EditSetName] varchar(100) not null collate nocase, \r\n";
    FSQL += "[LastModDate] datetime null, \r\n";
    FSQL += "[Description] text null \r\n";
    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_EditSets()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_Fields()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_Fields;\r\n";

    FSQL += "CREATE TABLE EE_Fields( \r\n";
	FSQL += "[PKey] integer not null primary key, \r\n";
    // EE_Agencies will set up a first record with PKey=1, meaning "NONE"
	FSQL += "[AgencyPKey] integer not null default 1 references EE_Agencies(PKey), \r\n";

	FSQL += "[FieldNumber] integer not null ";
    FSQL += "check(typeof(FieldNumber) = 'integer'), \r\n";

	FSQL += "[FieldName] varchar(50) not null collate nocase, \r\n";

	FSQL += "[LocalName] varchar(50) not null collate nocase, \r\n";

	FSQL += "[FieldLength] integer not null ";
	FSQL += "default 1 ";
	FSQL += "check(typeof(FieldLength) = 'integer'), \r\n";

	// FIX ME: I eliminated data types D3, D5, D6, D7, D9 and DA,
	// plus 'R ' (right-justified string). In ten years, I have NEVER
	// seen any of these used, and don't know if the current Engine
	// would even support them. Let's just get rid of them.
	// Also deleted 'D8'.
	FSQL += "[DataType] char(2) not null ";
	FSQL += "default 'S ' ";
	FSQL += "check(length(DataType) <= 2) ";
	FSQL += "check(DataType in ('S ', 'D1')), \r\n";

	FSQL += "[LastModDate] datetime null, \r\n";
	FSQL += "[Comments] text null \r\n";
	FSQL += ");";

	FQry->SetSQLText(FSQL.c_str());
	if (!FQry->ExecSQL())
	{
		t_result = false;
	}
	return t_result;
} // create_table_EE_Fields()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_Layouts()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_Layouts;\r\n";

    FSQL += "CREATE TABLE EE_Layouts( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    // EE_Agencies will set up a first record with PKey=1, meaning "NONE"
    FSQL += "[AgencyPKey] integer not null default 1 references EE_Agencies(PKey), \r\n";

    FSQL += "[LayoutTag] varchar(20) not null collate nocase, \r\n";

    FSQL += "[LayoutName] varchar(100) not null collate nocase, \r\n";

    FSQL += "[LastModDate] datetime null, \r\n";
    FSQL += "[Description] text null \r\n";
    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_Layouts()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_Messages()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_Messages;\r\n";

    FSQL += "CREATE TABLE EE_Messages( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    // EE_Agencies will set up a first record with PKey=1, meaning "NONE"
    FSQL += "[AgencyPKey] integer not null default 1 references EE_Agencies(PKey), \r\n";

    FSQL += "[MessageNumber] integer not null ";
    FSQL += "check(typeof(MessageNumber) = 'integer'), \r\n";

    FSQL += "[Message] varchar(100) not null collate nocase, \r\n";

    FSQL += "[LastModDate] datetime null \r\n";
    FSQL += ");";

	FQry->SetSQLText(FSQL.c_str());
	if (FQry->ExecSQL())
	{
		std::string t_insert_fields  = "[AgencyPKey],[MessageNumber],[Message],";
		t_insert_fields += "[LastModDate]";
		// EE_Messages will set up a first record with PKey=1,
		// meaning "This edit has no default error message."
		std::string t_insert_values  = "1,0,'This edit has no default error message',";
		t_insert_values += "NULL";

		FSQL += "INSERT INTO EE_Messages(" + t_insert_fields + ") ";
		FSQL += "VALUES(" + t_insert_values + ");";

		FQry->SetSQLText(FSQL.c_str());
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}
	}
	else
	{
        t_result = false;
    }
    return t_result;
} // create_table_EE_Messages()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_LookupTables()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_LookupTables;\r\n";

    FSQL += "CREATE TABLE EE_LookupTables( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    // EE_Agencies will set up a first record with PKey=1, meaning "NONE"
    FSQL += "[AgencyPKey] integer not null default 1 references EE_Agencies(PKey), \r\n";

    FSQL += "[TableName] varchar(50) not null collate nocase, \r\n";

    FSQL += "[RecCount] integer not null ";
    FSQL += "check(typeof(RecCount) = 'integer'), \r\n";

    FSQL += "[LastModDate] datetime null, \r\n";
    FSQL += "[Comments] text null \r\n";
    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_LookupTables()
//---------------------------------------------------------------------------
/****************************************************************************
 ***                          PIVOT TABLES                                ***
 ****************************************************************************/
bool TSMFTables::create_table_EE_EditSetEditsPivot()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_EditSetEditsPivot;\r\n";

    FSQL += "CREATE TABLE EE_EditSetEditsPivot( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    FSQL += "[EditPKey] integer not null references EE_Edits(PKey) on delete cascade, \r\n";
    FSQL += "[EditSetPKey] integer not null references EE_EditSets(PKey) on delete cascade \r\n";
    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_EditSetEditsPivot()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_EditErrorsPivot()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_EditErrorsPivot;\r\n";

    FSQL += "CREATE TABLE EE_EditErrorsPivot( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    FSQL += "[EditPKey] integer not null references EE_Edits(PKey) on delete cascade, \r\n";

    FSQL += "[ErrorText] varchar(80) not null collate nocase \r\n";
    FSQL += "check(length(ErrorText) <= 80), \r\n";

    FSQL += "[EditsFunc] char(1) not null ";
    FSQL += "check(EditsFunc in ('E','S')) \r\n";
    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_EditErrorsPivot()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_EditMessagesPivot()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_EditMessagesPivot;\r\n";

    FSQL += "CREATE TABLE EE_EditMessagesPivot( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    FSQL += "[EditPKey] integer not null references EE_Edits(PKey) on delete cascade, \r\n";
    FSQL += "[MessagePKey] integer not null references EE_Messages(PKey) on delete cascade, \r\n";

    FSQL += "[EditsFunc] char(1) not null ";
    FSQL += "check(EditsFunc in ('A','D','E','I','S','U','W')) \r\n";

    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_EditMessagesPivot()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_EditFieldsPivot()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_EditFieldsPivot;\r\n";

    FSQL += "CREATE TABLE EE_EditFieldsPivot( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    FSQL += "[EditPKey] integer not null references EE_Edits(PKey) on delete cascade, \r\n";
    FSQL += "[FieldPKey] integer not null references EE_Fields(PKey) on delete cascade, \r\n";

    FSQL += "[FieldOrder] integer not null ";
    FSQL += "check(typeof(FieldOrder) = 'integer') \r\n";

    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_EditFieldsPivot()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_EditTablesPivot()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_EditTablesPivot;\r\n";

    FSQL += "CREATE TABLE EE_EditTablesPivot( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    FSQL += "[EditPKey] integer not null references EE_Edits(PKey) on delete cascade, \r\n";
    FSQL += "[LookupTablePKey] integer not null references EE_LookupTables(PKey) on delete cascade \r\n";
    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_EditTablesPivot()
//---------------------------------------------------------------------------
bool TSMFTables::create_table_EE_LayoutFieldsPivot()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_LayoutFieldsPivot;\r\n";

    FSQL += "CREATE TABLE EE_LayoutFieldsPivot( \r\n";
    FSQL += "[PKey] integer not null primary key, \r\n";
    FSQL += "[LayoutPKey] integer not null references EE_Layouts(PKey) on delete cascade, \r\n";
    FSQL += "[FieldPKey] integer not null references EE_Fields(PKey) on delete cascade, \r\n";

    FSQL += "[StartPos] integer not null ";
    FSQL += "check(typeof(StartPos) = 'integer') \r\n";

    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_LayoutFieldsPivot()
//---------------------------------------------------------------------------
/*
bool TSMFTables::create_table_EE_TEMPLATE()
{
    bool t_result = true;

    FSQL  = "DROP TABLE IF EXISTS EE_TEMPLATE;\r\n";

    FSQL += "CREATE TABLE EE_TEMPLATE( ";
    FSQL += "[PKey] integer not null primary key, ";
    FSQL += "[EditPKey] integer not null references EE_Edits(PKey) on delete cascade, ";

    FSQL += ");";

    FQry->SetSQLText(FSQL.c_str());
    if (!FQry->ExecSQL())
    {
        t_result = false;
    }
    return t_result;
} // create_table_EE_TEMPLATE()
//---------------------------------------------------------------------------
*/

/*****************************************************************************
 ***                                TRIGGERS                               ***
 *****************************************************************************/
/*
bool TSMFTables::create_trigger_lastmoddate(TSMFTableId TableId)
{
// Example:
// CREATE TRIGGER IF NOT EXISTS update_ee_agencies_lastmoddate AFTER UPDATE on EE_Agencies
//  begin
//    UPDATE EE_Agencies
//    SET LASTMODDATE = (select strftime('%Y%m%d %H:%M:%S', 'now', 'localtime')
//    WHERE rowid=old.rowid;
//  end;
//
    bool t_result = true;
    std::string t_table_name;
    switch (TableId)
    {
        case tidAgencies:
            t_table_name.assign("EE_Agencies");
            break;

        case tidEdits:
            t_table_name.assign("EE_Edits");
            break;

        case tidEditSets:
            t_table_name.assign("EE_EditSets");
            break;

        case tidFields:
            t_table_name.assign("EE_Fields");
            break;

        case tidLayouts:
            t_table_name.assign("EE_Layouts");
            break;

        case tidMessages:
            t_table_name.assign("EE_Messages");
            break;

        case tidLookupTables:
            t_table_name.assign("EE_LookupTables");
            break;

        default:
            break;
    }

    if (t_table_name.length() > 0)
    {
        std::string t_sql,
                    t_errmsg;
        int t_pass = 0;

        // We want both an UPDATE and an INSERT trigger
        while (t_pass < 2)
        {
            if (t_pass == 0)
                t_sql = "CREATE TRIGGER IF NOT EXISTS update_"+t_table_name +"_lastmoddate AFTER UPDATE on "+t_table_name+"\r\n";
            else
                t_sql = "CREATE TRIGGER IF NOT EXISTS insert_"+t_table_name +"_lastmoddate AFTER INSERT on "+t_table_name+"\r\n";

            t_sql += "begin\r\n";
            t_sql += "    UPDATE "+t_table_name+"\r\n";
            t_sql += "    SET LastModDate = (select strftime('%Y%m%d %H:%M:%S', 'now', 'localtime'))\r\n";

            if (t_pass == 0)
                t_sql += "    WHERE rowid=old.rowid;\r\n";
            else
                t_sql += "    WHERE rowid=new.rowid;\r\n";

            t_sql += "  end;\r\n";

            FQry->SetSQLText(t_sql.c_str());
			FSQLDB->BeginTrans();
			try
			{
                if (FQry->ExecSQL())
                {
                    FSQLDB->CommitTrans();
                }
                else
                {
                    FSQLDB->RollbackTrans();
                    // Debugging
                    t_errmsg.assign(FQry->LastErrorMessage());
                    t_result = false;
                    break;
                }
            }
			catch(std::exception& e)
            {
                // Debugging
                t_errmsg.assign(e.what());
                FSQLDB->RollbackTrans();
                t_result = false;
                break;
            }
            t_pass++;
        }

    }

    return t_result;

} // create_trigger_lastmoddate
//---------------------------------------------------------------------------
*/
int TSMFTables::UpdateMetafile(const char* mf_structure_version,
	char* smf_init_error, const int error_buffer_size)
{
	// The EDITS API returns code 0 to mean "all is well", so initialize
	// the return value of this function to -1, the traditional failure code,
	// until we make it through the updates.
	int t_result = -1;
	std::string t_errors = "";

	t_errors.assign(FSQLDB->GetDBName());

	t_errors = "";

	// Here's my concept: This function will know what has to be updated
	// at each version level. We'll hold on to historical updates in case we
	// ever find ourselves needing to roll them forward a change at a time...
	// which will be tricky, and I'm not going to hurt my head thinking it out
	// right now, but at least we'll have what we need to know how we got here.

	///////////////////////
	/// Version 5.01
	///////////////////////
	if (std::strcmp(mf_structure_version, "5.01") == 0)
	{
		// This is the update where we dropped the "revert to EMF" fields from
		// the EE_* control tables.
		while (true)
		{
			if (!toggle_foreign_keys_pragma(false))
			{
				t_errors += "* unable to disable foreign keys\n";
				break;
			}

			//////////////////////
			// EE_Agencies
			//////////////////////
			if (!rename_table("EE_Agencies", "EE_Agencies_BAK"))
			{
				t_errors += "* unable to rename EE_Agencies\n";
				break;
			}

			if (!create_table_EE_Agencies())
			{
				t_errors += "* unable to create EE_Agencies\n";
				break;
			}

			if (!import_data_from_BAK(mf_structure_version, "EE_Agencies", "EE_Agencies_BAK"))
			{
				t_errors += "* unable to import data into EE_Agencies\n";
				break;
			}

			if (!drop_BAK_table("EE_Agencies_BAK"))
			{
				t_errors += "* unable to drop table EE_Agencies_BAK\n";
				break;
			}

			// NOTE 1: Create the indexes after dropping the BAK file,
			// or else you'll encounter a conflict about the index name
			// already in use.

			// NOTE 2: Of course, this strategy could also run afoul of
			// imported data violating an index constraint, but that isn't
			// supposed to happen, is it?
			if (!create_indexes(tidAgencies))
			{
				t_errors += "* unable to create indexes on EE_Agencies\n";
				break;
			}

			//////////////////////
			// EE_Edits
			//////////////////////
			if (!rename_table("EE_Edits", "EE_Edits_BAK"))
			{
				t_errors += "* unable to rename EE_Edits\n";
				break;
			}

			if (!create_table_EE_Edits())
			{
				t_errors += "* unable to create EE_Edits\n";
				break;
			}

			if (!import_data_from_BAK(mf_structure_version, "EE_Edits", "EE_Edits_BAK"))
			{
				t_errors += "* unable to import data into EE_Edits\n";
				break;
			}

			if (!drop_BAK_table("EE_Edits_BAK"))
			{
				t_errors += "* unable to drop table EE_Edits_BAK\n";
				break;
			}

			if (!create_indexes(tidEdits))
			{
				t_errors += "* unable to create indexes on EE_Edits\n";
				break;
			}

			//////////////////////
			// EE_EditSets
			//////////////////////
			if (!rename_table("EE_EditSets", "EE_EditSets_BAK"))
			{
				t_errors += "* unable to rename EE_EditSets\n";
				break;
			}

			if (!create_table_EE_EditSets())
			{
				t_errors += "* unable to create EE_EditSets\n";
				break;
			}

			if (!import_data_from_BAK(mf_structure_version, "EE_EditSets", "EE_EditSets_BAK"))
			{
				t_errors += "* unable to import data into EE_EditSets\n";
				break;
			}

			if (!drop_BAK_table("EE_EditSets_BAK"))
			{
				t_errors += "* unable to drop table EE_EditSets_BAK\n";
				break;
			}

			if (!create_indexes(tidEditSets))
			{
				t_errors += "* unable to create indexes on EE_EditSets\n";
				break;
			}

			//////////////////////
			// EE_Fields
			//////////////////////
			if (!rename_table("EE_Fields", "EE_Fields_BAK"))
			{
				t_errors += "* unable to rename EE_Fields\n";
				break;
			}

			if (!create_table_EE_Fields())
			{
				t_errors += "* unable to create EE_Fields\n";
				break;
			}

			if (!import_data_from_BAK(mf_structure_version, "EE_Fields", "EE_Fields_BAK"))
			{
				t_errors += "* unable to import data into EE_Fields\n";
				break;
			}

			if (!drop_BAK_table("EE_Fields_BAK"))
			{
				t_errors += "* unable to drop table EE_Fields_BAK\n";
				break;
			}

			if (!create_indexes(tidFields))
			{
				t_errors += "* unable to create indexes on EE_Fields\n";
				break;
			}

			//////////////////////
			// EE_Layouts
			//////////////////////
			if (!rename_table("EE_Layouts", "EE_Layouts_BAK"))
			{
				t_errors += "* unable to rename EE_Layouts\n";
				break;
			}

			if (!create_table_EE_Layouts())
			{
				t_errors += "* unable to create EE_Layouts\n";
				break;
			}

			if (!import_data_from_BAK(mf_structure_version, "EE_Layouts", "EE_Layouts_BAK"))
			{
				t_errors += "* unable to import data into EE_Layouts\n";
				break;
			}

			if (!drop_BAK_table("EE_Layouts_BAK"))
			{
				t_errors += "* unable to drop table EE_Layouts_BAK\n";
				break;
			}

			if (!create_indexes(tidLayouts))
			{
				t_errors += "* unable to create indexes on EE_Layouts\n";
				break;
			}

			//////////////////////
			// EE_Messages
			//////////////////////
			if (!rename_table("EE_Messages", "EE_Messages_BAK"))
			{
				t_errors += "* unable to rename EE_Messages\n";
				break;
			}

			if (!create_table_EE_Messages())
			{
				t_errors += "* unable to create EE_Messages\n";
				break;
			}

			if (!import_data_from_BAK(mf_structure_version, "EE_Messages", "EE_Messages_BAK"))
			{
				t_errors += "* unable to import data into EE_Messages\n";
				break;
			}

			if (!drop_BAK_table("EE_Messages_BAK"))
			{
				t_errors += "* unable to drop table EE_Messages_BAK\n";
				break;
			}

			if (!create_indexes(tidMessages))
			{
				t_errors += "* unable to create indexes on EE_Messages\n";
				break;
			}

			//////////////////////
			// EE_LookupTables
			//////////////////////
			if (!rename_table("EE_LookupTables", "EE_LookupTables_BAK"))
			{
				t_errors += "* unable to rename EE_LookupTables\n";
				break;
			}

			if (!create_table_EE_LookupTables())
			{
				t_errors += "* unable to create EE_LookupTables\n";
				break;
			}

			if (!import_data_from_BAK(mf_structure_version, "EE_LookupTables", "EE_LookupTables_BAK"))
			{
				t_errors += "* unable to import data into EE_LookupTables\n";
				break;
			}

			if (!drop_BAK_table("EE_LookupTables_BAK"))
			{
				t_errors += "* unable to drop table EE_LookupTables_BAK\n";
				break;
			}

			if (!create_indexes(tidLookupTables))
			{
				t_errors += "* unable to create indexes on EE_LookupTables\n";
				break;
			}

			// End of restructuring; turn foreign keys back on and update
			// MFStructureVersion
			if (!toggle_foreign_keys_pragma(true))
			{
				t_errors += "* unable to reinstate/enable foreign keys\n";
				break;
			}

			if (!update_mf_structure_version(mf_structure_version))
			{
				t_errors += "* unable to update metafile structure version in EE_Control\n";
			}

			// We made it... return a success code
			t_result = 0;

			// exit the loop
			break;
		} // while (true)
	}

	if (!t_errors.empty())
	{
		// Return the error messages to the caller
		std::strncpy(smf_init_error, t_errors.c_str(), error_buffer_size);
	}

	return t_result;
} // UpdateMetafile
//---------------------------------------------------------------------------

bool TSMFTables::toggle_foreign_keys_pragma(bool on_off)
{
	// We're going to be renaming tables. If we don't disable the
	// foreign keys pragma, the FKs will realign with the backup table...
	// and we don't want that! Setting to OFF puts the foreign keys into
	// suspense while we restructure and move things around, and setting
	// to ON realigns the FKs with the replacement tables. Magic.
	bool t_result = true;
	if (on_off)
	{
		FSQL.assign("PRAGMA foreign_keys=ON;");
	}
	else
	{
		FSQL.assign("PRAGMA foreign_keys=OFF;");
	}

	try
	{
		FQry->SetSQLText(FSQL.c_str());
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}
	}
	catch (...)
	{
		t_result = false;
	}

	return t_result;
} // toggle_foreign_keys_pragma
//---------------------------------------------------------------------------
bool TSMFTables::rename_table(const char* orig_name, const char* bak_name)
{
	bool t_result = true;

	char t_buffer[100];
	std::snprintf(t_buffer, 100,
		"ALTER TABLE %s RENAME TO %s;",
		orig_name, bak_name);

	try
	{
		FQry->SetSQLText(t_buffer);
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}
	}
	catch (...)
	{
		t_result = false;
	}

	return t_result;
} // rename_table
//---------------------------------------------------------------------------
bool TSMFTables::import_data_from_BAK(const char* mf_structure_version,
	const char* orig_name, const char* bak_name)
{
	bool t_result = false;

	FSQL.clear();

	if (std::strcmp(mf_structure_version, "5.01") == 0)
	{
		if (std::strcmp(orig_name, "EE_Agencies") == 0)
		{
			FSQL = "INSERT INTO EE_Agencies (" \
				   "PKey, AdminCode, OrgName, Password, LastModDate, Comments) " \
				   "SELECT PKey, AdminCode, OrgName, Password, LastModDate, " \
				   "Comments || char(10) || 'ADDRESS: ' || " \
				   "Address1 || char(10) || Address2 || char(10) || " \
				   "City || ' ' || State || ' ' || PostalCode || char(10) || " \
				   "'CONTACT: ' || ContactName || char(10) || " \
				   "'PHONE: ' || PhoneNumber " \
				   "FROM EE_Agencies_BAK " \
				   "WHERE PKey != 1;";
		}

		else if (std::strcmp(orig_name, "EE_Edits") == 0)
		{
			FSQL = "INSERT INTO EE_Edits ( " \
				"[PKey], " \
				"[AgencyPKey], " \
				"[IsMultiField], " \
				"[EditTag], " \
				"[EditName], " \
				"[MessagePKey], " \
				"[LastModDate], " \
				"[Description], " \
				"[EditLogic], " \
				"[CompiledLogic], " \
				"[AdminNotes] " \
				") " \
				"SELECT  " \
				"[PKey], " \
				"[AgencyPKey], " \
				"[IsMultiField], " \
				"[EditTag], " \
				"[EditName], " \
				"[MessagePKey], " \
				"[LastModDate], " \
				"[Description], " \
				"[EditLogic], " \
				"[CompiledLogic], " \
				"[AdminNotes] " \
				"FROM EE_Edits_BAK; ";
		}

		else if (std::strcmp(orig_name, "EE_EditSets") == 0)
		{
			FSQL = "INSERT INTO EE_EditSets ( " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[EditSetTag], " \
					"[EditSetName], " \
					"[LastModDate], " \
					"[Description] " \
					") " \
					"SELECT  " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[EditSetTag], " \
					"[EditSetName], " \
					"[LastModDate], " \
					"[Description] " \
					"FROM EE_EditSets_BAK; ";
		}

		else if (std::strcmp(orig_name, "EE_Fields") == 0)
		{
			FSQL = "INSERT INTO EE_Fields ( " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[FieldNumber], " \
					"[FieldName], " \
					"[LocalName], " \
					"[FieldLength], " \
					"[DataType], " \
					"[LastModDate], " \
					"[Comments] " \
					") " \
					"SELECT  " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[FieldNumber], " \
					"[FieldName], " \
					"[LocalName], " \
					"[FieldLength], " \
					"[DataType], " \
					"[LastModDate], " \
					"[Comments] " \
					"FROM EE_Fields_BAK; ";
		}

		else if (std::strcmp(orig_name, "EE_Layouts") == 0)
		{
			FSQL = "INSERT INTO EE_Layouts ( " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[LayoutTag], " \
					"[LayoutName], " \
					"[LastModDate], " \
					"[Description] " \
					") " \
					"SELECT  " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[LayoutTag], " \
					"[LayoutName], " \
					"[LastModDate], " \
					"[Description] " \
					"FROM EE_Layouts_BAK; ";
		}

		else if (std::strcmp(orig_name, "EE_Messages") == 0)
		{
			FSQL = "INSERT INTO EE_Messages ( " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[MessageNumber], " \
					"[Message], " \
					"[LastModDate] " \
					") " \
					"SELECT  " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[MessageNumber], " \
					"[Message], " \
					"[LastModDate] " \
					"FROM EE_Messages_BAK " \
					"WHERE [PKey] != 1;";
		}

		else if (std::strcmp(orig_name, "EE_LookupTables") == 0)
		{
			FSQL = "INSERT INTO EE_LookupTables ( " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[TableName], " \
					"[RecCount], " \
					"[LastModDate], " \
					"[Comments] " \
					") " \
					"SELECT  " \
					"[PKey], " \
					"[AgencyPKey], " \
					"[TableName], " \
					"[RecCount], " \
					"[LastModDate], " \
					"[Comments] " \
					"FROM EE_LookupTables_BAK;";
		}
	}

	if (!FSQL.empty())
	{
		try
		{
			FQry->SetSQLText(FSQL.c_str());
			if (FQry->ExecSQL())
			{
				t_result = true;
			}
		}
		catch (...)
		{
			t_result = false;
		}
	}
	return t_result;
} // import_data_from_BAK
//---------------------------------------------------------------------------
bool TSMFTables::drop_BAK_table(const char* bak_name)
{
	bool t_result = true;

	char t_buffer[100];
	std::snprintf(t_buffer, 100,
		"DROP TABLE %s;",
		bak_name);

	try
	{
		FQry->SetSQLText(t_buffer);
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}
	}
	catch (...)
	{
		t_result = false;
	}

	return t_result;
} // drop_BAK_table
//---------------------------------------------------------------------------
bool TSMFTables::update_mf_structure_version(const char* mf_structure_version)
{
	// This is the last step. If we make it through all the other updates,
	// we can reset the metafile structure version to whatever is current.
	bool t_result = true;

	char t_buffer[100];
	std::snprintf(t_buffer, 100,
		   "UPDATE EE_Control " \
		   "SET [ItemValue] = '%s' " \
		   "WHERE [ItemName] = 'MFStructureVersion';",
		   mf_structure_version);

	try
	{
		FQry->SetSQLText(t_buffer);
		if (!FQry->ExecSQL())
		{
			t_result = false;
		}
	}
	catch (...)
	{
		t_result = false;
	}



	return t_result;
} // update_mf_structure_version
//---------------------------------------------------------------------------

