//---------------------------------------------------------------------------

#ifndef SMFWriterH
#define SMFWriterH
//---------------------------------------------------------------------------
#ifdef __SQLEW__
#include <string>

#include "edit_external.h"
#include "TableStats.h"

// Forward declarations
class TSMF;
class TSQLDB;
class TSQLQuery;
class TEngineErrorsPump;
class TEw_AccessController;
class TTablesList;
class TTableDef;
class TDefineTable;

class TUpdateEditLogicObject
{
private:
	std::string FTableName;
	std::string FFieldName;
	std::string FOriginalValue;
	std::string FValue;
public:
	TUpdateEditLogicObject(const char* table_name,
						   const char* field_name,
						   const char* original_value,
						   const char* new_value);
	~TUpdateEditLogicObject() {;}

	const char* GetTableName();
	const char* GetFieldName();
	const char* GetOriginalValue();
	const char* GetNewValue();
};

class TImportMetafile
{
private:
	std::string FSourceName; // the source (attached) metafile
	std::string FAttachedDBAlias;

	bool FWasAlreadySuperUser;
	TSMF* FParent;
	TSQLDB* FSQLDB;
	TEngineErrorsPump* FEngineErrorsPump;
	TEw_AccessController* FAccessController;
	TTablesList* FTablesList;

	void* FCallbackOwner;
	CallbackImportMetafile FCallbackFunction;

	int attach_to_metafile(const std::string source_smf_name);
	int detach_from_metafile();

	// Get the import candidates by table name; the "match_on_field_name"
	// is for tables that have "Tags": EditTag, EditSetTag, LayoutTag.
	// Those can be matched on either the Name or the Tag, user's choice
	int get_agencies();
	int get_edits(const char* match_on_field_name);
	int get_edit_sets(const char* match_on_field_name);
	int get_fields();
	int get_layouts(const char* match_on_field_name);
	int get_messages();
	int get_tables();

	bool layout_fields_match(int src_layout_pkey, int targ_layout_pkey, bool match_on_tag);
	bool editset_edits_match(int src_editset_pkey, int targ_editset_pkey, bool match_on_tag);

public:
	TImportMetafile(TSMF* parent, TSQLDB* database, TEngineErrorsPump* pump, TEw_AccessController* access_controller, TTablesList* tables_list);
	~TImportMetafile();
	int Open(const char* source_metafile_name);
	int Close();
	int ImportLookupTable(const std::string table_name);
	int GetImportCandidates( const char* table_name, const char* match_on_field_name, void* owner, void* callback_func);
};

class TSMFWriter
{
private:
	TSMF* FParent;
	TSQLDB* FSQLDB;
	TEngineErrorsPump* FEngineErrorsPump;
	TEw_AccessController* FAccessController;
	TTablesList* FTablesList;

	TDefineTable* FDefineTable;

	void get_original_value(TTableDef* table, TFieldDef* field, std::string& original_value);
	bool update_edit_fields(TUpdateEditLogicObject* obj, const int field_pkey);
	bool update_edit_tables(TUpdateEditLogicObject* obj, const int table_pkey);
	bool update_edit_messages(TUpdateEditLogicObject* obj, const int message_pkey);

	int post_edit(TTableDef* table);
	int post_insert(TTableDef* table, int* new_pkey);

	bool is_system_table(TTableDef* table);
	int set_agency(TTableDef* table, std::string& old_admin_code, std::string& new_admin_code);

	bool have_access_rights(TEE_MCO mco, const int pkey);

	bool verify_delete_impact(std::string table_name, int delete_pkey);

	void compare_old_new_field_length(int field_pkey, int new_field_length);

	bool index_name_exists(const char* index_name);

public:
	TSMFWriter(TSMF* parent, TSQLDB* database, TEngineErrorsPump* pump, TEw_AccessController* access_controller, TTablesList* tables_list);
	~TSMFWriter();

	int Edit(const std::string table_name, const int pkey);
	int Insert(const std::string table_name);
	int Delete(const std::string table_name, const int delete_pkey);
	int Post(const std::string table_name, int* new_pkey);
	int Cancel(const std::string table_name);

	int ValidateUnique( const char* table_name, const int pkey, const char* field_name, const char* value, int* data_is_unique, int* dupe_pkey );

	int BeginTrans();
	int CommitTrans();
	int RollbackTrans();

	int AddEditSetsPivot(const int editset_pkey, const int edit_pkey);
	int DeleteEditSetsPivot(const int editset_pkey, const int edit_pkey);

	int AddLayoutsPivot(const int layout_pkey, const int field_pkey, const int start_pos);
	int DeleteLayoutsPivot(const int layout_pkey, const int field_pkey);

	int SetFieldByName(const std::string table_name, const std::string field_name, const std::string value);
	int SetFieldByNumber(const std::string table_name, const int column_number, const std::string value);

	int ClearEditLogicPivots( const int edit_pkey );

	// lookup table maintenance (creating and restructuring)
	int DefineTableStructure(const char* table_name, const int restructure_existing);
	int AddColumnDef(const char* col_name,
					 const char* copy_from_col_name,
					 const int data_type,
					 const int data_length,
					 const int is_nullable,
					 const int is_trigger_field,
					 const char* check_limit_to_list,
					 const char* check_range_low,
					 const char* check_range_high,
					 const char* default_value,
					 const int index_status);
	int CreateTable(int* lookup_pkey);
};

#endif // __SQLEW
#endif
