#ifndef SQLQUERYH
#define SQLQUERYH
#include "sqlite3.h"
#include <map>
#include <string>

// forward declaration
class TSQLDB;

class TSQLQuery
{
private:
	std::string FSQLText;
	std::string FLastErrorMessage;

	sqlite3_stmt* FStatement; // for sqlite3_prepare_v2, sqlite3_step
	int FColCount;
	int FRowCount;
	bool FEof;

	bool FIsActive;

	bool FIsParameterized;

	// This will require testing... but if the sqlite3_prepare_v2 calls
	// must always be matched by sqlite3_finalize calls, could we not
	// keep track of them and run all the finalize calls when we destroy
	// the query object?
	int FPreparedCount;

	// For SELECT queries, we need to store the ordinal positions
	// of fields by name, so we can use GetFieldByName to retrieve
	// data values. Sheesh.
	std::map<std::string, int> FColumnNames;
	std::map<int, std::string> FColumnNamesByNumber;


	std::string get_ucase_sql();
	int get_row_count(std::string ucase_sql);

	bool reset_statement();
	bool open_query();
	bool close_query();
	bool finalize_query();

	void set_is_parameterized(bool is_parameterized);

	int get_prepared_count();
	void incr_prepared_count();
	void decr_prepared_count();
	void force_finalize_query();

public:
	TSQLQuery(TSQLDB* database);
	~TSQLQuery();

	bool ExecSQL();
	bool FetchNext();
	// Return data as string
	char* GetFieldByName(std::string fldname);
	char* GetFieldByNumber(int column_number);
	// For int fields, return data as integer
	int GetIntFieldByName(std::string fldname);
	int GetIntFieldByNumber(int column_number);
	char* GetColumnName(int col_idx);
	char* GetColumnDataType( int col_idx );
	bool GetColumnConstraints( const char* table_name, const char* column_name, int& is_nullable, int& is_primary_key, int& is_autoinc);

	TSQLDB* SQLDB;
	char* GetSQLText();
	void SetSQLText(std::string sql_text, bool check_parameterized = true);

	bool GetIsActive();
	bool SetIsActive(bool is_active);

	bool GetIsParameterized();

	bool Eof();
	int ColCount();
	int RowCount();

	char* LastErrorMessage();

	// Support for parameterized queries
	bool PrepareParameterizedQuery();
	bool ResetParameterizedQuery();
	bool BindIntParameter(const int param_num, const int value);
	bool BindTextParameter(const int param_num, const char* value);
	bool OpenParameterizedQuery();
	bool ReleaseParameterizedQuery();

}; // class TSQLQuery
//---------------------------------------------------------------------------


#endif

