//---------------------------------------------------------------------------
#include <algorithm> // std::transform
#include <cctype>    // std::tolower
#include <cstring>   // std::strcmp

// For the Borland compilers, use #pragma hdrstop for precompiled headers (pch)
#ifdef BCB
	#pragma hdrstop
#endif

#include "TableStats.h"
//---------------------------------------------------------------------------
////////////////////////////////////////////////////////////
/// class TFieldDef
////////////////////////////////////////////////////////////
TFieldDef::TFieldDef(const char* field_name, const int column_order,
    const TEditDataType data_type, int column_size,
    bool is_nullable, bool is_primary_key, bool is_autoinc)
{
    FFieldName.assign(field_name);
    FColumnOrder = column_order;
    FDataType = data_type;
    FColumnSize = column_size;
    FIsNullable = is_nullable;
    FIsPrimaryKey = is_primary_key;
	FIsAutoInc = is_autoinc;
	FIsTriggerField = false; // see SetIsTriggerField
	FDefaultAsString = "";
	FLimitToList = ""; // CHECK constraint with list of allowed values
	FCheckBetween = ""; // CHECK constraint with range: BETWEEN 'X' AND 'Y'
#ifdef __SQLEW__
	FIsModified = false;
#endif

} // constructor
//---------------------------------------------------------------------------
TFieldDef::~TFieldDef()
{
    // does nothing
} // destructor
//---------------------------------------------------------------------------
char* TFieldDef::GetFieldName()
{
	return (char*)FFieldName.c_str();
}
//---------------------------------------------------------------------------
void TFieldDef::SetFieldName(const char* field_name)
{
	FFieldName.assign(field_name);
}
//---------------------------------------------------------------------------
int TFieldDef::GetColumnOrder()
{
    return FColumnOrder;
}
//---------------------------------------------------------------------------
TEditDataType TFieldDef::GetDataType()
{
    return FDataType;
}
//---------------------------------------------------------------------------
int TFieldDef::GetColumnSize()
{
    // FColumnSize will be 0 for data types OTHER THAN edtVarchar or edtChar
    return FColumnSize;
}
//---------------------------------------------------------------------------
bool TFieldDef::GetIsNullable()
{
    return FIsNullable;
}
//---------------------------------------------------------------------------
bool TFieldDef::GetIsPrimaryKey()
{
    return FIsPrimaryKey;
}
//---------------------------------------------------------------------------
bool TFieldDef::GetIsAutoInc()
{
    return FIsAutoInc;
}
//---------------------------------------------------------------------------
void TFieldDef::SetIsTriggerField(bool is_triggered)
{
	FIsTriggerField = is_triggered;
}
//---------------------------------------------------------------------------
bool TFieldDef::GetIsTriggerField()
{
	return FIsTriggerField;
}
//---------------------------------------------------------------------------
bool TFieldDef::HasDefaultValue()
{
	return !FDefaultAsString.empty();
}
//---------------------------------------------------------------------------
char* TFieldDef::GetDefaultValue()
{
	return (char*)FDefaultAsString.c_str();
}
//---------------------------------------------------------------------------

void TFieldDef::SetDefaultValue(const char* default_value)
{
	FDefaultAsString.assign(default_value);
}
//---------------------------------------------------------------------------

bool TFieldDef::HasLimitToList()
{
	return !FLimitToList.empty();
}
//---------------------------------------------------------------------------

char* TFieldDef::GetLimitToList()
{
	return (char*)FLimitToList.c_str();
}
//---------------------------------------------------------------------------

void TFieldDef::SetLimitToList(const char* limit_to_list)
{
	FLimitToList.assign(limit_to_list);
}
//---------------------------------------------------------------------------

bool TFieldDef::HasCheckBetween()
{
	return !FCheckBetween.empty();
}
//---------------------------------------------------------------------------

char* TFieldDef::GetCheckBetween()
{
	return (char*)FCheckBetween.c_str();
}
//---------------------------------------------------------------------------

void TFieldDef::SetCheckBetween(const char* check_between)
{
	FCheckBetween.assign(check_between);
}
//---------------------------------------------------------------------------

#ifdef __SQLEW__
bool TFieldDef::GetIsModified()
{
	return FIsModified;
}
//---------------------------------------------------------------------------
char* TFieldDef::GetValue()
{
	return (char*)FValue.c_str();
}
//---------------------------------------------------------------------------
void TFieldDef::SetValue(const char* value)
{
	FValue.assign(value);
	FIsModified = true;
}
//---------------------------------------------------------------------------
void TFieldDef::ClearWriteProperties()
{
	FValue.clear();
	FIsModified = false;
}
//---------------------------------------------------------------------------

#endif  // __SQLEW__

////////////////////////////////////////////////////////////
/// class TTableDef
////////////////////////////////////////////////////////////
/*   TTableDef constructor
 *      Stores the table name. The one for SQLEW.dll needs to talk to
 *      the access controller for the SMF object.
 */
TTableDef::TTableDef(const char* table_name)
{
	FTableName.assign(table_name);

	if (std::strcmp(table_name, "EE_Agencies") == 0)
		FMCO = _mco_Agencies;
	else if (std::strcmp(table_name, "EE_Edits") == 0)
		FMCO = _mco_Edits;
	else if (std::strcmp(table_name, "EE_EditSets") == 0)
		FMCO = _mco_EditSets;
	else if (std::strcmp(table_name, "EE_Fields") == 0)
		FMCO = _mco_Fields;
	else if (std::strcmp(table_name, "EE_Layouts") == 0)
		FMCO = _mco_Layouts;
	else if (std::strcmp(table_name, "EE_Messages") == 0)
		FMCO = _mco_Messages;
	else if (std::strcmp(table_name, "EE_LookupTables") == 0)
		FMCO = _mco_LookupTables;
	else if (std::strcmp(table_name, "EE_LayoutFieldsPivot") == 0)
		FMCO = _mco_LayoutFieldsPivot;
	else if (std::strcmp(table_name, "EE_EditSetEditsPivot") == 0)
		FMCO = _mco_EditSetEditsPivot;
	else if (std::strcmp(table_name, "EE_EditFieldsPivot") == 0)
		FMCO = _mco_EditFieldsPivot;
	else if (std::strcmp(table_name, "EE_EditTablesPivot") == 0)
		FMCO = _mco_EditTablesPivot;
	else if (std::strcmp(table_name, "EE_EditMessagesPivot") == 0)
		FMCO = _mco_EditMessagesPivot;
	else if (std::strcmp(table_name, "EE_EditErrorsPivot") == 0)
		FMCO = _mco_EditErrorsPivot;
	else if (std::strcmp(table_name, "EE_Control") == 0)
		FMCO = _mco_Control;
	else
		FMCO = _mco_Other;

	FState = ewInactive;
	FPostToPKey = 0;
} // constructor
//---------------------------------------------------------------------------

/*   TTableDef destructor
 *      Deletes each TFieldDef in its list, then clears the map.
 */
TTableDef::~TTableDef()
{
    std::map<int, TFieldDef*>::iterator it = FFieldsByColumnOrder.begin();
    while (it != FFieldsByColumnOrder.end())
    {
        delete (TFieldDef*)(*it).second;
        it++;
    }
    FFieldsByColumnOrder.clear();
} // destructor
//---------------------------------------------------------------------------
/*   TTableDef::GetTableName
 *      Returns the table name
 */
char* TTableDef::GetTableName()
{
	return (char*)FTableName.c_str();
}
//---------------------------------------------------------------------------
/*   TTableDef::GetMCO
 *      Returns the TEE_MCO (see TSMFWriter; it's for access rights control)
 */
TEE_MCO TTableDef::GetMCO()
{
	return FMCO;
}
//---------------------------------------------------------------------------
/*   TTableDef::GetColumnCount
 *      Returns the number of TFieldDef objects in its list
 */
int TTableDef::GetColumnCount()
{
    return (int)FFieldsByColumnOrder.size();
}
//---------------------------------------------------------------------------
/*   TTableDef::AddColumn
 *      Checks to see if a TFieldDef column using the same column order
 *      already exists in the list. If not, it adds it to FFieldsByColumnOrder.
 *      Otherwise, returns an error code (-1).
 */
int TTableDef::AddColumn(TFieldDef* field_def)
{
    int t_result = 0;

    int t_col_order = field_def->GetColumnOrder();
    if (!FFieldsByColumnOrder[t_col_order])
    {
        FFieldsByColumnOrder[t_col_order] = field_def;
        // We need to use case-sensitive matching here,
        // so (flip a coin) we'll lowercase the field names
        std::string t_fname = field_def->GetFieldName();
#ifdef BCB
        std::transform(t_fname.begin(),
                       t_fname.end(),
                       t_fname.begin(),
                       std::tolower);
#else
        // g++ uses tolower in global namespace
        std::transform(t_fname.begin(),
                       t_fname.end(),
                       t_fname.begin(),
                       ::tolower);
#endif
        FFieldsByName[t_fname] = field_def;
    }
    else
    {
        t_result = -1; // error condition
    }

    return t_result;
}
//---------------------------------------------------------------------------
/*   TTableDef::GetColumn   (by column order)
 *       Returns the TFieldDef using that column_order. If none found,
 *       returns NULL.
 *
 *   NOTE: The caller to this class will "know" that FFieldsByColumnOrder is 0-based,
 *   and will therefore request the field at "column order 1" by asking
 *   for column_order 0.
 */
TFieldDef* TTableDef::GetColumn(int column_order)
{
	TFieldDef* t_field = NULL;

	if ((column_order >= 0) && (column_order < (int)FFieldsByColumnOrder.size()))
	{
		t_field = FFieldsByColumnOrder[column_order];
	}

    return t_field;
} // GetColumn by column order
//---------------------------------------------------------------------------
/*   TTableDef::GetColumn   (by field name)
 *       Returns the TFieldDef using that field name. If none found,
 *       returns NULL.
 *
 *   NOTE: TTableDef maps FFieldsByName using lowercase(field_name).
 *         If the field name cannot be found in FFieldsByName, the
 *         result is NULL;
 */
TFieldDef* TTableDef::GetColumn(const char* field_name)
{
    std::string t_fname;
    t_fname.assign(field_name);
#ifdef BCB
    std::transform(t_fname.begin(),
                   t_fname.end(),
                   t_fname.begin(),
                   std::tolower);
#else
    std::transform(t_fname.begin(),
                   t_fname.end(),
                   t_fname.begin(),
                   ::tolower);
#endif
    TFieldDef* t_field = FFieldsByName[t_fname];

    return t_field;
} // GetColumn by name
//---------------------------------------------------------------------------
void TTableDef::AddIndex(const char* sql)
{
    FIndexDefs.push_back( sql );
} // AddIndex
//---------------------------------------------------------------------------
int TTableDef::GetIndexCount()
{
    return (int)FIndexDefs.size();
}
//---------------------------------------------------------------------------
/*   TTableDef::GetIndex
 *       Returns the sql for the idx'th index. If none found,
 *       returns an empty string.
 *
 *   NOTE: The caller to this class will "know" that FIndexDefs is 0-based.
 *   I'm hoping this won't be confusing, since getting a COLUMN definition
 *   does an adjustment to cope with the caller assuming a 1-based list.
 */
char* TTableDef::GetIndex(const int idx)
{
    if ((idx >= 0) && (idx < (int)FIndexDefs.size()))
    {
        return (char*)FIndexDefs[idx].c_str();
    }
    else
    {
		return (char*)"";
    }
} // GetIndex
//---------------------------------------------------------------------------
void TTableDef::AddTrigger(const char* sql)
{
    FTriggerDefs.push_back( sql );
} // AddTrigger
//---------------------------------------------------------------------------
int TTableDef::GetTriggerCount()
{
    return (int)FTriggerDefs.size();
} // GetTriggerCount
//---------------------------------------------------------------------------
/*   TTableDef::GetTrigger
 *       Returns the sql for the idx'th trigger. If none found,
 *       returns an empty string.
 *
 *   NOTE: The caller to this class will "know" that FTriggerDefs is 0-based.
 *   I'm hoping this won't be confusing, since getting a COLUMN definition
 *   does an adjustment to cope with the caller assuming a 1-based list.
 */
char* TTableDef::GetTrigger(const int idx)
{
    if ((idx >= 0) && (idx < (int)FTriggerDefs.size()))
    {
        return (char*)FTriggerDefs[idx].c_str();
    }
    else
    {
		return (char*)"";
    }
} // GetTrigger
//---------------------------------------------------------------------------

void TTableDef::AddForeignKey(const char* fkey)
{
    FForeignKeys.push_back( fkey );
} // AddForeignKey
//---------------------------------------------------------------------------
int TTableDef::GetForeignKeyCount()
{
    return (int)FForeignKeys.size();
} // GetForeignKeyCount
//---------------------------------------------------------------------------
/*   TTableDef::GetForeignKey
 *       Returns the fkey for the idx'th foreign key. If none found,
 *       returns an empty string.
 *
 *   NOTE: The caller to this class will "know" that FForeignKeys is 0-based.
 *   I'm hoping this won't be confusing, since getting a COLUMN definition
 *   does an adjustment to cope with the caller assuming a 1-based list.
 */
char* TTableDef::GetForeignKey(const int idx)
{
    if ((idx >= 0) && (idx < (int)FForeignKeys.size()))
    {
        return (char*)FForeignKeys[idx].c_str();
	}
    else
    {
		return (char*)"";
    }
} // GetForeignKey
//---------------------------------------------------------------------------

void TTableDef::AddCheck(const char* check_constraint)
{
    FCheckConstraints.push_back( check_constraint );
} // AddCheck
//---------------------------------------------------------------------------
int TTableDef::GetCheckCount()
{
    return (int)FCheckConstraints.size();
} // GetCheckCount
//---------------------------------------------------------------------------
/*   TTableDef::GetCheck
 *       Returns the fkey for the idx'th foreign key. If none found,
 *       returns an empty string.
 *
 *   NOTE: The caller to this class will "know" that FForeignKeys is 0-based.
 *   I'm hoping this won't be confusing, since getting a COLUMN definition
 *   does an adjustment to cope with the caller assuming a 1-based list.
 */
char* TTableDef::GetCheck(const int idx)
{
    if ((idx >= 0) && (idx < (int)FCheckConstraints.size()))
    {
        return (char*)FCheckConstraints[idx].c_str();
    }
    else
    {
		return (char*)"";
    }
} // GetCheck
//---------------------------------------------------------------------------

void TTableDef::AddDefault(const char* default_constraint)
{
    FDefaultConstraints.push_back( default_constraint );
} // AddDefault
//---------------------------------------------------------------------------
int TTableDef::GetDefaultCount()
{
    return (int)FDefaultConstraints.size();
} // GetDefaultCount
//---------------------------------------------------------------------------
/*   TTableDef::GetDefault
 *       Returns the default for the idx'th default in list. If none found,
 *       returns an empty string.
 *
 *   NOTE: The caller to this class will "know" that FDefaultConstraints is 0-based.
 *   I'm hoping this won't be confusing, since getting a COLUMN definition
 *   does an adjustment to cope with the caller assuming a 1-based list.
 */
char* TTableDef::GetDefault(const int idx)
{
    if ((idx >= 0) && (idx < (int)FDefaultConstraints.size()))
    {
        return (char*)FDefaultConstraints[idx].c_str();
    }
    else
    {
        return (char*)"";
    }
} // GetDefault
//---------------------------------------------------------------------------

#ifdef __SQLEW__
bool TTableDef::Edit(int pkey)
{
	if (IsEditing())
	{
		return false;
	}

	FPostToPKey = pkey;
	FState = ewEdit;
	return true;

} // Edit
//---------------------------------------------------------------------------
bool TTableDef::Insert()
{
	if (IsEditing())
	{
		return false;
	}

	FPostToPKey = 0;
	FState = ewInsert;
	return true;
} // Insert
//---------------------------------------------------------------------------
int TTableDef::GetPostToPKey()
{
	return FPostToPKey;
}
//---------------------------------------------------------------------------
bool TTableDef::Cancel()
{
	ClearWriteFields();

	FState = ewBrowse;

	FPostToPKey = 0;

	return true;

} // Cancel
//---------------------------------------------------------------------------
bool TTableDef::IsEditing()
{
	return ((FState == ewEdit) || (FState == ewInsert));
}
//---------------------------------------------------------------------------
TEw_DataSetState TTableDef::GetState()
{
	return FState;
}
//---------------------------------------------------------------------------
void TTableDef::ClearWriteFields()
{
	// Called after a Post() resolves to a successful transaction
	// (COMMIT or ROLLBACK, depending on the caller's preference).
	TFieldDef* t_field = NULL;
	size_t t_max_cols = FFieldsByColumnOrder.size();

	for (unsigned int t_idx = 0; t_idx < t_max_cols; t_idx++)
	{
		t_field = FFieldsByColumnOrder[t_idx];
		if (t_field)
		{
			t_field->ClearWriteProperties();
		}
	}
} // ClearWriteFields
//---------------------------------------------------------------------------

#endif // __SQLEW__

////////////////////////////////////////////////////////////
/// class TTablesList
////////////////////////////////////////////////////////////
/*   TTablesList constructor
 */
TTablesList::TTablesList()
{
    // do nothing constructor
}
//---------------------------------------------------------------------------
/*   TTablesList destructor
 *     Deletes each TTableDef in the list, then clears the map
 */
TTablesList::~TTablesList()
{
    std::map<std::string, TTableDef*>::iterator it = FTablesList.begin();
    while (it != FTablesList.end())
    {
        delete (TTableDef*)(*it).second;
        it++;
    }
    FTablesList.clear();
} // destructor
//---------------------------------------------------------------------------
/*   TTablesList::AddTable
 *      Determines whether a TTableDef using this same name already exists
 *      in the list. If not, it adds it. If so, it returns an error code (-1).
 */
int TTablesList::AddTable(TTableDef* table_def)
{
    int t_result = 0;
	std::string t_table_name;
	t_table_name.assign(table_def->GetTableName());
#ifdef BCB
	std::transform(t_table_name.begin(),
				   t_table_name.end(),
				   t_table_name.begin(),
				   std::tolower);
#else
	std::transform(t_table_name.begin(),
				   t_table_name.end(),
				   t_table_name.begin(),
				   ::tolower);
#endif
	if (!FTablesList[t_table_name])
	{
		FTablesList[t_table_name] = table_def;
	}
	else
	{
		t_result = -1; // error condition, table already in list
	}
	return t_result;
}
//---------------------------------------------------------------------------
int TTablesList::RemoveTable(TTableDef* table_def)
{
	// Called by TSMF::RenameLookupTable, this removes the definition
	// of the table under its OLD name.
	int t_result = 0;

	std::string t_table_name;
	t_table_name.assign(table_def->GetTableName());
#ifdef BCB
	std::transform(t_table_name.begin(),
				   t_table_name.end(),
				   t_table_name.begin(),
				   std::tolower);
#else
	std::transform(t_table_name.begin(),
				   t_table_name.end(),
				   t_table_name.begin(),
				   ::tolower);
#endif
	TTableDef* t_test_def = FTablesList[t_table_name];
	if (t_test_def && t_test_def == table_def)
	{
		// found it... delete the TTableDef object
		// and erase the element from the map
		delete t_test_def;
		t_test_def = NULL;
		FTablesList.erase(t_table_name);
	}

	return t_result;
}
//---------------------------------------------------------------------------
/*   TTablesList::GetTable
 *      If a table by that name can be found in the list, the TTableDef
 *      is returned to the caller. If not, NULL is returned.
 */
TTableDef* TTablesList::GetTable(std::string table_name)
{
    std::string t_table_name;
    t_table_name.assign(table_name);
#ifdef BCB
    std::transform(t_table_name.begin(),
                   t_table_name.end(),
                   t_table_name.begin(),
                   std::tolower);
#else
    std::transform(t_table_name.begin(),
                   t_table_name.end(),
                   t_table_name.begin(),
                   ::tolower);
#endif

    return FTablesList[t_table_name];
}
//---------------------------------------------------------------------------
int TTablesList::Count()
{
    return (int)FTablesList.size();
}
//---------------------------------------------------------------------------
std::map<std::string, TTableDef*> TTablesList::TablesList()
{
    return FTablesList;
}
//---------------------------------------------------------------------------

/****************************************************************************
 *** Support for defining new lookup tables and restructuring existing ones
 ****************************************************************************/
 #ifdef __SQLEW__
TDefineColumn::TDefineColumn(
	const char* col_name,
	const char* formatted_def,
	const char* copy_from_col_name)
{
	FColDstName.assign(col_name);
	// formatted_def will look like this:
	// FOOBAR varchar(10) NULL collate nocase check(FOOBAR IN ('foo','bar')) DEFAULT 'foo'
	// IOW, it is ready to add to a CREATE TABLE statement
	FFormattedDef.assign(formatted_def);
	// If copy_from_col_name is empty, we won't try to import data for this column
	// when restructuring an existing table. Otherwise, one of the changes might be
	// that the column name was changed... it is the job of the caller to keep
	// track of such shenanigans and tell us where to retrieve the data.
	if (copy_from_col_name)
		FColSrcName.assign(copy_from_col_name);
	else
		FColSrcName.assign("");

	FIsTriggerMaintained = 0; // default to false
} // ctor
//---------------------------------------------------------------------------
void TDefineColumn::SetIsTriggerMaintained(const int is_trigger_maintained)
{
	if (is_trigger_maintained == 0)
	{
		FIsTriggerMaintained = 0; // false
	}
	else
	{
		FIsTriggerMaintained = 1; // true
	}
} // SetIsTriggerMaintained
//---------------------------------------------------------------------------
TDefineTable::TDefineTable()
{
	FIsRestructuring = false;
} // ctor
//---------------------------------------------------------------------------
TDefineTable::~TDefineTable()
{
	ClearAll();
} // dtor
//---------------------------------------------------------------------------
void TDefineTable::ClearAll()
{
	FTableName.clear();
	FIsRestructuring = false;
	// delete the TDefineColumn instances
	std::vector<TDefineColumn*>::iterator it = FColumnDefs.begin();
	while (it != FColumnDefs.end())
	{
		delete (TDefineColumn*)(*it);
		it++;
	}

	// and clear the vectors
	FColumnDefs.clear();
	FIndexDefs.clear();
	FTriggerDefs.clear();

} // ClearAll
//---------------------------------------------------------------------------
void TDefineTable::SetTableName(const char* table_name)
{
	FTableName.assign(table_name);
}
//---------------------------------------------------------------------------
void TDefineTable::SetIsRestructuring(bool is_restructuring)
{
	FIsRestructuring = is_restructuring;
}
//---------------------------------------------------------------------------
size_t TDefineTable::GetColumnsCount()
{
	return FColumnDefs.size();
}
//---------------------------------------------------------------------------

void TDefineTable::AddColumnDef(TDefineColumn* col_def)
{
	FColumnDefs.push_back(col_def);
}
//---------------------------------------------------------------------------

TDefineColumn* TDefineTable::GetColumnDef(size_t idx)
{
	if (idx < FColumnDefs.size())
	{
		return FColumnDefs[idx];
	}

	return NULL;
}
//---------------------------------------------------------------------------

size_t TDefineTable::GetIndexCount()
{
	return FIndexDefs.size();
}
//---------------------------------------------------------------------------

void TDefineTable::AddIndexDef(std::string index_def)
{
	FIndexDefs.push_back(index_def);
}
//---------------------------------------------------------------------------

char* TDefineTable::GetIndexDef(size_t idx)
{
	if (idx < FIndexDefs.size())
	{
		return (char*)FIndexDefs[idx].c_str();
	}

	return NULL;
}
//---------------------------------------------------------------------------

size_t TDefineTable::GetTriggerCount()
{
	return FTriggerDefs.size();
}
//---------------------------------------------------------------------------

void TDefineTable::AddTriggerDef(std::string trigger_def)
{
	FTriggerDefs.push_back(trigger_def);
}
//---------------------------------------------------------------------------

char* TDefineTable::GetTriggerDef(size_t idx)
{
	if (idx < FTriggerDefs.size())
	{
		return (char*)FTriggerDefs[idx].c_str();
	}
	return NULL;
}
//---------------------------------------------------------------------------


#endif // __SQLEW__

