//---------------------------------------------------------------------------
#include <stdexcept>
#include <cstdlib> // std::atoi
#include <cstring> // std::memset

#include "TokenMaster.h"
#include "EngineErrorsPump.h"
//---------------------------------------------------------------------------
/*
 *     class TToken
 */
TToken::TToken(const char* original, TTokenID token_id, TTokenType token_type,
	char compiled)
{
	if (original)
		FOriginal.assign(original);
	else
		FOriginal.clear();

	FTokenID = token_id;
	FTokenType = token_type;

	FCompiled = compiled;
} // constructor
//---------------------------------------------------------------------------

/*
 *     class TCargo
 */
TCargo::TCargo(size_t string_size)
{
	// make the string at least 100 characters
	FCargoSize = (string_size < 100 ? 101 : string_size +1);
	FCargoString = (char*)std::malloc(FCargoSize);
	FCargoString[0] = '\x0';
} // ctor
//---------------------------------------------------------------------------
TCargo::~TCargo()
{
	std::free(FCargoString);
} // dtor
//---------------------------------------------------------------------------
size_t TCargo::GetCargoSize()
{
	return FCargoSize;
}
//---------------------------------------------------------------------------
intptr_t TCargo::GetCargoAddress()
{
	return (intptr_t)FCargoString;
}
//---------------------------------------------------------------------------
void TCargo::SetCargoString(const char* cargo_string, size_t cargo_len)
{
	size_t t_len = (cargo_len < FCargoSize ? cargo_len : FCargoSize);
	std::strncpy(FCargoString, cargo_string, t_len);
	FCargoString[t_len] = '\x0';
} // SetCargoString
//---------------------------------------------------------------------------
/*
 *     class TTokenMaster
 */
TTokenMaster::TTokenMaster(TEngineErrorsPump* pump)
{
	FEngineErrorsPump = pump;
	FCompiledLogic = NULL;
	FCompiledLength = 0;
	FPos = 0;

	FStringsVectorSize = 0;
	FTopOfStringsStack = 0;

	FOneChar[0] = ' ';
	FOneChar[1] = '\x0';

	// In 2015, most edits won't need more than 100 characters, so let's
	// not oversize the buffers until it becomes necessary
	expand_strings_vector();

	set_up_tokens();
} // constructor
//---------------------------------------------------------------------------
TTokenMaster::~TTokenMaster()
{
	// release the memory used by the strings in FStrings
	for (size_t t_idx = 0; t_idx < FStringsVectorSize; t_idx++)
	{
		delete FStrings[t_idx];
	}

	// We need to destroy the TToken objects
	for (size_t t_idx = 0; t_idx < 127; t_idx++)
	{
		TToken* t_token = FTokensByCompiled[t_idx];
		delete t_token;
	}
} // destructor
//---------------------------------------------------------------------------
void TTokenMaster::set_up_tokens()
{
	for (size_t t_idx = 0; t_idx < 127; t_idx++)
	{
		FTokensByCompiled[t_idx] = new TToken(" ", (TTokenID)0, (TTokenType)0, (char)32);
	}
	// original, token_id, token_type, compiled
	add_token("!",          NOT,     OPERATOR,'!');

	add_token("\"",         (TTokenID) STRING,  STRING,'\"');
	/* " + xx(length of string) + len bytes of string */

	add_token("#",          FIELD,   APPFIELD,'#'); // # + field_num

	add_token("mod",        MOD,     OPERATOR,'%');
	add_token("&&",         AND,     OPERATOR,'&');
	add_token("'",          CCONST,  CHARCONST,'\'');
	add_token("(",          LPAREN,  DELIMITER,'(');
	add_token(")",          RPAREN,  DELIMITER,')');
	add_token("*",          MULT,    OPERATOR,'*');
	add_token("+",          ADD,     OPERATOR,'+');
	add_token(",",          COMMA,   OPERATOR,',');
	add_token("-",          SUB,     OPERATOR,'-');
	add_token(".",          PERIOD,  DELIMITER,'.');
	add_token("/",          DIV,     OPERATOR,'/');
	add_token(":",          COLON,   DELIMITER,':');
	add_token(";",          SCOLON,  DELIMITER,';');
	add_token("<",          LT,      OPERATOR,'<');
	add_token("=",          ASSIGN,  OPERATOR,'=');
	add_token(">",          GT,      OPERATOR,'>');
	add_token("else",       ELSE,    CONSTRUCT,  'A');
	add_token("while",      WHILE,   CONSTRUCT,  'B');
	add_token("for",        FOR,     CONSTRUCT,  'C');
	add_token("return",     RETURN,  KEYWORD, 'D');
	add_token("==",         EQ,      OPERATOR,'E');
	add_token("if",         IF,      CONSTRUCT,  'F');
	add_token(">=",         GE,      OPERATOR,'G');
	add_token("char",       EEi_CHAR,      KEYWORD, 'H');
	add_token("int",        EEi_INT,    KEYWORD, 'I');
	add_token("!=",         NE,      OPERATOR,   'J');
	add_token("<=",         LE,      OPERATOR,   'K');
	add_token("do",         DO,      CONSTRUCT,  'L');
	add_token("libfnc",     LIBFUNC, FUNCTION,   'M');  // M + function ID letter
	add_token("var",        VAR,     IDENTIFIER, 'N');  // N + xx(length of id) + len bytes of name
	add_token("dt_today",   (TTokenID) 'O',     NUMCONST,   'O');
	add_token("dt_unknown", (TTokenID) 'P',     NUMCONST,   'P');
	add_token("fail",       (TTokenID) '0',     NUMCONST,   'Q');
	add_token("pass",       (TTokenID) '1',     NUMCONST,   'R'); // NOTE: also used for DT_MIN
	add_token("dt_max",     (TTokenID) '2',     NUMCONST,   'S');
	add_token("dt_exact",   (TTokenID) '3',     NUMCONST,   'T');
	add_token("number",     (TTokenID) 'U',     NUMBER,     'U'); // U + xx(length of number) +  len bytes of number
	add_token("\x5b",       LBRACKET,DELIMITER,  LBRACKET);
	add_token("\x5d",       RBRACKET,DELIMITER,  RBRACKET);
	add_token("funccall",   UDFUNCCALL, FUNCTION,   'm');  // m + function ID letter
	add_token("function",   UDFUNCDECL, FUNCTION,   'n');
	add_token("{",          LBRACE,  BLOCK,      '{');
	add_token("||",         OR,      OPERATOR,   '|');
	add_token("}",          RBRACE,  BLOCK,      '}');
	add_token("dt_empty",   (TTokenID) 'W',     NUMCONST,   'W');
	add_token("dt_valid",   (TTokenID) 'X',     NUMCONST,   'X');
	add_token("dt_missing", (TTokenID) 'Y',     NUMCONST,   'Y');
	add_token("dt_error",   (TTokenID) 'Z',     NUMCONST,   'Z');
	add_token("dt_day_empty",   (TTokenID) 'y',   NUMCONST,   'y');
	add_token("dt_month_empty", (TTokenID) 'z',   NUMCONST,   'z');
} // set_up_tokens
//---------------------------------------------------------------------------
void TTokenMaster::add_token( const char* original, const TTokenID token_id,
	const TTokenType token_type, const unsigned char compiled_char)
{
	TToken* t_token = FTokensByCompiled[compiled_char];

	t_token->SetOriginal(original);
	t_token->SetTokenID(token_id);
	t_token->SetTokenType(token_type);
	t_token->SetCompiled(compiled_char);
} // add_token
//---------------------------------------------------------------------------
void TTokenMaster::expand_strings_vector()
{
	if (FTopOfStringsStack >= FStringsVectorSize)
	{
		if (FStringsVectorSize == 0)
		{
			FStringsVectorSize = 100;
		}
		else
		{
			// increment the possible number of variables by 20;
			// it's an arbitrary number, but serves as an acceptable
			// growth rate
			FStringsVectorSize += 20;
		}
		FStrings.resize(FStringsVectorSize);
		for (size_t t_idx = FTopOfStringsStack; t_idx < FStringsVectorSize; t_idx++)
		{
			// Default string sizes to 100 characters plus NULL terminator;
			// the individual strings may be resized (increased) as needed
			FStrings[t_idx] = new TCargo(100);
		}
	}
} // expand_strings_vector
//---------------------------------------------------------------------------
void TTokenMaster::SetCompiledLogic(const char* logic)
{
	FCompiledLogic = logic;
	FCompiledLength = std::strlen(FCompiledLogic);

	FPos = 0;
} // SetCompiledLogic
//---------------------------------------------------------------------------
TToken* TTokenMaster::GetCompiledToken()
{
	TToken* t_token = NULL,
		  * t_test_token;

	unsigned char t_curr_tok = 0x32;
	char t_buffer[10];
	std::memset(t_buffer, '\x0', 10);

	size_t t_value_len;

	size_t t_old_pos = FPos;

	// From EditCompiler::tokenize()
	// Although we are removing limits on strings in this new Engine,
	// the compiler has to store the length of string constants. We're
	// allowing only up to 999 bytes, which one would hope would be
	// adequate. (Remember: You can build variables of indefinite
	// length out of string constants up to 999 bytes.) I HAD sized
	// t_string to be 101 bytes during development, and then sent
	// a string constant (to INLIST) that was 103 bytes. The Borland
	// compiler "let me get away with this", but the Microsoft compiler
	// blew up in my face... as it should have. Anyway, the fix is to
	// increase the declared size of this automatic variable to
	// accommodate the maximum size allowed by the EditCompiler.

	char t_string[1000]; // for CopyToCargo

	// Clear FOneChar to an blank, for ease of filling in the token
	FOneChar[0] = ' ';
	// FOneChar[1] = '\x0'; // the ctor already does this for us


	SkipNewLines();

	if (FPos < FCompiledLength)
	{
		// Retrieve the token
		t_curr_tok = FCompiledLogic[FPos];

		// Can we find a TToken for this t_curr_tok?

		t_test_token = FTokensByCompiled[t_curr_tok];

		if (t_test_token)
		{
			t_token = t_test_token;
			TTokenID t_token_id = t_token->GetTokenID();
			TTokenType t_token_type = t_token->GetTokenType();

			if (t_token_type == STRING)
			{
				// Assign the contents of the string to FCallCargo
				// The token is an open quote: "\""
				// The size of the string is indicated by the next
				// three characters, so that "abc" is represented as
				//                         /"003abc

				// Advance past the token
				FPos++;

				// This is NOT a string representation of a literal number.
				// We have to perform the dance to turn each ASCII byte back
				// into an integer FCallCargo.
				t_value_len  = (FCompiledLogic[FPos] - '0') * 100;
				t_value_len += ((FCompiledLogic[FPos+1] - '0') * 10) - t_value_len;
				t_value_len += (FCompiledLogic[FPos+2] - '0');

				FPos += 3;

				std::strncpy(t_string, &FCompiledLogic[FPos], t_value_len);
				t_string[t_value_len] = '\x0';
				CopyToCargo(t_string);
				// Advance past the size designation
				FPos += t_value_len;
			}
			else
			{
				switch (t_token_id)
				{
					// keyword
					case RETURN:

					// arithmetic ops
					case ADD:
					case DIV:
					case MOD:
					case MULT:
					case SUB:

					// logical ops
					case AND:
					case NOT:
					case OR:

					// punctuation, delimiters
					case COLON:
					case COMMA:
					case PERIOD:
					case SCOLON:
					case LBRACE:
					case RBRACE:
					case LBRACKET:
					case RBRACKET:
					case LPAREN:
					case RPAREN:

					// variable foo
					case EEi_CHAR:
					case EEi_INT:
					case ASSIGN:

					// relational ops
					case EQ:
					case GE:
					case GT:
					case LE:
					case LT:
					case NE:
						FOneChar[0] = t_curr_tok;
						FPos++;
						break;


					case IF:
					case ELSE:
					case WHILE:
					case FOR:
					case DO:
						// Advance past the token - we're putting the
						// "jump" into the FCallCargo parameter
						FPos++;

						std::strncpy(t_string, &FCompiledLogic[FPos], 4);
						t_string[4] = '\x0';
						CopyToCargo(t_string);
						// Advance past the "jump" to the opening
						// parenthesis of the condition
						FPos += 4;

						break;

					case FIELD:
						// will look like this: #2230
						FPos++;
						extract_field_number();

						break;

					case LIBFUNC:
						// This is an EDITS language function, token 'M'
						// The character that follows is an index into a
						// map of all functions in the language. Return
						// this index token in the FCallCargo.
						FOneChar[0] = FCompiledLogic[FPos+1];
						FPos += 2;

						break;

					case UDFUNCDECL:
						// Nothing needs to happen here... the caller will discover
						// that this is a UDF declaration from the token and handle
						// appropriately
						FPos++;

						break;

					case UDFUNCCALL:
						// Assign the alias for the function name to FCallCargo
						FOneChar[0] = FCompiledLogic[FPos+1];
						// and skip over the open/close parentheses
						FPos += 4;

						break;


					case CCONST:
						// An alpha character embedded in single-quotes: 'a'
						// Skip past the opening quote
						FPos++;
						// Turn this into a one-character string
						FOneChar[0] = FCompiledLogic[FPos];
						// And skip past the character and its closing quote
						FPos += 2;

						break;



					case VAR:
						// the token is 'N'; move past it
						FPos++;
						// Return the name of the variable to the caller
						std::strncpy(t_buffer, &FCompiledLogic[FPos], 3);
						t_buffer[3] = '\x0';
						t_value_len = std::atoi(t_buffer);
						FPos += 3;
						std::strncpy(t_string, &FCompiledLogic[FPos], t_value_len);
						t_string[t_value_len] = '\x0';
						CopyToCargo(t_string);

						FPos += t_value_len;

						break;


					default:
						if ((t_token_id == (TTokenID) '0') ||    // FAIL
							(t_token_id == (TTokenID) '1') ||    // PASS
							(t_token_id == (TTokenID) '2') ||    // DT_MAX
							(t_token_id == (TTokenID) '3') ||    // DT_EXACT
							(t_token_id == (TTokenID) 'O') ||    // DT_TODAY
							(t_token_id == (TTokenID) 'P') ||    // DT_UNKNOWN
							(t_token_id == (TTokenID) 'W') ||    // DT_EMPTY
							(t_token_id == (TTokenID) 'X') ||    // DT_VALID
							(t_token_id == (TTokenID) 'Y') ||    // DT_MISSING
							(t_token_id == (TTokenID) 'y') ||    // DT_ERROR
							(t_token_id == (TTokenID) 'Z') ||    // DT_DAY_EMPTY
							(t_token_id == (TTokenID) 'z'))      // DT_MONTH_EMPTY
						{
							FOneChar[0] = t_curr_tok;
							FPos++;
						}

						else if (t_token_id == (TTokenID) 'U')
						{
							// token is a NUMBER, described as U+xx for size of size, so
							// that "char str1[5]"  looks like U0015, but
							//      "char str1[56]" looks like U00256. So the first three
							// digits past the 'U' tell how many digits to read past
							// THOSE three digits.

							// Advance past the 'U'
							FPos++;
							std::strncpy(t_buffer, &FCompiledLogic[FPos], 3);
							t_buffer[3] = '\x0';
							t_value_len = std::atoi(t_buffer);
							FPos += 3;
							std::strncpy(t_string, &FCompiledLogic[FPos], t_value_len);
							t_string[t_value_len] = '\x0';
							CopyToCargo(t_string);
							// Advance past the size designation
							FPos += t_value_len;
						}

						break;

				} // end switch(t_token_id)
			} // if (t_token_type == STRING)
		} // if (t_test_token)
	} // if (FPos < FCompiledLength)

	// Just so we don't get stuck in an endless loop.
	if (FPos == t_old_pos)
		FPos += 1;

	// Let's do this instead of constantly testing the return value of
	// this function call.
	if ((FPos < FCompiledLength) && !t_token)
	{
		std::string t_error = "Token [ ] not found";
		t_error[7] = t_curr_tok;

		FEngineErrorsPump->AddError("TTokenMaster", "GetCompiledToken", t_error);

		throw std::runtime_error("");
	}

	return t_token;
} // GetCompiledToken
//---------------------------------------------------------------------------

TToken* TTokenMaster::GetHardToken( const TTokenID token_id )
{
	// This function checks to be sure the next token is the one the caller
	// is expecting. If it is not, the return value is NULL. The caller should
	// probably throw an error.

	TToken* t_token = GetCompiledToken() ;

	if (t_token && t_token->GetTokenID() != token_id)
	{
		t_token = NULL;
	}

	return t_token;
} // GetHardToken
//---------------------------------------------------------------------------
void TTokenMaster::extract_field_number()
{
	// Called by GetCompiledToken to move the field number from FCompiledLogic
	// into an entry in FStrings. There is no particular limit on
	// the number of digits in a field number, so we have to walk the code
	// until we hit something that is not a digit.
	char t_fld_num[20];
	t_fld_num[0] = '\x0';
	char t_char;
	for (int t_idx = 0; t_idx < 20 && FPos < FCompiledLength; t_idx++)
	{
		t_char = FCompiledLogic[FPos];
		if (t_char >= '0' && t_char <= '9')
		{
			FPos++;
			t_fld_num[t_idx] = t_char;
		}
		else
		{
			// Not a digit; we're done
			t_fld_num[t_idx] = '\x0';
			break;
		}
	}

	CopyToCargo(t_fld_num);
} // extract_field_number
//---------------------------------------------------------------------------
void TTokenMaster::ResetTopOfStringsStack()
{
	// Called by TRunEdits::interpret_block. Each "block", whether
	// delimited by a SCOLON or an RBRACE, reuses the FStrings vector
	// as a temporary scratch pad for string values.
	FTopOfStringsStack = 0;
} // ResetTopOfStringsStack
//---------------------------------------------------------------------------
intptr_t TTokenMaster::CopyToCargo(const char* cargo)
{
	// Make room in FStrings for this string
	if (FTopOfStringsStack >= FStringsVectorSize)
	{
		expand_strings_vector();
	}

	size_t t_cargo_len = std::strlen(cargo);
	// Check the size of the allocated string at the current position in FStrings
	if (t_cargo_len > FStrings[FTopOfStringsStack]->GetCargoSize())
	{
		// Need more room! Delete the string that's there
		delete FStrings[FTopOfStringsStack];
		// Create a new, bigger one
		TCargo* t_cargo = new TCargo(t_cargo_len);
		FStrings[FTopOfStringsStack] = t_cargo;
	}

	FStrings[FTopOfStringsStack]->SetCargoString(cargo, t_cargo_len);

	intptr_t t_retval = (intptr_t)FStrings[FTopOfStringsStack]->GetCargoAddress();

	FTopOfStringsStack += 1;

	return t_retval;

} // CopyToCargo
//---------------------------------------------------------------------------
intptr_t TTokenMaster::GetCallCargo()
{
	// For TLibFuncs::call_ routines that need to know what was stuffed into
	// FCallCargo by FTokenMaster::GetCompiledToken
	// Return the address of the most-recently assigned string in FStrings list.
	intptr_t t_retval = 0;
	if (FTopOfStringsStack > 0)
	{
		t_retval = FStrings[FTopOfStringsStack -1]->GetCargoAddress();
	}

	return t_retval;
}
//---------------------------------------------------------------------------
intptr_t TTokenMaster::GetOneChar()
{
	return (intptr_t)FOneChar;
}
//---------------------------------------------------------------------------
const char* TTokenMaster::GetCompiledLogic()
{
	// TRunEdits::setup_UDF_pointers needs to know
	return FCompiledLogic;
}
//---------------------------------------------------------------------------
size_t TTokenMaster::GetCompiledLength()
{
	// TRunEdits::interpret_block needs to know
	return FCompiledLength;
}
//---------------------------------------------------------------------------
void TTokenMaster::SavePos()
{
	FOldPos = FPos;
}
//---------------------------------------------------------------------------
void TTokenMaster::RestorePos()
{
	FPos = FOldPos;
}
//---------------------------------------------------------------------------
size_t TTokenMaster::GetPos()
{
	return FPos;
}
//---------------------------------------------------------------------------
void TTokenMaster::MovePosBy( int move_by )
{
	// This is called from TRunEdits for the CONSTRUCT handlers
	// (ExecuteIF, etc.) to advance FPos by "jump" positions
	if (move_by > 0)
	{
		size_t t_move_by = (size_t)move_by;
		if (FPos + t_move_by <= FCompiledLength)
		{
			FPos += t_move_by;
			return;
		}
		// If we get to here, this was an invalid call... scream!
		FEngineErrorsPump->AddError("TTokenMaster", "MovePosBy",
			"move_by value falls outside range of compiled logic");

		throw std::runtime_error("");
	}
	// And if we get to here, the caller tried to move the FPos back...
	// don't allow that!
	FEngineErrorsPump->AddError("TTokenMaster", "MovePos",
		"move_by must be > 0");

	throw std::runtime_error("");
} // MovePos
//---------------------------------------------------------------------------
void TTokenMaster::MovePosTo( size_t move_to )
{
	if (move_to < FCompiledLength)
	{
		FPos = move_to;
		return;
	}
	FEngineErrorsPump->AddError("TTokenMaster", "MovePosTo",
		"move_to must be < length of compiled logic");

	throw std::runtime_error("");
}
//---------------------------------------------------------------------------
void TTokenMaster::SkipNewLines()
{
	// Skip over newline characters
	while ((FPos < FCompiledLength) && (FCompiledLogic[FPos] == '\n'))
	{
		FPos++;
	}
}
//---------------------------------------------------------------------------

