//---------------------------------------------------------------------------
#include <cstdio> // snprintf
#include <cstring> // strncpy, strchr
#include <algorithm> // std::transform
#include <cctype> // std::tolower, toupper
#include <sstream> // std::ostringstream
#include <ctime>   // date/time foo

#include "Utility.h"
//---------------------------------------------------------------------------

void lowercase( std::string& in_string )
{
#ifdef BCB
	std::transform(in_string.begin(),
		   in_string.end(),
		   in_string.begin(),
		   std::tolower);
#else
        // g++ uses the tolower in the global namespace
	std::transform(in_string.begin(),
		   in_string.end(),
		   in_string.begin(),
		   ::tolower);
#endif
} // lowercase
//---------------------------------------------------------------------------

void uppercase( std::string& in_string )
{
#ifdef BCB
	std::transform(in_string.begin(),
		   in_string.end(),
		   in_string.begin(),
		   std::toupper);
#else
        // g++ uses the toupper in the global namespace
	std::transform(in_string.begin(),
		   in_string.end(),
		   in_string.begin(),
		   ::toupper);
#endif
} // uppercase
//---------------------------------------------------------------------------

std::string lowercase_string( std::string in_string )
{
	std::string t_out_string;
	t_out_string.assign(in_string);
	lowercase(t_out_string);
	return t_out_string;
} // lowercase_string
//---------------------------------------------------------------------------

std::string uppercase_string( std::string in_string )
{
	std::string t_out_string;
	t_out_string.assign(in_string);
	uppercase(t_out_string);
	return t_out_string;
} // uppercase_string
//---------------------------------------------------------------------------

std::string trim(const std::string& str,
				 const std::string& whitespace /* = " \t" */)
{
	size_t strBegin = str.find_first_not_of(whitespace);
	if (strBegin == std::string::npos)
		return ""; // no content

	size_t strEnd = str.find_last_not_of(whitespace);
	size_t strRange = strEnd - strBegin + 1;

	return str.substr(strBegin, strRange);
}
//---------------------------------------------------------------------------
std::string ltrim(const std::string& str,
				 const std::string& whitespace /* = " \t" */)
{
	size_t strBegin = str.find_first_not_of(whitespace);
	if (strBegin == std::string::npos)
		return ""; // no content

	return str.substr(strBegin, str.length());
}
//---------------------------------------------------------------------------
std::string rtrim(const std::string& str,
				  const std::string& whitespace /* = " \t" */)
{
	size_t strEnd = str.find_last_not_of(whitespace);
	if (strEnd == std::string::npos)
		return ""; // no content

	return str.substr(0, strEnd +1);
}
//---------------------------------------------------------------------------
std::string padright(const std::string inString, unsigned int cnt, char padchar /* = ' ' */)
{
	std::string pad(cnt, padchar);
	std::string s;

	s = inString + pad;

	return s.substr(0, cnt);
}
//---------------------------------------------------------------------------
std::string padleft(const std::string inString, unsigned int cnt, char padchar /* = ' ' */)
{
	std::string pad(cnt, padchar);
	std::string s = pad + inString;
	return s.substr(s.size() - cnt, cnt);
}
//---------------------------------------------------------------------------
std::string intToString( int value )
{
	// Another option would have been to use std::itoa <cstdlib>. But this works.
	std::ostringstream oss;
	oss << value;
	return oss.str();
}
//---------------------------------------------------------------------------
std::string stringReplace(const std::string instring, const std::string oldpatt,
	const std::string newpatt)
{
	std::string outstring = instring;
	std::string::size_type pos = outstring.find(oldpatt);
	size_t len = oldpatt.size();
	while (pos != std::string::npos)
	{
		outstring.replace(pos, len, newpatt);
		pos = outstring.find(oldpatt, pos+len+1);
	}
	return outstring;
}  // stringReplace
//---------------------------------------------------------------------------

// case-insensitive version of stringReplace
// i.e., the oldpatt will be searched case-insensitively

#define is_white(c) (c == ' '|| c == '\t'|| c == '\r'|| c == '\n')
#define is_delim(c) (strchr(" !;,+-<>'%^=()[]{}",c) || c==9 || c=='\r' || c==0)

std::string ci_stringReplace(const std::string instring,
	const std::string oldpatt, const std::string newpatt,
	const bool whole_word /* = false */)
{
	// For example:
	//    oldpatt = "AbCDefg"
	//    outstring contains "aBcdEFG" which should still replace with newpatt
	std::string outstring;
	size_t t_instring_pos = 0,
		   t_oldpatt_pos = 0,
		   t_instring_len = instring.length(),
		   t_oldpatt_len = oldpatt.length();
	std::string t_compare = "";

	// default c1 to whitespace in case pattern is the first string embedded
	// in instring and whole_word is true.
	char c1 = ' ';
	char c2;

	while (t_instring_pos < t_instring_len)
	{
		if (t_instring_pos > 0)
			c1 = std::toupper(instring[t_instring_pos-1]);
		c2 = std::toupper(instring[t_instring_pos]);

		if (!t_compare.empty())
		{
			// If we're working on a section that matches oldpatt,
			// keep going until we either hit the end, or determine
			// it isn't a match after all.
			while ( (std::toupper(instring[t_instring_pos]) ==
					 std::toupper(oldpatt[t_oldpatt_pos]) ) &&
					(t_oldpatt_pos < t_oldpatt_len) )
			{
				t_compare += oldpatt[t_oldpatt_pos++];
				t_instring_pos++;
			}

			if (t_compare.compare(oldpatt) == 0)
			{
				outstring += newpatt;
			}
			else
			{
				// not a match after all... append what we have to
				// outstring and carry on
				outstring += t_compare;
			}

			t_compare = "";
			t_oldpatt_pos = 0;
		}
		else
		{
			if (whole_word)
			{
				// All bets are off if c2 does not start a new word
				if (is_white(c1) || is_delim(c1))
				{
					// Have we located the start of oldpatt?
					if (c2 == std::toupper(oldpatt[0]))
					{
						t_compare += oldpatt[t_oldpatt_pos++];
					}
				}
			}
			else
			{
				// In this case, we don't care about whole_word;
				// if we found a matching character, add it to
				// t_compare; otherwise, append it to outstring
				if (std::toupper(instring[t_instring_pos]) ==
					std::toupper(oldpatt[0]))
				{
					t_compare += oldpatt[t_oldpatt_pos++];
				}
			}

            // If we didn't start a t_compare, then append to the outstring
			if (t_compare.empty())
			{
				outstring += instring[t_instring_pos];
			}

			t_instring_pos++;

		} // if (!t_compare.empty())

	} // while (t_instring_pos < t_instring_len)

	return outstring;
}  // ci_stringReplace
//---------------------------------------------------------------------------

// case-insensitive "find" for strings. optionally ensures that the pattern
// is found as a whole word, and not as part of another word.

size_t ci_stringFind(const std::string instring, const std::string pattern,
	const bool whole_word /* = false */)
{
	size_t t_found_pos = std::string::npos,
		   t_pattern_start = std::string::npos;

	size_t t_instring_len = instring.length(),
		   t_pattern_len = pattern.length(),
		   t_pos = 0,
		   t_patt_pos = 0;

	std::string t_compare = "";
	// default c1 to whitespace in case pattern is the first string embedded
	// in instring and whole_word is true.
	char c1 = ' ';
	char c2;

	while (t_pos < t_instring_len)
	{
		if (t_pos > 0)
			c1 = std::toupper(instring[t_pos-1]);
		c2 = std::toupper(instring[t_pos]);

		if (!t_compare.empty())
		{
			while ( (std::toupper(instring[t_pos]) ==
					 std::toupper(pattern[t_patt_pos])) &&
					(t_patt_pos < t_pattern_len))
			{
				t_compare += pattern[t_patt_pos++];
				t_pos++;
			}

			// Did we find it?
			if (t_compare.compare(pattern) == 0)
			{
				t_found_pos = t_pattern_start;
				break;
			}
			else
			{
				// Not a match after all; reset and carry on looking
				t_compare = "";
				t_patt_pos = 0;
			}

		}
		else
		{
			if (whole_word)
			{
				// all bets are off if c2 does not start a new word
				if (is_white(c1) || is_delim(c1))
				{
					// Have we located the start of the pattern?
					if (c2 == std::toupper(pattern[0]))
					{
						t_compare += pattern[t_patt_pos++];
						t_pattern_start = t_pos;
					}
				}
			}
			else
			{
				// in this case, we don't care about whole_word;
				// if we found a matching character, add it to t_compare
				if (c2 == std::toupper(pattern[0]))
				{
					t_compare += pattern[t_patt_pos++];
					t_pattern_start = t_pos;
				}

			}  // if (whole_word)
		} // if (!t_compare.empty())

		t_pos++;

	} // while (t_pos < t_instring_len)

	return t_found_pos;

} // ci_stringFind
//---------------------------------------------------------------------------

int posix_stricmp(const char *s1, const char *s2)
{
	// Function "stricmp" is available on Windows, but not Linux.
	// This is an implementation we used in the old Edit Engine.
    while (toupper(*s1++) == toupper(*s2++))
    {
	if (*s1 == '\x0' && *s2 == '\x0')
	    return 0;
    }

    if (toupper(*s1) < toupper(*s2))
	return -1;
    else
        return 1;

} // posix_stricmp
//---------------------------------------------------------------------------

int get_int_val(const char* s, size_t start_pos, size_t length)
{
	char t_buffer[101];
	for (size_t t_idx = 0; (t_idx < 100) && (t_idx < length); t_idx++)
	{
		t_buffer[t_idx] = s[start_pos + t_idx];
	}

	if (length < 100)
	{
		t_buffer[length] = '\x0';
	}
	else
	{
		t_buffer[100] = '\x0';
	}
	return std::atoi(t_buffer);
} // get_int_val
//---------------------------------------------------------------------------

void get_current_date_time(char* s, size_t length)
{
	// posix date/time string formatted the way we are using LastModDate in
	// metafile control tables
	time_t now = time(0);
	struct tm t_struct;
	char buf[80];
	t_struct = *localtime(&now);
	strftime(buf, sizeof(buf), "%Y%m%d %H:%M:%S", &t_struct);
	std::strncpy(s, buf, length);
} // get_current_date_time
//---------------------------------------------------------------------------

#ifdef EE_WINDOWS

#include <windows.h>

std::string system_error_message( int error_code )
{
	std::string t_message;
	int t_len;
	char t_buffer[255];

	t_len = FormatMessage(
		FORMAT_MESSAGE_FROM_SYSTEM,  // dwFlags
		NULL,                        // lpSource
		error_code,                  // dwMessageId
		0,                           // dwLanguageId
		t_buffer,                    // lpBuffer
		sizeof(t_buffer),            // nSize
		NULL);                       // Arguments

	if (!t_len)
	{
		_snprintf(t_buffer, 255, "FormatMessage failed with %d", (int)GetLastError());
	}

	t_message.assign(t_buffer);

	return t_message;
}
#endif

