//---------------------------------------------------------------------------
#include <stdexcept>
#include <cstdio>  // std::snprintf
#include <cstring> // std::strncpy

#include "Variable.h"
//---------------------------------------------------------------------------
/*
 *     class TVariable
 */
TVariable::TVariable(const char* var_name, const int var_type,
	const size_t max_size)
{
	FVariableName.assign(var_name);
	FVariableType = var_type;
	FStringValue = 0;

	if (var_type & EEVT_CHAR)
	{
		FMaxLength = max_size;
		FStringValue = (intptr_t)((char*)std::malloc(FMaxLength+1));

		std::strncpy((char*)FStringValue, "", FMaxLength);
	}
	else
	{
		FValues.resize(max_size);
		for (size_t t_idx = 0; t_idx < max_size; t_idx++)
		{
			// Initialize the values to zeroes
			FValues[t_idx] = 0;
		}
	}

} // constructor
//---------------------------------------------------------------------------
TVariable::~TVariable()
{
	if (FStringValue > 0)
	{
		std::free((char*)FStringValue);
		FStringValue = 0;
	}
} // destructor
//---------------------------------------------------------------------------
void TVariable::Reset(const char* var_name, const int var_type,
	const size_t max_size)
{
	FVariableName.assign(var_name);
	FVariableType = var_type;

	if (var_type & EEVT_CHAR)
	{
		FMaxLength = max_size;
		if (FStringValue > 0)
		{
			std::free((char*)FStringValue);
		}
		FStringValue = (intptr_t)((char*)std::malloc(FMaxLength+1));
		std::strncpy((char*)FStringValue, "", FMaxLength);
	}
	else
	{
		FValues.resize(max_size);
		for (size_t t_idx = 0; t_idx < max_size; t_idx++)
		{
			// Initialize the values to zeroes
			FValues[t_idx] = 0;
		}
	}
} // Reset
//---------------------------------------------------------------------------
intptr_t TVariable::GetStringValueAsIntPtr()
{
	return FStringValue;
}
//---------------------------------------------------------------------------

void TVariable::SetValue( const char* value )
{
	size_t t_value_len = std::strlen(value);
	if (t_value_len <= FMaxLength)
	{
		std::strncpy((char*)FStringValue, value, t_value_len);
		((char*)FStringValue)[t_value_len] = '\x0';
	}
	else
	{
		//std::strncpy((char*)FStringValue, value, FMaxLength);

		char t_buffer[200];
		std::snprintf(t_buffer, 200,
			"TVariable::SetValue - Variable %s, value \"%s\" longer than declared size (%d)",
			FVariableName.c_str(), value, (int)FMaxLength);

		throw std::runtime_error(t_buffer);

	}
} // SetValue
//---------------------------------------------------------------------------
void TVariable::SetTableVarValue( const char* value, size_t length )
{
	// TableVars are a different animal. At the time of compilation, the
	// length of a field cannot be definitively determined (somebody could
	// change the table structure after the edit is compiled). So TABLEVARs
	// are declared with size 1. If we set the value (see TLibFuncs::internal_lookup)
	// we won't check for a FMaxLength, but we will assign the new FMaxLength

	if (length > FMaxLength)
	{
		std::free((char*)FStringValue);

		FMaxLength = length;

		FStringValue = (intptr_t)((char*)std::malloc(FMaxLength+1));
		std::strncpy((char*)FStringValue, "", FMaxLength);
	}
	std::strncpy((char*)FStringValue, value, FMaxLength);
	((char*)FStringValue)[FMaxLength] = '\x0';
} // SetTableVarValue
//---------------------------------------------------------------------------

void TVariable::SetValue( const size_t idx, const int value )
{
	if (idx < FValues.size())
	{
		FValues[idx] = value;
	}
} // SetValue
//---------------------------------------------------------------------------
int TVariable::GetValue(const size_t idx)
{
	int t_result = 0;
	if (idx < FValues.size())
	{
		t_result = FValues[idx];
	}
	else
	{
		throw std::runtime_error("TVariable::GetValue - index out of range");
	}
	return t_result;

} // GetValue
//---------------------------------------------------------------------------
size_t TVariable::Count()
{
	return FValues.size();
}
//---------------------------------------------------------------------------

