#ifndef edit_external_h
#define edit_external_h

///////////////////////////////////////////////////////////////////////////
/// EDITS50.dll includes this header file in order to return appropriate
/// data via callback function pointers submitted by the calling application
/// via parameters to certain functions of the EDITS50 API.
///////////////////////////////////////////////////////////////////////////

// edit option specifiers
//      Constant           Hex               Dec
#define EE_NOSKIP          0x0000         //   0 All edits run
#define EE_SKIPEMPTY       0x0001         //   1 Edit skips if any field is empty
#define EE_SKIPFAIL        0x0002         //   2 Multi-field edits skip if single-field edit failed
#define EE_SUPPRESSWARN    0x0080         // 128 Suppresses warnings (the edit will run, but warnings
										  //     will not be returned to the caller)

/* Create a callback function pointer type. Arguments include a void*
 * pointer to the object making the callback; in practice, this will be the
 * "this" pointer of the GUI form that will display the contents of
 * the char* arg passed to it. If you don't need it, pass NULL.
 */
typedef void (*CallbackInt)( void* ptr2object, const int AnInteger );
typedef void (*CallbackString)( void* ptr2object, const char* AString );

typedef void (*CallbackPKeys)( void* ptr2object, const char* table_name, const int pkey );

typedef void (*CallbackTableStructure)( void* ptr2Object, const char* table_name,
	const int column_order, const char* field_name, const char* data_type,
	const char* nullability, int is_trigger_field, const char* default_value,
	const char* check_limit_to_list, const char* check_between);

typedef void (*CallbackReadData)( void* ptr2Object, const char* table_name,
	const int pkey, const int column_order, const char* field_name,
	const char* data_type, const char* data_string );

typedef void (*CallbackDisplayDataRecord)( void* ptr2Object, const int pkey,
	const int start_pos, const int field_number, const int field_macro_order,
	const char* field_name, const char* display_name, const int field_length,
	const char* data_type, const char* value );

typedef void (*CallbackCheckEditSetAgainstLayout)( void* ptr2Object,
	const int field_pkey, const char* field_name,
	const int edit_pkey, const char* edit_name, const char* edit_tag );

typedef void (*CallbackRunEdits)( void* ptr2object, const char* edit_tag,
	const char* edit_name, const char* admin_code, const char* error_type,
	const char* message );
typedef void (*CallbackDebugFieldsFunction)( void* ptr2object, const char* AString );
typedef void (*CallbackDebugFunction)( void* ptr2object, const int line_number,
	const int line_executed, const char* AString );
typedef void (*CallbackDebugVarsFunction)( void* ptr2object, const int line_number,
	const int iteration, const char* variable, const char* var_value, const int is_string_var );

typedef void (*CallbackGetRunEditsSummary)( void* ptr2object,
	const char* edit_tag, const char* edit_name, const char* admin_code,
	const int total_errors, const int total_warnings, const int total_skips,
	const int total_messages, const int total_warning_messages,
	const int edit_skipped_silently, const int edit_is_suppressed,
	const int edit_logic_has_syntax_error);

typedef void (*CallbackImportMetafile)( void* ptr2object,
	const char* table_name,
	const int source_pkey,
	const int target_pkey,
	const char* record_key,
	const char* admin_code,
	const int status );

#endif
